@@ -37,7 +37,7 @@ def get_build_type() -> str:
 
 
 def refresh_logger() -> None:
-    """Refreshes Daft's internal rust logging to the current python log level"""
+    """Refreshes Daft's internal rust logging to the current python log level."""
     _refresh_logger()
 
 
@@ -50,17 +50,16 @@ __version__ = get_version()
 
 from daft.analytics import init_analytics
 
-dev_build = get_build_type() == "dev"
 user_opted_out = os.getenv("DAFT_ANALYTICS_ENABLED") == "0"
-if not dev_build and not user_opted_out:
-    analytics_client = init_analytics(get_version(), get_build_type())
-    analytics_client.track_import()
+analytics_client = init_analytics(get_version(), get_build_type(), user_opted_out)
+analytics_client.track_import()
 
 ###
 # Daft top-level imports
 ###
 
-from daft.context import set_execution_config, set_planning_config
+from daft.catalog import read_table, register_table
+from daft.context import set_execution_config, set_planning_config, execution_config_ctx, planning_config_ctx
 from daft.convert import (
     from_arrow,
     from_dask_dataframe,
@@ -73,14 +72,13 @@ from daft.daft import ImageFormat, ImageMode, ResourceRequest
 from daft.dataframe import DataFrame
 from daft.logical.schema import Schema
 from daft.datatype import DataType, TimeUnit
-from daft.expressions import Expression, col, lit
+from daft.expressions import Expression, col, lit, interval, coalesce
 from daft.io import (
     DataCatalogTable,
     DataCatalogType,
     from_glob_path,
     read_csv,
     read_deltalake,
-    read_delta_lake,
     read_hudi,
     read_iceberg,
     read_json,
@@ -89,42 +87,53 @@ from daft.io import (
     read_lance,
 )
 from daft.series import Series
+from daft.sql import sql, sql_expr
 from daft.udf import udf
 from daft.viz import register_viz_hook
 
+to_struct = Expression.to_struct
+
 __all__ = [
-    "from_pylist",
-    "from_pydict",
-    "from_arrow",
-    "from_pandas",
-    "from_ray_dataset",
-    "from_dask_dataframe",
-    "from_glob_path",
-    "read_csv",
-    "read_json",
-    "read_parquet",
-    "read_hudi",
-    "read_iceberg",
-    "read_deltalake",
-    "read_delta_lake",
-    "read_sql",
-    "read_lance",
-    "DataCatalogType",
     "DataCatalogTable",
+    "DataCatalogType",
     "DataFrame",
-    "Expression",
-    "col",
     "DataType",
-    "ImageMode",
+    "Expression",
     "ImageFormat",
-    "lit",
-    "Series",
-    "TimeUnit",
-    "register_viz_hook",
-    "refresh_logger",
-    "udf",
+    "ImageMode",
     "ResourceRequest",
     "Schema",
-    "set_planning_config",
+    "Series",
+    "TimeUnit",
+    "coalesce",
+    "col",
+    "execution_config_ctx",
+    "from_arrow",
+    "from_dask_dataframe",
+    "from_glob_path",
+    "from_pandas",
+    "from_pydict",
+    "from_pylist",
+    "from_ray_dataset",
+    "interval",
+    "lit",
+    "planning_config_ctx",
+    "read_csv",
+    "read_deltalake",
+    "read_hudi",
+    "read_iceberg",
+    "read_json",
+    "read_lance",
+    "read_parquet",
+    "read_sql",
+    "read_table",
+    "refresh_logger",
+    "register_table",
+    "register_viz_hook",
     "set_execution_config",
+    "set_planning_config",
+    "sql",
+    "sql_expr",
+    "to_struct",
+    "udf",
 ]
