@@ -15,8 +15,6 @@ import urllib.error
 import urllib.request
 from typing import Any, Callable
 
-from daft import context
-
 _ANALYTICS_CLIENT = None
 _WRITE_KEY = "ZU2LLq6HFW0kMEY6TiGZoGnRzogXBUwa"
 _SEGMENT_BATCH_ENDPOINT = "https://api.segment.io/v1/batch"
@@ -62,8 +60,8 @@ def _build_segment_batch_payload(
     }
 
 
-def _post_segment_track_endpoint(payload: dict[str, Any]) -> None:
-    """Posts a batch of JSON data to Segment"""
+def _post_segment_track_endpoint(analytics_client: AnalyticsClient, payload: dict[str, Any]) -> None:
+    """Posts a batch of JSON data to Segment."""
     req = urllib.request.Request(
         _SEGMENT_BATCH_ENDPOINT,
         method="POST",
@@ -74,21 +72,29 @@ def _post_segment_track_endpoint(payload: dict[str, Any]) -> None:
         },
         data=json.dumps(payload).encode("utf-8"),
     )
-    resp = urllib.request.urlopen(req)
+    try:
+        resp = urllib.request.urlopen(
+            req,
+            timeout=1,
+        )
+    except urllib.error.URLError:
+        analytics_client._is_active = False
     if resp.status != 200:
         raise RuntimeError(f"HTTP request to segment returned status code: {resp.status}")
 
 
 class AnalyticsClient:
-    """Non-threadsafe client for sending analytics events, which is a singleton for each Python process"""
+    """Non-threadsafe client for sending analytics events, which is a singleton for each Python process."""
 
     def __init__(
         self,
         daft_version: str,
         daft_build_type: str,
-        publish_payload_function: Callable[[dict[str, Any]], None] = _post_segment_track_endpoint,
+        enabled: bool,
+        publish_payload_function: Callable[[AnalyticsClient, dict[str, Any]], None] = _post_segment_track_endpoint,
         buffer_capacity: int = 100,
     ) -> None:
+        self._is_active = enabled
         self._daft_version = daft_version
         self._daft_build_type = daft_build_type
         self._session_key = _get_session_key()
@@ -101,21 +107,24 @@ class AnalyticsClient:
         self._buffer: list[AnalyticsEvent] = []
 
     def _append_to_log(self, event_name: str, data: dict[str, Any]) -> None:
-        self._buffer.append(
-            AnalyticsEvent(
-                session_id=self._session_key,
-                event_name=event_name,
-                event_time=datetime.datetime.utcnow(),
-                data=data,
+        if self._is_active:
+            self._buffer.append(
+                AnalyticsEvent(
+                    session_id=self._session_key,
+                    event_name=event_name,
+                    event_time=datetime.datetime.utcnow(),
+                    data=data,
+                )
             )
-        )
-        if len(self._buffer) >= self._buffer_capacity:
-            self._flush()
+            if len(self._buffer) >= self._buffer_capacity:
+                self._flush()
+        else:
+            logger.debug("Analytics is disabled; not sending data to segment")
 
     def _flush(self) -> None:
         try:
             payload = _build_segment_batch_payload(self._buffer, self._daft_version, self._daft_build_type)
-            self._publish(payload)
+            self._publish(self, payload)
         except Exception as e:
             # No-op on failure to avoid crashing the program - TODO: add retries for more robust logging
             logger.debug("Error in analytics publisher thread: %s", e)
@@ -126,7 +135,6 @@ class AnalyticsClient:
         self._append_to_log(
             "Imported Daft",
             {
-                "runner": context.get_context().runner_config.name,
                 "platform": platform.platform(),
                 "python_version": platform.python_version(),
                 "DAFT_ANALYTICS_ENABLED": os.getenv("DAFT_ANALYTICS_ENABLED"),
@@ -160,24 +168,26 @@ class AnalyticsClient:
         )
 
 
-def init_analytics(daft_version: str, daft_build_type: str) -> AnalyticsClient:
-    """Initialize the analytics module
+def init_analytics(daft_version: str, daft_build_type: str, user_opted_out: bool) -> AnalyticsClient:
+    """Initialize the analytics module.
 
     Returns:
         AnalyticsClient: initialized singleton AnalyticsClient
     """
+    enabled = (not user_opted_out) and daft_build_type != "dev"
+
     global _ANALYTICS_CLIENT
 
     if _ANALYTICS_CLIENT is not None:
         return _ANALYTICS_CLIENT
 
-    _ANALYTICS_CLIENT = AnalyticsClient(daft_version, daft_build_type)
+    _ANALYTICS_CLIENT = AnalyticsClient(daft_version, daft_build_type, enabled)
     atexit.register(_ANALYTICS_CLIENT._flush)
     return _ANALYTICS_CLIENT
 
 
 def time_df_method(method):
-    """Decorator to track metrics about Dataframe method calls"""
+    """Decorator to track metrics about Dataframe method calls."""
 
     @functools.wraps(method)
     def tracked_method(*args, **kwargs):
@@ -203,7 +213,7 @@ def time_df_method(method):
 
 
 def time_func(fn):
-    """Decorator to track metrics for daft API calls"""
+    """Decorator to track metrics for daft API calls."""
 
     @functools.wraps(fn)
     def tracked_fn(*args, **kwargs):
