@@ -1,8 +1,10 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
+from typing import TYPE_CHECKING
 
-from daft.table import MicroPartition
+if TYPE_CHECKING:
+    from daft.table import MicroPartition
 
 
 @dataclass(frozen=True)
