@@ -1,40 +1,17 @@
 from __future__ import annotations
 
-import builtins
-from typing import TYPE_CHECKING
-
-import pyarrow as pa
+import threading
+from typing import TYPE_CHECKING, Union
 
 from daft.context import get_context
 from daft.daft import ImageMode, PyDataType, PyTimeUnit
+from daft.dependencies import pa
 
 if TYPE_CHECKING:
+    import builtins
+
     import numpy as np
 
-_RAY_DATA_EXTENSIONS_AVAILABLE = True
-_TENSOR_EXTENSION_TYPES = []
-try:
-    import ray
-except ImportError:
-    _RAY_DATA_EXTENSIONS_AVAILABLE = False
-else:
-    _RAY_VERSION = tuple(int(s) for s in ray.__version__.split(".")[0:3])
-    try:
-        # Variable-shaped tensor column support was added in Ray 2.1.0.
-        if _RAY_VERSION >= (2, 2, 0):
-            from ray.data.extensions import (
-                ArrowTensorType,
-                ArrowVariableShapedTensorType,
-            )
-
-            _TENSOR_EXTENSION_TYPES = [ArrowTensorType, ArrowVariableShapedTensorType]
-        else:
-            from ray.data.extensions import ArrowTensorType
-
-            _TENSOR_EXTENSION_TYPES = [ArrowTensorType]
-    except ImportError:
-        _RAY_DATA_EXTENSIONS_AVAILABLE = False
-
 
 class TimeUnit:
     _timeunit: PyTimeUnit
@@ -97,7 +74,7 @@ class TimeUnit:
 
 
 class DataType:
-    """A Daft DataType defines the type of all the values in an Expression or DataFrame column"""
+    """A Daft DataType defines the type of all the values in an Expression or DataFrame column."""
 
     _dtype: PyDataType
 
@@ -107,6 +84,40 @@ class DataType:
             "use a creator method like DataType.int32() or use DataType.from_arrow_type(pa_type)"
         )
 
+    @classmethod
+    def _infer_type(cls, user_provided_type: DataTypeLike) -> DataType:
+        from typing import get_args, get_origin
+
+        if isinstance(user_provided_type, DataType):
+            return user_provided_type
+        elif isinstance(user_provided_type, dict):
+            return DataType.struct({k: DataType._infer_type(user_provided_type[k]) for k in user_provided_type})
+        elif get_origin(user_provided_type) is not None:
+            origin_type = get_origin(user_provided_type)
+            if origin_type is list:
+                child_type = get_args(user_provided_type)[0]
+                return DataType.list(DataType._infer_type(child_type))
+            elif origin_type is dict:
+                (key_type, val_type) = get_args(user_provided_type)
+                return DataType.map(DataType._infer_type(key_type), DataType._infer_type(val_type))
+            else:
+                raise ValueError(f"Unrecognized Python origin type, cannot convert to Daft type: {origin_type}")
+        elif isinstance(user_provided_type, type):
+            if user_provided_type is str:
+                return DataType.string()
+            elif user_provided_type is int:
+                return DataType.int64()
+            elif user_provided_type is float:
+                return DataType.float64()
+            elif user_provided_type is bytes:
+                return DataType.binary()
+            elif user_provided_type is object:
+                return DataType.python()
+            else:
+                raise ValueError(f"Unrecognized Python type, cannot convert to Daft type: {user_provided_type}")
+        else:
+            raise ValueError(f"Unable to infer Daft DataType for provided value: {user_provided_type}")
+
     @staticmethod
     def _from_pydatatype(pydt: PyDataType) -> DataType:
         dt = DataType.__new__(DataType)
@@ -115,79 +126,79 @@ class DataType:
 
     @classmethod
     def int8(cls) -> DataType:
-        """Create an 8-bit integer DataType"""
+        """Create an 8-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.int8())
 
     @classmethod
     def int16(cls) -> DataType:
-        """Create an 16-bit integer DataType"""
+        """Create an 16-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.int16())
 
     @classmethod
     def int32(cls) -> DataType:
-        """Create an 32-bit integer DataType"""
+        """Create an 32-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.int32())
 
     @classmethod
     def int64(cls) -> DataType:
-        """Create an 64-bit integer DataType"""
+        """Create an 64-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.int64())
 
     @classmethod
     def uint8(cls) -> DataType:
-        """Create an unsigned 8-bit integer DataType"""
+        """Create an unsigned 8-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.uint8())
 
     @classmethod
     def uint16(cls) -> DataType:
-        """Create an unsigned 16-bit integer DataType"""
+        """Create an unsigned 16-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.uint16())
 
     @classmethod
     def uint32(cls) -> DataType:
-        """Create an unsigned 32-bit integer DataType"""
+        """Create an unsigned 32-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.uint32())
 
     @classmethod
     def uint64(cls) -> DataType:
-        """Create an unsigned 64-bit integer DataType"""
+        """Create an unsigned 64-bit integer DataType."""
         return cls._from_pydatatype(PyDataType.uint64())
 
     @classmethod
     def float32(cls) -> DataType:
-        """Create a 32-bit float DataType"""
+        """Create a 32-bit float DataType."""
         return cls._from_pydatatype(PyDataType.float32())
 
     @classmethod
     def float64(cls) -> DataType:
-        """Create a 64-bit float DataType"""
+        """Create a 64-bit float DataType."""
         return cls._from_pydatatype(PyDataType.float64())
 
     @classmethod
     def string(cls) -> DataType:
-        """Create a String DataType: A string of UTF8 characters"""
+        """Create a String DataType: A string of UTF8 characters."""
         return cls._from_pydatatype(PyDataType.string())
 
     @classmethod
     def bool(cls) -> DataType:
-        """Create the Boolean DataType: Either ``True`` or ``False``"""
+        """Create the Boolean DataType: Either ``True`` or ``False``."""
         return cls._from_pydatatype(PyDataType.bool())
 
     @classmethod
     def binary(cls) -> DataType:
-        """Create a Binary DataType: A string of bytes"""
+        """Create a Binary DataType: A string of bytes."""
         return cls._from_pydatatype(PyDataType.binary())
 
     @classmethod
     def fixed_size_binary(cls, size: int) -> DataType:
-        """Create a FixedSizeBinary DataType: A fixed-size string of bytes"""
+        """Create a FixedSizeBinary DataType: A fixed-size string of bytes."""
         if not isinstance(size, int) or size <= 0:
             raise ValueError("The size for a fixed-size binary must be a positive integer, but got: ", size)
         return cls._from_pydatatype(PyDataType.fixed_size_binary(size))
 
     @classmethod
     def null(cls) -> DataType:
-        """Creates the Null DataType: Always the ``Null`` value"""
+        """Creates the Null DataType: Always the ``Null`` value."""
         return cls._from_pydatatype(PyDataType.null())
 
     @classmethod
@@ -197,7 +208,7 @@ class DataType:
 
     @classmethod
     def date(cls) -> DataType:
-        """Create a Date DataType: A date with a year, month and day"""
+        """Create a Date DataType: A date with a year, month and day."""
         return cls._from_pydatatype(PyDataType.date())
 
     @classmethod
@@ -221,9 +232,14 @@ class DataType:
             timeunit = TimeUnit.from_str(timeunit)
         return cls._from_pydatatype(PyDataType.duration(timeunit._timeunit))
 
+    @classmethod
+    def interval(cls) -> DataType:
+        """Interval DataType."""
+        return cls._from_pydatatype(PyDataType.interval())
+
     @classmethod
     def list(cls, dtype: DataType) -> DataType:
-        """Create a List DataType: Variable-length list, where each element in the list has type ``dtype``
+        """Create a List DataType: Variable-length list, where each element in the list has type ``dtype``.
 
         Args:
             dtype: DataType of each element in the list
@@ -232,8 +248,7 @@ class DataType:
 
     @classmethod
     def fixed_size_list(cls, dtype: DataType, size: int) -> DataType:
-        """Create a FixedSizeList DataType: Fixed-size list, where each element in the list has type ``dtype``
-        and each list has length ``size``.
+        """Create a FixedSizeList DataType: Fixed-size list, where each element in the list has type ``dtype`` and each list has length ``size``.
 
         Args:
             dtype: DataType of each element in the list
@@ -246,6 +261,7 @@ class DataType:
     @classmethod
     def map(cls, key_type: DataType, value_type: DataType) -> DataType:
         """Create a Map DataType: A map is a nested type of key-value pairs that is implemented as a list of structs with two fields, key and value.
+
         Args:
             key_type: DataType of the keys in the map
             value_type: DataType of the values in the map
@@ -254,7 +270,7 @@ class DataType:
 
     @classmethod
     def struct(cls, fields: dict[str, DataType]) -> DataType:
-        """Create a Struct DataType: a nested type which has names mapped to child types
+        """Create a Struct DataType: a nested type which has names mapped to child types.
 
         Example:
         >>> DataType.struct({"name": DataType.string(), "age": DataType.int64()})
@@ -270,8 +286,7 @@ class DataType:
 
     @classmethod
     def embedding(cls, dtype: DataType, size: int) -> DataType:
-        """Create an Embedding DataType: embeddings are fixed size arrays, where each element
-        in the array has a **numeric** ``dtype`` and each array has a fixed length of ``size``.
+        """Create an Embedding DataType: embeddings are fixed size arrays, where each element in the array has a **numeric** ``dtype`` and each array has a fixed length of ``size``.
 
         Args:
             dtype: DataType of each element in the list (must be numeric)
@@ -328,8 +343,7 @@ class DataType:
         dtype: DataType,
         shape: tuple[int, ...] | None = None,
     ) -> DataType:
-        """Create a tensor DataType: tensor arrays contain n-dimensional arrays of data of the provided ``dtype`` as elements, each of the provided
-        ``shape``.
+        """Create a tensor DataType: tensor arrays contain n-dimensional arrays of data of the provided ``dtype`` as elements, each of the provided ``shape``.
 
         If a ``shape`` is given, each ndarray in the column will have this shape.
 
@@ -346,9 +360,32 @@ class DataType:
                 raise ValueError("Tensor shape must be a non-empty tuple of ints, but got: ", shape)
         return cls._from_pydatatype(PyDataType.tensor(dtype._dtype, shape))
 
+    @classmethod
+    def sparse_tensor(
+        cls,
+        dtype: DataType,
+        shape: tuple[int, ...] | None = None,
+    ) -> DataType:
+        """Create a SparseTensor DataType: SparseTensor arrays implemented as 'COO Sparse Tensor' representation of n-dimensional arrays of data of the provided ``dtype`` as elements, each of the provided ``shape``.
+
+        If a ``shape`` is given, each ndarray in the column will have this shape.
+
+        If ``shape`` is not given, the ndarrays in the column can have different shapes. This is much more flexible,
+        but will result in a less compact representation and may be make some operations less efficient.
+
+        Args:
+            dtype: The type of the data contained within the tensor elements.
+            shape: The shape of each SparseTensor in the column. This is ``None`` by default, which allows the shapes of
+                each tensor element to vary.
+        """
+        if shape is not None:
+            if not isinstance(shape, tuple) or not shape or any(not isinstance(n, int) for n in shape):
+                raise ValueError("SparseTensor shape must be a non-empty tuple of ints, but got: ", shape)
+        return cls._from_pydatatype(PyDataType.sparse_tensor(dtype._dtype, shape))
+
     @classmethod
     def from_arrow_type(cls, arrow_type: pa.lib.DataType) -> DataType:
-        """Maps a PyArrow DataType to a Daft DataType"""
+        """Maps a PyArrow DataType to a Daft DataType."""
         if pa.types.is_int8(arrow_type):
             return cls.int8()
         elif pa.types.is_int16(arrow_type):
@@ -406,16 +443,14 @@ class DataType:
             assert isinstance(arrow_type, pa.StructType)
             fields = [arrow_type[i] for i in range(arrow_type.num_fields)]
             return cls.struct({field.name: cls.from_arrow_type(field.type) for field in fields})
+        elif pa.types.is_interval(arrow_type):
+            return cls.interval()
         elif pa.types.is_map(arrow_type):
             assert isinstance(arrow_type, pa.MapType)
             return cls.map(
                 key_type=cls.from_arrow_type(arrow_type.key_type),
                 value_type=cls.from_arrow_type(arrow_type.item_type),
             )
-        elif _RAY_DATA_EXTENSIONS_AVAILABLE and isinstance(arrow_type, tuple(_TENSOR_EXTENSION_TYPES)):
-            scalar_dtype = cls.from_arrow_type(arrow_type.scalar_type)
-            shape = arrow_type.shape if isinstance(arrow_type, ArrowTensorType) else None
-            return cls.tensor(scalar_dtype, shape)
         elif isinstance(arrow_type, getattr(pa, "FixedShapeTensorType", ())):
             scalar_dtype = cls.from_arrow_type(arrow_type.value_type)
             return cls.tensor(scalar_dtype, tuple(arrow_type.shape))
@@ -428,7 +463,7 @@ class DataType:
         elif isinstance(arrow_type, pa.BaseExtensionType):
             name = arrow_type.extension_name
 
-            if (get_context().runner_config.name == "ray") and (
+            if (get_context().get_or_create_runner().name == "ray") and (
                 type(arrow_type).__reduce__ == pa.BaseExtensionType.__reduce__
             ):
                 raise ValueError(
@@ -458,16 +493,16 @@ class DataType:
 
     @classmethod
     def from_numpy_dtype(cls, np_type: np.dtype) -> DataType:
-        """Maps a Numpy datatype to a Daft DataType"""
+        """Maps a Numpy datatype to a Daft DataType."""
         arrow_type = pa.from_numpy_dtype(np_type)
         return cls.from_arrow_type(arrow_type)
 
-    def to_arrow_dtype(self, cast_tensor_to_ray_type: builtins.bool = False) -> pa.DataType:
-        return self._dtype.to_arrow(cast_tensor_to_ray_type)
+    def to_arrow_dtype(self) -> pa.DataType:
+        return self._dtype.to_arrow()
 
     @classmethod
     def python(cls) -> DataType:
-        """Create a Python DataType: a type which refers to an arbitrary Python object"""
+        """Create a Python DataType: a type which refers to an arbitrary Python object."""
         return cls._from_pydatatype(PyDataType.python())
 
     def _is_python_type(self) -> builtins.bool:
@@ -483,6 +518,12 @@ class DataType:
     def _is_fixed_shape_tensor_type(self) -> builtins.bool:
         return self._dtype.is_fixed_shape_tensor()
 
+    def _is_sparse_tensor_type(self) -> builtins.bool:
+        return self._dtype.is_sparse_tensor()
+
+    def _is_fixed_shape_sparse_tensor_type(self) -> builtins.bool:
+        return self._dtype.is_fixed_shape_sparse_tensor()
+
     def _is_image_type(self) -> builtins.bool:
         return self._dtype.is_image()
 
@@ -492,6 +533,18 @@ class DataType:
     def _is_numeric_type(self) -> builtins.bool:
         return self._dtype.is_numeric()
 
+    def _is_integer(self) -> builtins.bool:
+        return self._dtype.is_integer()
+
+    def _is_list(self) -> builtins.bool:
+        return self._dtype.is_list()
+
+    def _is_boolean(self) -> builtins.bool:
+        return self._dtype.is_boolean()
+
+    def _is_string(self) -> builtins.bool:
+        return self._dtype.is_string()
+
     def _is_map(self) -> builtins.bool:
         return self._dtype.is_map()
 
@@ -520,25 +573,50 @@ class DataType:
         return self._dtype.__hash__()
 
 
-class DaftExtension(pa.ExtensionType):
-    def __init__(self, dtype, metadata=b""):
-        # attributes need to be set first before calling
-        # super init (as that calls serialize)
-        self._metadata = metadata
-        super().__init__(dtype, "daft.super_extension")
-
-    def __reduce__(self):
-        return type(self).__arrow_ext_deserialize__, (self.storage_type, self.__arrow_ext_serialize__())
-
-    def __arrow_ext_serialize__(self):
-        return self._metadata
-
-    @classmethod
-    def __arrow_ext_deserialize__(cls, storage_type, serialized):
-        return cls(storage_type, serialized)
+# Type alias for a union of types that can be inferred into a DataType
+DataTypeLike = Union[DataType, type]
 
 
-pa.register_extension_type(DaftExtension(pa.null()))
-import atexit
+_EXT_TYPE_REGISTRATION_LOCK = threading.Lock()
+_EXT_TYPE_REGISTERED = False
+_STATIC_DAFT_EXTENSION = None
 
-atexit.register(lambda: pa.unregister_extension_type("daft.super_extension"))
+
+def _ensure_registered_super_ext_type():
+    global _EXT_TYPE_REGISTERED
+    global _STATIC_DAFT_EXTENSION
+
+    # Double-checked locking: avoid grabbing the lock if we know that the ext type
+    # has already been registered.
+    if not _EXT_TYPE_REGISTERED:
+        with _EXT_TYPE_REGISTRATION_LOCK:
+            if not _EXT_TYPE_REGISTERED:
+
+                class DaftExtension(pa.ExtensionType):
+                    def __init__(self, dtype, metadata=b""):
+                        # attributes need to be set first before calling
+                        # super init (as that calls serialize)
+                        self._metadata = metadata
+                        super().__init__(dtype, "daft.super_extension")
+
+                    def __reduce__(self):
+                        return type(self).__arrow_ext_deserialize__, (self.storage_type, self.__arrow_ext_serialize__())
+
+                    def __arrow_ext_serialize__(self):
+                        return self._metadata
+
+                    @classmethod
+                    def __arrow_ext_deserialize__(cls, storage_type, serialized):
+                        return cls(storage_type, serialized)
+
+                _STATIC_DAFT_EXTENSION = DaftExtension
+                pa.register_extension_type(DaftExtension(pa.null()))
+                import atexit
+
+                atexit.register(lambda: pa.unregister_extension_type("daft.super_extension"))
+                _EXT_TYPE_REGISTERED = True
+
+
+def get_super_ext_type():
+    _ensure_registered_super_ext_type()
+    return _STATIC_DAFT_EXTENSION
