@@ -1,14 +1,12 @@
 from __future__ import annotations
 
 import itertools
-import pathlib
 from dataclasses import dataclass, field
 from typing import TYPE_CHECKING, Generic, Protocol
 
-from daft.daft import FileFormat, IOConfig, JoinType, ResourceRequest, ScanTask
+from daft.context import get_context
+from daft.daft import JoinSide, ResourceRequest
 from daft.expressions import Expression, ExpressionsProjection, col
-from daft.logical.map_partition_ops import MapPartitionOp
-from daft.logical.schema import Schema
 from daft.runners.partitioning import (
     Boundaries,
     MaterializedResult,
@@ -19,9 +17,15 @@ from daft.runners.partitioning import (
 from daft.table import MicroPartition, table_io
 
 if TYPE_CHECKING:
+    import pathlib
+
     from pyiceberg.schema import Schema as IcebergSchema
     from pyiceberg.table import TableProperties as IcebergTableProperties
 
+    from daft.daft import FileFormat, IOConfig, JoinType, ScanTask
+    from daft.logical.map_partition_ops import MapPartitionOp
+    from daft.logical.schema import Schema
+
 
 ID_GEN = itertools.count()
 
@@ -42,6 +46,17 @@ class PartitionTask(Generic[PartitionT]):
     num_results: int
     stage_id: int
     partial_metadatas: list[PartialPartitionMetadata]
+
+    # Indicates that this PartitionTask must be executed on the executor with the supplied ID
+    # This is used when a specific executor (e.g. an Actor pool) must be provisioned and used for the task
+    actor_pool_id: str | None
+
+    # Indicates if the PartitionTask is "done" or not
+    is_done: bool = False
+
+    # Desired node_id to schedule this task on
+    node_id: str | None = None
+
     _id: int = field(default_factory=lambda: next(ID_GEN))
 
     def id(self) -> str:
@@ -49,14 +64,23 @@ class PartitionTask(Generic[PartitionT]):
 
     def done(self) -> bool:
         """Whether the PartitionT result of this task is available."""
-        raise NotImplementedError()
+        return self.is_done
+
+    def set_done(self):
+        """Sets the PartitionTask as done."""
+        assert not self.is_done, "Cannot set PartitionTask as done more than once"
+        self.is_done = True
 
     def cancel(self) -> None:
         """If possible, cancel the execution of this PartitionTask."""
         raise NotImplementedError()
 
     def set_result(self, result: list[MaterializedResult[PartitionT]]) -> None:
-        """Set the result of this Task. For use by the Task executor."""
+        """Set the result of this Task. For use by the Task executor.
+
+        NOTE: A PartitionTask may contain a `result` without being `.done()`. This is because
+        results can potentially contain futures which are yet to be completed.
+        """
         raise NotImplementedError
 
     def is_empty(self) -> bool:
@@ -86,6 +110,8 @@ class PartitionTaskBuilder(Generic[PartitionT]):
         inputs: list[PartitionT],
         partial_metadatas: list[PartialPartitionMetadata] | None,
         resource_request: ResourceRequest = ResourceRequest(),
+        actor_pool_id: str | None = None,
+        node_id: str | None = None,
     ) -> None:
         self.inputs = inputs
         if partial_metadatas is not None:
@@ -95,6 +121,8 @@ class PartitionTaskBuilder(Generic[PartitionT]):
         self.resource_request: ResourceRequest = resource_request
         self.instructions: list[Instruction] = list()
         self.num_results = len(inputs)
+        self.actor_pool_id = actor_pool_id
+        self.node_id = node_id
 
     def add_instruction(
         self,
@@ -132,6 +160,8 @@ class PartitionTaskBuilder(Generic[PartitionT]):
             num_results=1,
             resource_request=resource_request_final_cpu,
             partial_metadatas=self.partial_metadatas,
+            actor_pool_id=self.actor_pool_id,
+            node_id=self.node_id,
         )
 
     def finalize_partition_task_multi_output(self, stage_id: int) -> MultiOutputPartitionTask[PartitionT]:
@@ -152,6 +182,8 @@ class PartitionTaskBuilder(Generic[PartitionT]):
             num_results=self.num_results,
             resource_request=resource_request_final_cpu,
             partial_metadatas=self.partial_metadatas,
+            actor_pool_id=self.actor_pool_id,
+            node_id=self.node_id,
         )
 
     def __str__(self) -> str:
@@ -175,9 +207,6 @@ class SingleOutputPartitionTask(PartitionTask[PartitionT]):
         [partition] = result
         self._result = partition
 
-    def done(self) -> bool:
-        return self._result is not None
-
     def result(self) -> MaterializedResult[PartitionT]:
         assert self._result is not None, "Cannot call .result() on a PartitionTask that is not done"
         return self._result
@@ -199,9 +228,9 @@ class SingleOutputPartitionTask(PartitionTask[PartitionT]):
         [partial_metadata] = self.partial_metadatas
         return self.result().metadata().merge_with_partial(partial_metadata)
 
-    def vpartition(self) -> MicroPartition:
+    def micropartition(self) -> MicroPartition:
         """Get the raw vPartition of the result."""
-        return self.result().vpartition()
+        return self.result().micropartition()
 
     def __str__(self) -> str:
         return super().__str__()
@@ -212,7 +241,9 @@ class SingleOutputPartitionTask(PartitionTask[PartitionT]):
 
 @dataclass
 class MultiOutputPartitionTask(PartitionTask[PartitionT]):
-    """A PartitionTask that is ready to run. More instructions cannot be added.
+    """A PartitionTask that is ready to run.
+
+    More instructions cannot be added.
     This PartitionTask will return a list of any number of partitions.
     """
 
@@ -223,9 +254,6 @@ class MultiOutputPartitionTask(PartitionTask[PartitionT]):
         assert self._results is None, f"Cannot set result twice. Result is already {self._results}"
         self._results = result
 
-    def done(self) -> bool:
-        return self._results is not None
-
     def cancel(self) -> None:
         if self._results is not None:
             for result in self._results:
@@ -247,10 +275,10 @@ class MultiOutputPartitionTask(PartitionTask[PartitionT]):
             for result, partial_metadata in zip(self._results, self.partial_metadatas)
         ]
 
-    def vpartition(self, index: int) -> MicroPartition:
+    def micropartition(self, index: int) -> MicroPartition:
         """Get the raw vPartition of the result."""
         assert self._results is not None
-        return self._results[index].vpartition()
+        return self._results[index].micropartition()
 
     def __str__(self) -> str:
         return super().__str__()
@@ -304,10 +332,12 @@ class ScanWithTask(SingleOutputInstruction):
     def run_partial_metadata(self, input_metadatas: list[PartialPartitionMetadata]) -> list[PartialPartitionMetadata]:
         assert len(input_metadatas) == 0
 
+        cfg = get_context().daft_execution_config
+
         return [
             PartialPartitionMetadata(
                 num_rows=self.scan_task.num_rows(),
-                size_bytes=self.scan_task.size_bytes(),
+                size_bytes=self.scan_task.estimate_in_memory_size_bytes(cfg),
             )
         ]
 
@@ -375,7 +405,8 @@ class WriteIceberg(SingleOutputInstruction):
     base_path: str
     iceberg_schema: IcebergSchema
     iceberg_properties: IcebergTableProperties
-    spec_id: int
+    partition_spec_id: int
+    partition_cols: ExpressionsProjection
     io_config: IOConfig | None
 
     def run(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
@@ -403,7 +434,8 @@ class WriteIceberg(SingleOutputInstruction):
             base_path=self.base_path,
             schema=self.iceberg_schema,
             properties=self.iceberg_properties,
-            spec_id=self.spec_id,
+            partition_spec_id=self.partition_spec_id,
+            partition_cols=self.partition_cols,
             io_config=self.io_config,
         )
 
@@ -413,6 +445,7 @@ class WriteDeltaLake(SingleOutputInstruction):
     base_path: str
     large_dtypes: bool
     version: int
+    partition_cols: list[str] | None
     io_config: IOConfig | None
 
     def run(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
@@ -440,6 +473,7 @@ class WriteDeltaLake(SingleOutputInstruction):
             large_dtypes=self.large_dtypes,
             base_path=self.base_path,
             version=self.version,
+            partition_cols=self.partition_cols,
             io_config=self.io_config,
         )
 
@@ -527,11 +561,25 @@ class Project(SingleOutputInstruction):
         ]
 
 
+@dataclass(frozen=True)
+class ActorPoolProject(SingleOutputInstruction):
+    projection: ExpressionsProjection
+
+    def run(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
+        raise NotImplementedError("UDFProject instruction cannot be run from outside an Actor. Please file an issue.")
+
+    def run_partial_metadata(self, input_metadatas: list[PartialPartitionMetadata]) -> list[PartialPartitionMetadata]:
+        return [
+            PartialPartitionMetadata(
+                num_rows=None,  # UDFs can potentially change cardinality
+                size_bytes=None,
+                boundaries=None,  # TODO: figure out if the actor pool UDF projection changes boundaries
+            )
+        ]
+
+
 def _prune_boundaries(boundaries: Boundaries, projection: ExpressionsProjection) -> Boundaries | None:
-    """
-    If projection expression is a nontrivial computation (i.e. not a direct col() reference and not an alias) on top of a boundary
-    expression, then invalidate the boundary.
-    """
+    """If projection expression is a nontrivial computation (i.e. not a direct col() reference and not an alias) on top of a boundary expression, then invalidate the boundary."""
     proj_all_names = projection.to_name_set()
     proj_names_needing_compute = proj_all_names - projection.input_mapping().keys()
     for i, e in enumerate(boundaries.sort_by):
@@ -747,6 +795,7 @@ class Unpivot(SingleOutputInstruction):
 class HashJoin(SingleOutputInstruction):
     left_on: ExpressionsProjection
     right_on: ExpressionsProjection
+    null_equals_nulls: list[bool] | None
     how: JoinType
     is_swapped: bool
 
@@ -769,6 +818,7 @@ class HashJoin(SingleOutputInstruction):
             right,
             left_on=self.left_on,
             right_on=self.right_on,
+            null_equals_nulls=self.null_equals_nulls,
             how=self.how,
         )
         return [result]
@@ -876,6 +926,7 @@ class ReduceToQuantiles(ReduceInstruction):
     num_quantiles: int
     sort_by: ExpressionsProjection
     descending: list[bool]
+    nulls_first: list[bool] | None = None
 
     def run(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
         return self._reduce_to_quantiles(inputs)
@@ -883,8 +934,12 @@ class ReduceToQuantiles(ReduceInstruction):
     def _reduce_to_quantiles(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
         merged = MicroPartition.concat(inputs)
 
+        nulls_first = self.nulls_first if self.nulls_first is not None else self.descending
+
         # Skip evaluation of expressions by converting to Column Expression, since evaluation was done in Sample
-        merged_sorted = merged.sort(self.sort_by.to_column_expressions(), descending=self.descending)
+        merged_sorted = merged.sort(
+            self.sort_by.to_column_expressions(), descending=self.descending, nulls_first=nulls_first
+        )
 
         result = merged_sorted.quantiles(self.num_quantiles)
         return [result]
@@ -898,6 +953,52 @@ class ReduceToQuantiles(ReduceInstruction):
         ]
 
 
+def calculate_cross_join_stats(
+    left_meta: PartialPartitionMetadata, right_meta: PartialPartitionMetadata
+) -> tuple[int | None, int | None]:
+    """Given the left and right partition metadata, returns the expected (num rows, size bytes) of the cross join output."""
+    left_rows, left_bytes = left_meta.num_rows, left_meta.size_bytes
+    right_rows, right_bytes = right_meta.num_rows, right_meta.size_bytes
+
+    if left_rows is not None and right_rows is not None:
+        num_rows = left_rows * right_rows
+
+        if left_bytes is not None and right_bytes is not None:
+            size_bytes = left_bytes * right_rows + right_bytes * left_rows
+        else:
+            size_bytes = None
+    else:
+        num_rows = None
+        size_bytes = None
+
+    return num_rows, size_bytes
+
+
+@dataclass(frozen=True)
+class CrossJoin(SingleOutputInstruction):
+    outer_loop_side: JoinSide
+
+    def run(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
+        return self._cross_join(inputs)
+
+    def _cross_join(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
+        left, right = inputs
+        result = left.cross_join(
+            right,
+            self.outer_loop_side,
+        )
+        return [result]
+
+    def run_partial_metadata(self, input_metadatas: list[PartialPartitionMetadata]) -> list[PartialPartitionMetadata]:
+        left_meta, right_meta = input_metadatas
+
+        num_rows, size_bytes = calculate_cross_join_stats(left_meta, right_meta)
+
+        boundaries = left_meta.boundaries if self.outer_loop_side == JoinSide.Left else right_meta.boundaries
+
+        return [PartialPartitionMetadata(num_rows=num_rows, size_bytes=size_bytes, boundaries=boundaries)]
+
+
 @dataclass(frozen=True)
 class FanoutInstruction(Instruction):
     _num_outputs: int
@@ -1009,3 +1110,33 @@ class FanoutSlices(FanoutInstruction):
             )
 
         return results
+
+
+@dataclass(frozen=True)
+class FanoutEvenSlices(FanoutInstruction):
+    def run(self, inputs: list[MicroPartition]) -> list[MicroPartition]:
+        [input] = inputs
+        results = []
+
+        input_length = len(input)
+        num_outputs = self.num_outputs()
+
+        chunk_size, remainder = divmod(input_length, num_outputs)
+        ptr = 0
+        for output_idx in range(self.num_outputs()):
+            end = ptr + chunk_size + 1 if output_idx < remainder else ptr + chunk_size
+            results.append(input.slice(ptr, end))
+            ptr = end
+        assert ptr == input_length
+
+        return results
+
+    def run_partial_metadata(self, input_metadatas: list[PartialPartitionMetadata]) -> list[PartialPartitionMetadata]:
+        # TODO: Derive this based on the ratios of num rows
+        return [
+            PartialPartitionMetadata(
+                num_rows=None,
+                size_bytes=None,
+            )
+            for _ in range(self._num_outputs)
+        ]
