@@ -1,5 +1,5 @@
-"""
-This file contains physical plan building blocks.
+"""This file contains physical plan building blocks.
+
 To get a physical plan for a logical plan, see physical_plan_factory.py.
 
 Conceptually, a physical plan decides what steps, and the order of steps, to run to build some target.
@@ -14,10 +14,11 @@ because it is waiting for the result of a previous PartitionTask to can decide w
 from __future__ import annotations
 
 import collections
+import contextlib
 import itertools
 import logging
 import math
-import pathlib
+from abc import abstractmethod
 from collections import deque
 from typing import (
     TYPE_CHECKING,
@@ -30,7 +31,7 @@ from typing import (
 )
 
 from daft.context import get_context
-from daft.daft import FileFormat, IOConfig, JoinType, ResourceRequest
+from daft.daft import JoinSide, ResourceRequest
 from daft.execution import execution_step
 from daft.execution.execution_step import (
     Instruction,
@@ -39,9 +40,9 @@ from daft.execution.execution_step import (
     PartitionTaskBuilder,
     ReduceInstruction,
     SingleOutputPartitionTask,
+    calculate_cross_join_stats,
 )
 from daft.expressions import ExpressionsProjection
-from daft.logical.schema import Schema
 from daft.runners.partitioning import (
     MaterializedResult,
     PartitionT,
@@ -53,9 +54,14 @@ logger = logging.getLogger(__name__)
 T = TypeVar("T")
 
 if TYPE_CHECKING:
+    import pathlib
+
     from pyiceberg.schema import Schema as IcebergSchema
     from pyiceberg.table import TableProperties as IcebergTableProperties
 
+    from daft.daft import FileFormat, IOConfig, JoinType
+    from daft.logical.schema import Schema
+
 
 # A PhysicalPlan that is still being built - may yield both PartitionTaskBuilders and PartitionTasks.
 InProgressPhysicalPlan = Iterator[Union[None, PartitionTask[PartitionT], PartitionTaskBuilder[PartitionT]]]
@@ -94,7 +100,6 @@ def file_write(
     io_config: IOConfig | None,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Write the results of `child_plan` into files described by `write_info`."""
-
     yield from (
         step.add_instruction(
             execution_step.WriteFile(
@@ -117,18 +122,19 @@ def iceberg_write(
     base_path: str,
     iceberg_schema: IcebergSchema,
     iceberg_properties: IcebergTableProperties,
-    spec_id: int,
+    partition_spec_id: int,
+    partition_cols: ExpressionsProjection,
     io_config: IOConfig | None,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Write the results of `child_plan` into pyiceberg data files described by `write_info`."""
-
     yield from (
         step.add_instruction(
             execution_step.WriteIceberg(
                 base_path=base_path,
                 iceberg_schema=iceberg_schema,
                 iceberg_properties=iceberg_properties,
-                spec_id=spec_id,
+                partition_spec_id=partition_spec_id,
+                partition_cols=partition_cols,
                 io_config=io_config,
             ),
         )
@@ -143,16 +149,17 @@ def deltalake_write(
     base_path: str,
     large_dtypes: bool,
     version: int,
+    partition_cols: list[str] | None,
     io_config: IOConfig | None,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Write the results of `child_plan` into pyiceberg data files described by `write_info`."""
-
     yield from (
         step.add_instruction(
             execution_step.WriteDeltaLake(
                 base_path=base_path,
                 large_dtypes=large_dtypes,
                 version=version,
+                partition_cols=partition_cols,
                 io_config=io_config,
             ),
         )
@@ -170,7 +177,6 @@ def lance_write(
     kwargs: dict | None,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Write the results of `child_plan` into lance data files described by `write_info`."""
-
     yield from (
         step.add_instruction(
             execution_step.WriteLance(
@@ -192,18 +198,130 @@ def pipeline_instruction(
     resource_request: execution_step.ResourceRequest,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Apply an instruction to the results of `child_plan`."""
-
     yield from (
         step.add_instruction(pipeable_instruction, resource_request) if isinstance(step, PartitionTaskBuilder) else step
         for step in child_plan
     )
 
 
+class ActorPoolManager:
+    @abstractmethod
+    @contextlib.contextmanager
+    def actor_pool_context(
+        self,
+        name: str,
+        actor_resource_request: ResourceRequest,
+        task_resource_request: ResourceRequest,
+        num_actors: int,
+        projection: ExpressionsProjection,
+    ) -> Iterator[str]:
+        """Creates a pool of actors which can execute work, and yield a context in which the pool can be used.
+
+        Also yields a `str` ID which clients can use to refer to the actor pool when submitting tasks.
+
+        Note that attempting to do work outside this context will result in errors!
+
+        Args:
+            name: Name of the actor pool for debugging/observability
+            resource_request: Requested amount of resources for each actor
+            num_actors: Number of actors to spin up
+            projection: Projection to be run on the incoming data (contains actor pool UDFs as well as other stateless expressions such as aliases)
+        """
+        ...
+
+
+def actor_pool_project(
+    child_plan: InProgressPhysicalPlan[PartitionT],
+    projection: ExpressionsProjection,
+    actor_pool_manager: ActorPoolManager,
+    resource_request: execution_step.ResourceRequest,
+    num_actors: int,
+) -> InProgressPhysicalPlan[PartitionT]:
+    stage_id = next(stage_id_counter)
+
+    from daft.daft import get_udf_names
+
+    udf_names = "-".join(name for expr in projection for name in get_udf_names(expr._expr))
+    actor_pool_name = f"{udf_names}-stage={stage_id}"
+
+    # Keep track of materializations of the children tasks
+    child_materializations: deque[SingleOutputPartitionTask[PartitionT]] = deque()
+
+    # Keep track of materializations of the actor_pool tasks
+    actor_pool_materializations: deque[SingleOutputPartitionTask[PartitionT]] = deque()
+
+    # Perform separate accounting for the tasks' resource request and the actors' resource request:
+    # * When spinning up an actor, we consider resources that are required for the persistent state in an actor (namely, GPUs and memory)
+    # * When running a task, we consider resources that are required for placement of tasks (namely CPUs)
+    task_resource_request = ResourceRequest(num_cpus=resource_request.num_cpus)
+    actor_resource_request = ResourceRequest(
+        num_gpus=resource_request.num_gpus, memory_bytes=resource_request.memory_bytes
+    )
+
+    with actor_pool_manager.actor_pool_context(
+        actor_pool_name,
+        actor_resource_request,
+        task_resource_request,
+        num_actors,
+        projection,
+    ) as actor_pool_id:
+        child_plan_exhausted = False
+
+        # Loop until the child plan is exhausted and there is no more work in the pipeline
+        while not (child_plan_exhausted and len(child_materializations) == 0 and len(actor_pool_materializations) == 0):
+            # Exhaustively pop ready child_steps and submit them to be run on the actor_pool
+            while len(child_materializations) > 0 and child_materializations[0].done():
+                next_ready_child = child_materializations.popleft()
+                actor_project_step = (
+                    PartitionTaskBuilder[PartitionT](
+                        inputs=[next_ready_child.partition()],
+                        partial_metadatas=[next_ready_child.partition_metadata()],
+                        actor_pool_id=actor_pool_id,
+                    )
+                    .add_instruction(
+                        instruction=execution_step.ActorPoolProject(projection),
+                        resource_request=task_resource_request,
+                    )
+                    .finalize_partition_task_single_output(
+                        stage_id=stage_id,
+                    )
+                )
+                actor_pool_materializations.append(actor_project_step)
+                yield actor_project_step
+
+            # Exhaustively pop ready actor_pool steps and bubble it upwards as the start of a new pipeline
+            while len(actor_pool_materializations) > 0 and actor_pool_materializations[0].done():
+                next_ready_actor_pool_task = actor_pool_materializations.popleft()
+                new_pipeline_starter_task = PartitionTaskBuilder[PartitionT](
+                    inputs=[next_ready_actor_pool_task.partition()],
+                    partial_metadatas=[next_ready_actor_pool_task.partition_metadata()],
+                    resource_request=ResourceRequest(),
+                )
+                yield new_pipeline_starter_task
+
+            # No more child work to be done: if there is pending work in the pipeline we yield None
+            if child_plan_exhausted:
+                if len(child_materializations) > 0 or len(actor_pool_materializations) > 0:
+                    yield None
+
+            # Attempt to schedule child work
+            else:
+                try:
+                    child_step = next(child_plan)
+                except StopIteration:
+                    child_plan_exhausted = True
+                else:
+                    # Finalize and yield the child step to be run if it is a PartitionTaskBuilder
+                    if isinstance(child_step, PartitionTaskBuilder):
+                        child_step = child_step.finalize_partition_task_single_output(stage_id=stage_id)
+                        child_materializations.append(child_step)
+                    yield child_step
+
+
 def monotonically_increasing_id(
     child_plan: InProgressPhysicalPlan[PartitionT], column_name: str
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Apply a monotonically_increasing_id instruction to the results of `child_plan`."""
-
     partition_counter = (
         0  # This counter gives each partition a monotonically increasing int to use as the leftmost 28 bits of the id
     )
@@ -222,10 +340,10 @@ def hash_join(
     right_plan: InProgressPhysicalPlan[PartitionT],
     left_on: ExpressionsProjection,
     right_on: ExpressionsProjection,
+    null_equals_nulls: None | list[bool],
     how: JoinType,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Hash-based pairwise join the partitions from `left_child_plan` and `right_child_plan` together."""
-
     # Materialize the steps from the left and right sources to get partitions.
     # As the materializations complete, emit new steps to join each left and right partition.
     left_requests: deque[SingleOutputPartitionTask[PartitionT]] = deque()
@@ -263,6 +381,7 @@ def hash_join(
                 instruction=execution_step.HashJoin(
                     left_on=left_on,
                     right_on=right_on,
+                    null_equals_nulls=null_equals_nulls,
                     how=how,
                     is_swapped=False,
                 )
@@ -308,6 +427,7 @@ def _create_broadcast_join_step(
     receiver_part: SingleOutputPartitionTask[PartitionT],
     left_on: ExpressionsProjection,
     right_on: ExpressionsProjection,
+    null_equals_nulls: None | list[bool],
     how: JoinType,
     is_swapped: bool,
 ) -> PartitionTaskBuilder[PartitionT]:
@@ -353,6 +473,7 @@ def _create_broadcast_join_step(
         instruction=execution_step.BroadcastJoin(
             left_on=left_on,
             right_on=right_on,
+            null_equals_nulls=null_equals_nulls,
             how=how,
             is_swapped=is_swapped,
         )
@@ -364,11 +485,11 @@ def broadcast_join(
     receiver_plan: InProgressPhysicalPlan[PartitionT],
     left_on: ExpressionsProjection,
     right_on: ExpressionsProjection,
+    null_equals_nulls: None | list[bool],
     how: JoinType,
     is_swapped: bool,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Broadcast join all partitions from the broadcaster child plan to each partition in the receiver child plan."""
-
     # Materialize the steps from the broadcaster and receiver sources to get partitions.
     # As the receiver-side materializations complete, emit new steps to join each broadcaster and receiver partition.
     stage_id = next(stage_id_counter)
@@ -406,7 +527,15 @@ def broadcast_join(
         # Broadcast all broadcaster partitions to each new receiver partition that was materialized on this dispatch loop.
         while receiver_requests and receiver_requests[0].done():
             receiver_part = receiver_requests.popleft()
-            yield _create_broadcast_join_step(broadcaster_parts, receiver_part, left_on, right_on, how, is_swapped)
+            yield _create_broadcast_join_step(
+                broadcaster_parts,
+                receiver_part,
+                left_on,
+                right_on,
+                null_equals_nulls,
+                how,
+                is_swapped,
+            )
 
         # Execute single child step to pull in more input partitions.
         try:
@@ -426,9 +555,80 @@ def broadcast_join(
                 return
 
 
+def cross_join(
+    left_plan: InProgressPhysicalPlan[PartitionT],
+    right_plan: InProgressPhysicalPlan[PartitionT],
+    outer_loop_side: JoinSide,
+):
+    stage_id = next(stage_id_counter)
+
+    outer_plan, inner_plan = (left_plan, right_plan) if outer_loop_side == JoinSide.Left else (right_plan, left_plan)
+
+    # Materialize inner side first
+    inner_requests: deque[SingleOutputPartitionTask[PartitionT]] = deque()
+    for step in inner_plan:
+        if isinstance(step, PartitionTaskBuilder):
+            step = step.finalize_partition_task_single_output(stage_id=stage_id)
+            inner_requests.append(step)
+        yield step
+
+    outer_requests: deque[SingleOutputPartitionTask[PartitionT]] = deque()
+    while True:
+        while outer_requests and outer_requests[0].done():
+            next_outer = outer_requests.popleft()
+
+            for next_inner in inner_requests:
+                while not next_inner.done():
+                    logger.debug(
+                        "cross join blocked on completion of inner side of join.\n inner sources: %s",
+                        inner_requests,
+                    )
+                    yield None
+
+                next_left, next_right = (
+                    (next_outer, next_inner) if outer_loop_side == JoinSide.Left else (next_inner, next_outer)
+                )
+
+                # Calculate memory request for task.
+                left_meta = next_left.partition_metadata()
+                right_meta = next_right.partition_metadata()
+
+                size_bytes = None
+
+                # If left or right side metadata missing, assume that left and right side are ~ the same size.
+                for first, second in [(left_meta, right_meta), (left_meta, left_meta), (right_meta, right_meta)]:
+                    _, size_bytes = calculate_cross_join_stats(first, second)
+                    if size_bytes is not None:
+                        break
+
+                join_step = PartitionTaskBuilder[PartitionT](
+                    inputs=[next_left.partition(), next_right.partition()],
+                    partial_metadatas=[next_left.partition_metadata(), next_right.partition_metadata()],
+                    resource_request=ResourceRequest(memory_bytes=size_bytes),
+                ).add_instruction(instruction=execution_step.CrossJoin(outer_loop_side=outer_loop_side))
+
+                yield join_step
+
+        # Execute single child step to pull in more outer partitions.
+        try:
+            step = next(outer_plan)
+            if isinstance(step, PartitionTaskBuilder):
+                step = step.finalize_partition_task_single_output(stage_id=stage_id)
+                outer_requests.append(step)
+            yield step
+        except StopIteration:
+            if outer_requests:
+                logger.debug(
+                    "broadcast join blocked on completion of receiver side of join.\n receiver sources: %s",
+                    outer_requests,
+                )
+                yield None
+            else:
+                return
+
+
 class MergeJoinTaskTracker(Generic[PartitionT]):
-    """
-    Tracks merge-join tasks for each larger-side partition.
+    """Tracks merge-join tasks for each larger-side partition.
 
     Merge-join tasks are added to the tracker, and the tracker handles empty tasks, finalizing PartitionTaskBuilders,
     determining whether tasks are ready to be executed, checking whether tasks are done, and deciding whether a coalesce
@@ -454,8 +654,7 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
         self._stage_id = stage_id
 
     def add_task(self, part_id: str, task: PartitionTaskBuilder[PartitionT]) -> None:
-        """
-        Add a merge-join task to the tracker for the provided larger-side partition.
+        """Add a merge-join task to the tracker for the provided larger-side partition.
 
         This task needs to be unfinalized, i.e. a PartitionTaskBuilder.
         """
@@ -478,9 +677,7 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
             self._finalized_tasks[part_id].append(task.finalize_partition_task_single_output(self._stage_id))
 
     def finalize(self, part_id: str) -> None:
-        """
-        Indicates to the tracker that we are done adding merge-join tasks for this partition.
-        """
+        """Indicates to the tracker that we are done adding merge-join tasks for this partition."""
         # All finalized tasks should have been yielded before the tracker.finalize() call.
         finalized_tasks = self._finalized_tasks.pop(part_id, deque())
         assert len(finalized_tasks) == 0
@@ -490,8 +687,9 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
     def yield_ready(
         self, part_id: str
     ) -> Iterator[SingleOutputPartitionTask[PartitionT] | PartitionTaskBuilder[PartitionT]]:
-        """
-        Returns an iterator of all tasks for this partition that are ready for execution. Each merge-join task will be
+        """Returns an iterator of all tasks for this partition that are ready for execution.
+
+        Each merge-join task will be
         yielded once, even across multiple calls.
         """
         assert self._is_contained(part_id)
@@ -507,8 +705,9 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
             yield self._task_staging.pop(part_id)
 
     def pop_uncoalesced(self, part_id: str) -> deque[SingleOutputPartitionTask[PartitionT]] | None:
-        """
-        Returns all tasks for this partition that need to be coalesced. If this partition only involved a single
+        """Returns all tasks for this partition that need to be coalesced.
+
+        If this partition only involved a single
         merge-join task (i.e. we don't need to coalesce), this this function will return None.
 
         NOTE: tracker.finalize(part_id) must be called before this function.
@@ -517,9 +716,7 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
         return self._uncoalesced_tasks.pop(part_id, None)
 
     def all_tasks_done_for_partition(self, part_id: str) -> bool:
-        """
-        Return whether all merge-join tasks for this partition are done.
-        """
+        """Return whether all merge-join tasks for this partition are done."""
         assert self._is_contained(part_id)
         if part_id in self._task_staging:
             # Unfinalized tasks are trivially "done".
@@ -532,9 +729,7 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
         )
 
     def all_tasks_done(self) -> bool:
-        """
-        Return whether all merge-join tasks for all partitions are done.
-        """
+        """Return whether all merge-join tasks for all partitions are done."""
         return all(
             self.all_tasks_done_for_partition(part_id)
             for part_id in itertools.chain(
@@ -543,9 +738,7 @@ class MergeJoinTaskTracker(Generic[PartitionT]):
         )
 
     def _is_contained(self, part_id: str) -> bool:
-        """
-        Return whether the provided partition is being tracked by this tracker.
-        """
+        """Return whether the provided partition is being tracked by this tracker."""
         return part_id in self._task_staging or part_id in self._finalized_tasks or part_id in self._uncoalesced_tasks
 
 
@@ -559,9 +752,7 @@ def _emit_merge_joins_on_window(
     right_on: ExpressionsProjection,
     how: JoinType,
 ) -> Iterator[PartitionTaskBuilder[PartitionT] | PartitionTask[PartitionT]]:
-    """
-    Emits merge-join steps of next_part with each partition in other_window.
-    """
+    """Emits merge-join steps of next_part with each partition in other_window."""
     # Emit a merge-join step for all partitions in the other window that intersect with this new partition.
     for other_next_part in other_window:
         memory_bytes = _memory_bytes_for_merge(next_part, other_next_part)
@@ -620,13 +811,11 @@ def merge_join_sorted(
     how: JoinType,
     left_is_larger: bool,
 ) -> InProgressPhysicalPlan[PartitionT]:
-    """
-    Merge the sorted partitions from `left_plan` and `right_plan` together.
+    """Merge the sorted partitions from `left_plan` and `right_plan` together.
 
     This assumes that `left_plan` and `right_plan` are both sorted on the join key(s), although with potentially
     different range partitionings (partition boundaries).
     """
-
     # Large vs. smaller side of join.
     larger_plan = left_plan if left_is_larger else right_plan
     smaller_plan = right_plan if left_is_larger else left_plan
@@ -816,10 +1005,7 @@ def merge_join_sorted(
 def _is_strictly_bounded_above_by(
     lower_part: SingleOutputPartitionTask[PartitionT], upper_part: SingleOutputPartitionTask[PartitionT]
 ) -> bool:
-    """
-    Returns whether lower_part is strictly bounded above by upper part; i.e., whether lower_part's upper bound is
-    strictly less than upper_part's upper bound.
-    """
+    """Returns whether lower_part is strictly bounded above by upper part."""
     lower_boundaries = lower_part.partition_metadata().boundaries
     upper_boundaries = upper_part.partition_metadata().boundaries
     assert lower_boundaries is not None and upper_boundaries is not None
@@ -854,8 +1040,7 @@ def sort_merge_join_aligned_boundaries(
     num_partitions: int,
     left_is_larger: bool,
 ) -> InProgressPhysicalPlan[PartitionT]:
-    """
-    Sort-merge join the partitions from `left_plan` and `right_plan` together.
+    """Sort-merge join the partitions from `left_plan` and `right_plan` together.
 
     This assumes that both `left_plan` and `right_plan` need to be sorted, and will be sorted using the same
     partitioning boundaries.
@@ -1012,7 +1197,7 @@ def sort_merge_join_aligned_boundaries(
         ]
 
         # Execute a sorting reduce on it.
-        per_partition_bounds = _to_per_partition_bounds(boundaries.vpartition(), num_partitions)
+        per_partition_bounds = _to_per_partition_bounds(boundaries.micropartition(), num_partitions)
         sorted_plans.append(
             reduce(
                 fanout_plan=iter(range_fanout_plan),
@@ -1052,7 +1237,7 @@ def _to_per_partition_bounds(boundaries: MicroPartition, num_partitions: int) ->
 def concat(
     top_plan: InProgressPhysicalPlan[PartitionT], bottom_plan: InProgressPhysicalPlan[PartitionT]
 ) -> InProgressPhysicalPlan[PartitionT]:
-    """Vertical concat of the partitions in `top_plan` and `bottom_plan`"""
+    """Vertical concat of the partitions in `top_plan` and `bottom_plan`."""
     # Yield steps in order from the top_plan to bottom_plan
     yield from top_plan
     yield from bottom_plan
@@ -1088,7 +1273,6 @@ def global_limit(
     num_partitions: int,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Return the first n rows from the `child_plan`."""
-
     remaining_rows = limit_rows
     assert remaining_rows >= 0, f"Invalid value for limit: {remaining_rows}"
     remaining_partitions = num_partitions
@@ -1186,7 +1370,6 @@ def global_limit(
 
 def flatten_plan(child_plan: InProgressPhysicalPlan[PartitionT]) -> InProgressPhysicalPlan[PartitionT]:
     """Wrap a plan that emits multi-output tasks to a plan that emits single-output tasks."""
-
     materializations: deque[MultiOutputPartitionTask[PartitionT]] = deque()
     stage_id = next(stage_id_counter)
     while True:
@@ -1219,61 +1402,30 @@ def split(
     num_input_partitions: int,
     num_output_partitions: int,
 ) -> InProgressPhysicalPlan[PartitionT]:
-    """Repartition the child_plan into more partitions by splitting partitions only. Preserves order."""
+    """Repartition the child_plan into more partitions by splitting partitions only. Preserves order.
 
+    This performs a naive split, which might lead to data skews but does not require a full materialization of
+    input partitions when performing the split.
+    """
     assert (
         num_output_partitions >= num_input_partitions
     ), f"Cannot split from {num_input_partitions} to {num_output_partitions}."
 
-    # Materialize the input partitions so we can see the number of rows and try to split evenly.
-    # Splitting evenly is fairly important if this operation is to be used for parallelism.
-    # (optimization TODO: don't materialize if num_rows is already available in physical plan metadata.)
-    materializations: deque[SingleOutputPartitionTask[PartitionT]] = deque()
-    stage_id = next(stage_id_counter)
+    base_splits_per_partition, num_partitions_with_extra_output = divmod(num_output_partitions, num_input_partitions)
+
+    input_partition_idx = 0
     for step in child_plan:
         if isinstance(step, PartitionTaskBuilder):
-            step = step.finalize_partition_task_single_output(stage_id=stage_id)
-            materializations.append(step)
-        yield step
-
-    while any(not _.done() for _ in materializations):
-        logger.debug("split_to blocked on completion of all sources: %s", materializations)
-        yield None
-
-    splits_per_partition = deque([1 for _ in materializations])
-    num_splits_to_apply = num_output_partitions - num_input_partitions
-
-    # Split by rows for now.
-    # In the future, maybe parameterize to allow alternatively splitting by size.
-    rows_by_partitions = [task.partition_metadata().num_rows for task in materializations]
-
-    # Calculate how to spread the required splits across all the partitions.
-    # Iteratively apply a split and update how many rows would be in the resulting partitions.
-    # After this loop, splits_per_partition has the final number of splits to apply to each partition.
-    rows_after_splitting = [float(_) for _ in rows_by_partitions]
-    for _ in range(num_splits_to_apply):
-        _, split_at = max((rows, index) for (index, rows) in enumerate(rows_after_splitting))
-        splits_per_partition[split_at] += 1
-        rows_after_splitting[split_at] = float(rows_by_partitions[split_at] / splits_per_partition[split_at])
-
-    # Emit the split partitions.
-    for task, num_out, num_rows in zip(consume_deque(materializations), splits_per_partition, rows_by_partitions):
-        if num_out == 1:
-            yield PartitionTaskBuilder[PartitionT](
-                inputs=[task.partition()],
-                partial_metadatas=[task.partition_metadata()],
-                resource_request=ResourceRequest(memory_bytes=task.partition_metadata().size_bytes),
+            num_out = (
+                base_splits_per_partition + 1
+                if input_partition_idx < num_partitions_with_extra_output
+                else base_splits_per_partition
             )
+            step = step.add_instruction(instruction=execution_step.FanoutEvenSlices(_num_outputs=num_out))
+            input_partition_idx += 1
+            yield step
         else:
-            boundaries = [math.ceil(num_rows * i / num_out) for i in range(num_out + 1)]
-            starts, ends = boundaries[:-1], boundaries[1:]
-            yield PartitionTaskBuilder[PartitionT](
-                inputs=[task.partition()],
-                partial_metadatas=[task.partition_metadata()],
-                resource_request=ResourceRequest(memory_bytes=task.partition_metadata().size_bytes),
-            ).add_instruction(
-                instruction=execution_step.FanoutSlices(_num_outputs=num_out, slices=list(zip(starts, ends)))
-            )
+            yield step
 
 
 def coalesce(
@@ -1285,7 +1437,6 @@ def coalesce(
 
     The current implementation only does partition merging, no rebalancing.
     """
-
     assert (
         to_num_partitions <= from_num_partitions
     ), f"Cannot coalesce upwards from {from_num_partitions} to {to_num_partitions} partitions."
@@ -1361,7 +1512,6 @@ def reduce(
 
     Then, the reduce instruction is applied to each `i`th slice across the child lists.
     """
-
     materializations = list()
     stage_id = next(stage_id_counter)
 
@@ -1403,10 +1553,10 @@ def sort(
     child_plan: InProgressPhysicalPlan[PartitionT],
     sort_by: ExpressionsProjection,
     descending: list[bool],
+    nulls_first: list[bool],
     num_partitions: int,
 ) -> InProgressPhysicalPlan[PartitionT]:
     """Sort the result of `child_plan` according to `sort_info`."""
-
     # First, materialize the child plan.
     source_materializations: deque[SingleOutputPartitionTask[PartitionT]] = deque()
     stage_id_children = next(stage_id_counter)
@@ -1458,6 +1608,7 @@ def sort(
                 num_quantiles=num_partitions,
                 sort_by=sort_by,
                 descending=descending,
+                nulls_first=nulls_first,
             ),
         )
         .finalize_partition_task_single_output(stage_id=stage_id_reduce)
@@ -1486,7 +1637,7 @@ def sort(
         )
         for source in consume_deque(source_materializations)
     )
-    per_partition_bounds = _to_per_partition_bounds(boundaries.vpartition(), num_partitions)
+    per_partition_bounds = _to_per_partition_bounds(boundaries.micropartition(), num_partitions)
 
     # Execute a sorting reduce on it.
     yield from reduce(
@@ -1503,7 +1654,7 @@ def sort(
 
 
 def fanout_random(child_plan: InProgressPhysicalPlan[PartitionT], num_partitions: int):
-    """Splits the results of `child_plan` randomly into a list of `node.num_partitions()` number of partitions"""
+    """Splits the results of `child_plan` randomly into a list of `node.num_partitions()` number of partitions."""
     seed = 0
     for step in child_plan:
         if isinstance(step, PartitionTaskBuilder):
@@ -1513,39 +1664,136 @@ def fanout_random(child_plan: InProgressPhysicalPlan[PartitionT], num_partitions
         seed += 1
 
 
-def materialize(
-    child_plan: InProgressPhysicalPlan[PartitionT],
-) -> MaterializedPhysicalPlan:
+def _best_effort_next_step(
+    stage_id: int, child_plan: InProgressPhysicalPlan[PartitionT]
+) -> tuple[PartitionTask[PartitionT] | None, bool]:
+    """Performs a best-effort attempt at retrieving the next step from a child plan.
+
+    Returns None in cases where there is nothing to run, or the plan has been exhausted.
+
+    Returns:
+        step: the step (potentially None) to run
+        is_final_task: a boolean indicating whether or not this step was a final step
+    """
+    try:
+        step = next(child_plan)
+    except StopIteration:
+        return (None, False)
+    else:
+        if isinstance(step, PartitionTaskBuilder):
+            step = step.finalize_partition_task_single_output(stage_id=stage_id)
+            return (step, True)
+        elif isinstance(step, PartitionTask):
+            return (step, False)
+        else:
+            return (None, False)
+
+
+class Materialize:
     """Materialize the child plan.
 
     Repeatedly yields either a PartitionTask (to produce an intermediate partition)
     or a PartitionT (which is part of the final result).
     """
 
-    materializations: deque[SingleOutputPartitionTask[PartitionT]] = deque()
-    stage_id = next(stage_id_counter)
-    while True:
-        # Check if any inputs finished executing.
-        while len(materializations) > 0 and materializations[0].done():
-            done_task = materializations.popleft()
-            yield done_task.result()
+    def __init__(
+        self,
+        child_plan: InProgressPhysicalPlan[PartitionT],
+        results_buffer_size: int | None,
+    ):
+        self.child_plan = child_plan
+        self.materializations: deque[SingleOutputPartitionTask[PartitionT]] = deque()
+        self.results_buffer_size = results_buffer_size
 
-        # Materialize a single dependency.
-        try:
-            step = next(child_plan)
-            if isinstance(step, PartitionTaskBuilder):
-                step = step.finalize_partition_task_single_output(stage_id=stage_id)
-                materializations.append(step)
-            assert isinstance(step, (PartitionTask, type(None)))
+    def __iter__(self) -> MaterializedPhysicalPlan:
+        num_materialized_yielded = 0
+        num_intermediate_yielded = 0
+        num_final_yielded = 0
+        stage_id = next(stage_id_counter)
 
-            yield step
+        logger.debug(
+            "[plan-%s] Starting to emit tasks from `materialize` with results_buffer_size=%s",
+            stage_id,
+            self.results_buffer_size,
+        )
 
-        except StopIteration:
-            if len(materializations) > 0:
-                logger.debug("materialize blocked on completion of all sources: %s", materializations)
+        while True:
+            # If any inputs have finished executing, we want to drain the `materializations` buffer
+            while len(self.materializations) > 0 and self.materializations[0].done():
+                # Make space on buffer by popping the task that was done
+                done_task = self.materializations.popleft()
+
+                # Best-effort attempt to yield new work and fill up the buffer to the desired `results_buffer_size`
+                if self.results_buffer_size is not None:
+                    for _ in range(self.results_buffer_size - len(self.materializations)):
+                        best_effort_step, is_final_task = _best_effort_next_step(stage_id, self.child_plan)
+                        if best_effort_step is None:
+                            break
+                        elif is_final_task:
+                            assert isinstance(best_effort_step, SingleOutputPartitionTask)
+                            self.materializations.append(best_effort_step)
+                            num_final_yielded += 1
+                            logger.debug(
+                                "[plan-%s] YIELDING final task to replace done materialized task (%s so far)",
+                                stage_id,
+                                num_final_yielded,
+                            )
+                        else:
+                            num_intermediate_yielded += 1
+                            logger.debug(
+                                "[plan-%s] YIELDING an intermediate task to replace done materialized task (%s so far)",
+                                stage_id,
+                                num_intermediate_yielded,
+                            )
+                        yield best_effort_step
+
+                # Yield the task that was done
+                num_materialized_yielded += 1
+                logger.debug("[plan-%s] YIELDING a materialized task (%s so far)", stage_id, num_materialized_yielded)
+                yield done_task.result()
+
+            # If the buffer has too many results already, we yield None until some are completed
+            if self.results_buffer_size is not None and len(self.materializations) >= self.results_buffer_size:
+                logger.debug(
+                    "[plan-%s] YIELDING none, waiting on tasks in buffer to complete: %s in buffer, but maximum is %s",
+                    stage_id,
+                    len(self.materializations),
+                    self.results_buffer_size,
+                )
                 yield None
-            else:
-                return
+
+                # Important: start again at the top and drain materialized results
+                # Otherwise it may lead to a weird corner-case where the plan has ended (raising StopIteration)
+                # but some of the completed materializations haven't been drained from the buffer.
+                continue
+
+            # Materialize a single dependency.
+            try:
+                step = next(self.child_plan)
+                if isinstance(step, PartitionTaskBuilder):
+                    step = step.finalize_partition_task_single_output(stage_id=stage_id)
+                    self.materializations.append(step)
+                    num_final_yielded += 1
+                    logger.debug("[plan-%s] YIELDING final task (%s so far)", stage_id, num_final_yielded)
+                elif isinstance(step, PartitionTask):
+                    num_intermediate_yielded += 1
+                    logger.debug(
+                        "[plan-%s] YIELDING an intermediate task (%s so far)", stage_id, num_intermediate_yielded
+                    )
+
+                assert isinstance(step, (PartitionTask, type(None)))
+                yield step
+
+            except StopIteration:
+                if len(self.materializations) > 0:
+                    logger.debug(
+                        "[plan-%s] YIELDING none, iterator completed but materialize is blocked on completion of all sources: %s",
+                        stage_id,
+                        self.materializations,
+                    )
+                    yield None
+                else:
+                    return
 
 
 def enumerate_open_executions(
