@@ -17,12 +17,13 @@ from daft.expressions import Expression, ExpressionsProjection
 from daft.logical.map_partition_ops import MapPartitionOp
 from daft.logical.schema import Schema
 from daft.runners.partitioning import PartitionT
-from daft.table import MicroPartition
 
 if TYPE_CHECKING:
     from pyiceberg.schema import Schema as IcebergSchema
     from pyiceberg.table import TableProperties as IcebergTableProperties
 
+    from daft.table import MicroPartition
+
 
 def scan_with_tasks(
     scan_tasks: list[ScanTask],
@@ -51,7 +52,7 @@ def scan_with_tasks(
 def empty_scan(
     schema: Schema,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
-    """yield a plan to create an empty Partition"""
+    """Yield a plan to create an empty Partition."""
     scan_step = execution_step.PartitionTaskBuilder[PartitionT](
         inputs=[],
         partial_metadatas=None,
@@ -63,9 +64,14 @@ def empty_scan(
 
 
 def project(
-    input: physical_plan.InProgressPhysicalPlan[PartitionT], projection: list[PyExpr], resource_request: ResourceRequest
+    input: physical_plan.InProgressPhysicalPlan[PartitionT],
+    projection: list[PyExpr],
+    resource_request: ResourceRequest | None,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
     expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in projection])
+    resource_request = (
+        ResourceRequest(num_cpus=1, num_gpus=None, memory_bytes=None) if resource_request is None else resource_request
+    )
     return physical_plan.pipeline_instruction(
         child_plan=input,
         pipeable_instruction=execution_step.Project(expr_projection),
@@ -73,6 +79,26 @@ def project(
     )
 
 
+def actor_pool_project(
+    input: physical_plan.InProgressPhysicalPlan[PartitionT],
+    projection: list[PyExpr],
+    actor_pool_manager: physical_plan.ActorPoolManager,
+    resource_request: ResourceRequest | None,
+    num_actors: int,
+) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
+    expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in projection])
+    resource_request = (
+        ResourceRequest(num_cpus=1, num_gpus=None, memory_bytes=None) if resource_request is None else resource_request
+    )
+    return physical_plan.actor_pool_project(
+        child_plan=input,
+        projection=expr_projection,
+        actor_pool_manager=actor_pool_manager,
+        resource_request=resource_request,
+        num_actors=num_actors,
+    )
+
+
 class ShimExplodeOp(MapPartitionOp):
     explode_columns: ExpressionsProjection
 
@@ -176,6 +202,7 @@ def sort(
     input: physical_plan.InProgressPhysicalPlan[PartitionT],
     sort_by: list[PyExpr],
     descending: list[bool],
+    nulls_first: list[bool],
     num_partitions: int,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
     expr_projection = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in sort_by])
@@ -183,11 +210,12 @@ def sort(
         child_plan=input,
         sort_by=expr_projection,
         descending=descending,
+        nulls_first=nulls_first,
         num_partitions=num_partitions,
     )
 
 
-def split_by_hash(
+def fanout_by_hash(
     input: physical_plan.InProgressPhysicalPlan[PartitionT],
     num_partitions: int,
     partition_by: list[PyExpr],
@@ -216,6 +244,7 @@ def hash_join(
     right: physical_plan.InProgressPhysicalPlan[PartitionT],
     left_on: list[PyExpr],
     right_on: list[PyExpr],
+    null_equals_nulls: list[bool] | None,
     join_type: JoinType,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
     left_on_expr_proj = ExpressionsProjection([Expression._from_pyexpr(expr) for expr in left_on])
@@ -226,6 +255,7 @@ def hash_join(
         left_on=left_on_expr_proj,
         right_on=right_on_expr_proj,
         how=join_type,
+        null_equals_nulls=null_equals_nulls,
     )
 
 
@@ -276,6 +306,7 @@ def broadcast_join(
     receiver: physical_plan.InProgressPhysicalPlan[PartitionT],
     left_on: list[PyExpr],
     right_on: list[PyExpr],
+    null_equals_nulls: list[bool] | None,
     join_type: JoinType,
     is_swapped: bool,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
@@ -288,6 +319,7 @@ def broadcast_join(
         right_on=right_on_expr_proj,
         how=join_type,
         is_swapped=is_swapped,
+        null_equals_nulls=null_equals_nulls,
     )
 
 
@@ -320,7 +352,8 @@ def write_iceberg(
     base_path: str,
     iceberg_schema: IcebergSchema,
     iceberg_properties: IcebergTableProperties,
-    spec_id: int,
+    partition_spec_id: int,
+    partition_cols: list[PyExpr],
     io_config: IOConfig | None,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
     return physical_plan.iceberg_write(
@@ -328,7 +361,8 @@ def write_iceberg(
         base_path=base_path,
         iceberg_schema=iceberg_schema,
         iceberg_properties=iceberg_properties,
-        spec_id=spec_id,
+        partition_spec_id=partition_spec_id,
+        partition_cols=ExpressionsProjection([Expression._from_pyexpr(expr) for expr in partition_cols]),
         io_config=io_config,
     )
 
@@ -338,6 +372,7 @@ def write_deltalake(
     path: str,
     large_dtypes: bool,
     version: int,
+    partition_cols: list[str] | None,
     io_config: IOConfig | None,
 ) -> physical_plan.InProgressPhysicalPlan[PartitionT]:
     return physical_plan.deltalake_write(
@@ -345,6 +380,7 @@ def write_deltalake(
         path,
         large_dtypes,
         version,
+        partition_cols,
         io_config,
     )
 
