@@ -9,9 +9,11 @@ if TYPE_CHECKING:
 
 
 def expr_structurally_equal(e1: Expression, e2: Expression) -> bool:
-    """Returns a boolean indicating whether two Expressions are structurally equal:
+    """Returns a boolean indicating whether two Expressions are structurally equal.
+
+    Structurally equal is defined as:
     1. Expressions' local parameters are value-wise equal
     2. Expressions have the same number of children Expressions
-    3. (Recursive) Expressions' children are structurally equal to each other as well
+    3. (Recursive) Expressions' children are structurally equal to each other as well.
     """
     return _eq(e1._expr, e2._expr)
