@@ -6,24 +6,24 @@ import os
 import pathlib
 import sys
 import urllib.parse
-from typing import Any, Literal
-
-import fsspec
-from fsspec.registry import get_filesystem_class
-from pyarrow.fs import FileSystem, LocalFileSystem, S3FileSystem
-from pyarrow.fs import _resolve_filesystem_and_path as pafs_resolve_filesystem_and_path
+from typing import TYPE_CHECKING, Any, Literal
 
+from daft.convert import from_pydict
 from daft.daft import FileFormat, FileInfos, IOConfig, io_glob
+from daft.dependencies import fsspec, pafs
+from daft.expressions.expressions import col, lit
 from daft.table import MicroPartition
 
+if TYPE_CHECKING:
+    from daft import DataFrame
+
 logger = logging.getLogger(__name__)
 
-_CACHED_FSES: dict[tuple[str, IOConfig | None], FileSystem] = {}
+_CACHED_FSES: dict[tuple[str, IOConfig | None], pafs.FileSystem] = {}
 
 
-def _get_fs_from_cache(protocol: str, io_config: IOConfig | None) -> FileSystem | None:
-    """
-    Get an instantiated pyarrow filesystem from the cache based on the URI protocol.
+def _get_fs_from_cache(protocol: str, io_config: IOConfig | None) -> pafs.FileSystem | None:
+    """Get an instantiated pyarrow filesystem from the cache based on the URI protocol.
 
     Returns None if no such cache entry exists.
     """
@@ -32,7 +32,7 @@ def _get_fs_from_cache(protocol: str, io_config: IOConfig | None) -> FileSystem
     return _CACHED_FSES.get((protocol, io_config))
 
 
-def _put_fs_in_cache(protocol: str, fs: FileSystem, io_config: IOConfig | None) -> None:
+def _put_fs_in_cache(protocol: str, fs: pafs.FileSystem, io_config: IOConfig | None) -> None:
     """Put pyarrow filesystem in cache under provided protocol."""
     global _CACHED_FSES
 
@@ -43,7 +43,7 @@ def _put_fs_in_cache(protocol: str, fs: FileSystem, io_config: IOConfig | None)
 class ListingInfo:
     path: str
     size: int
-    type: Literal["file"] | Literal["directory"]
+    type: Literal["file", "directory"]
     rows: int | None = None
 
 
@@ -105,20 +105,17 @@ _CANONICAL_PROTOCOLS = {
 
 
 def canonicalize_protocol(protocol: str) -> str:
-    """
-    Return the canonical protocol from the provided protocol, such that there's a 1:1
-    mapping between protocols and pyarrow/fsspec filesystem implementations.
-    """
+    """Return the canonical protocol from the provided protocol, such that there's a 1:1 mapping between protocols and pyarrow/fsspec filesystem implementations."""
     return _CANONICAL_PROTOCOLS.get(protocol, protocol)
 
 
 def _resolve_paths_and_filesystem(
     paths: str | pathlib.Path | list[str],
     io_config: IOConfig | None = None,
-) -> tuple[list[str], FileSystem]:
-    """
-    Resolves and normalizes all provided paths, infers a filesystem from the
-    paths, and ensures that all paths use the same filesystem.
+) -> tuple[list[str], pafs.FileSystem]:
+    """Resolves and normalizes the provided path and infers it's filesystem.
+
+    Also ensures that the inferred filesystem is compatible with the passed filesystem, if provided.
 
     Args:
         paths: A single file/directory path or a list of file/directory paths.
@@ -166,7 +163,7 @@ def _resolve_paths_and_filesystem(
 
     # filesystem should be a non-None pyarrow FileSystem at this point, either
     # user-provided, taken from the cache, or inferred from the first path.
-    assert resolved_filesystem is not None and isinstance(resolved_filesystem, FileSystem)
+    assert resolved_filesystem is not None and isinstance(resolved_filesystem, pafs.FileSystem)
 
     # Resolve all other paths and validate with the user-provided/cached/inferred filesystem.
     resolved_paths = [resolved_path]
@@ -177,7 +174,7 @@ def _resolve_paths_and_filesystem(
     return resolved_paths, resolved_filesystem
 
 
-def _validate_filesystem(path: str, fs: FileSystem, io_config: IOConfig | None) -> str:
+def _validate_filesystem(path: str, fs: pafs.FileSystem, io_config: IOConfig | None) -> str:
     resolved_path, inferred_fs = _infer_filesystem(path, io_config)
     if not isinstance(fs, type(inferred_fs)):
         raise RuntimeError(
@@ -189,11 +186,10 @@ def _validate_filesystem(path: str, fs: FileSystem, io_config: IOConfig | None)
 def _infer_filesystem(
     path: str,
     io_config: IOConfig | None,
-) -> tuple[str, FileSystem]:
-    """
-    Resolves and normalizes the provided path, infers a filesystem from the
-    path, and ensures that the inferred filesystem is compatible with the passed
-    filesystem, if provided.
+) -> tuple[str, pafs.FileSystem]:
+    """Resolves and normalizes the provided path and infers it's filesystem.
+
+    Also ensures that the inferred filesystem is compatible with the passedfilesystem, if provided.
 
     Args:
         path: A single file/directory path.
@@ -204,7 +200,7 @@ def _infer_filesystem(
     translated_kwargs: dict[str, Any]
 
     def _set_if_not_none(kwargs: dict[str, Any], key: str, val: Any | None):
-        """Helper method used when setting kwargs for pyarrow"""
+        """Helper method used when setting kwargs for pyarrow."""
         if val is not None:
             kwargs[key] = val
 
@@ -221,8 +217,15 @@ def _infer_filesystem(
             _set_if_not_none(translated_kwargs, "session_token", s3_config.session_token)
             _set_if_not_none(translated_kwargs, "region", s3_config.region_name)
             _set_if_not_none(translated_kwargs, "anonymous", s3_config.anonymous)
+            if s3_config.num_tries is not None:
+                try:
+                    from pyarrow.fs import AwsStandardS3RetryStrategy
 
-        resolved_filesystem = S3FileSystem(**translated_kwargs)
+                    translated_kwargs["retry_strategy"] = AwsStandardS3RetryStrategy(max_attempts=s3_config.num_tries)
+                except ImportError:
+                    pass  # Config does not exist in pyarrow 7.0.0
+
+        resolved_filesystem = pafs.S3FileSystem(**translated_kwargs)
         resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
         return resolved_path, resolved_filesystem
 
@@ -230,7 +233,7 @@ def _infer_filesystem(
     # Local
     ###
     elif protocol == "file":
-        resolved_filesystem = LocalFileSystem()
+        resolved_filesystem = pafs.LocalFileSystem()
         resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(path))
         return resolved_path, resolved_filesystem
 
@@ -260,9 +263,9 @@ def _infer_filesystem(
     # HTTP: Use FSSpec as a fallback
     ###
     elif protocol in {"http", "https"}:
-        fsspec_fs_cls = get_filesystem_class(protocol)
+        fsspec_fs_cls = fsspec.get_filesystem_class(protocol)
         fsspec_fs = fsspec_fs_cls()
-        resolved_filesystem, resolved_path = pafs_resolve_filesystem_and_path(path, fsspec_fs)
+        resolved_filesystem, resolved_path = pafs._resolve_filesystem_and_path(path, fsspec_fs)
         resolved_path = resolved_filesystem.normalize_path(resolved_path)
         return resolved_path, resolved_filesystem
 
@@ -270,7 +273,7 @@ def _infer_filesystem(
     # Azure: Use FSSpec as a fallback
     ###
     elif protocol in {"az", "abfs", "abfss"}:
-        fsspec_fs_cls = get_filesystem_class(protocol)
+        fsspec_fs_cls = fsspec.get_filesystem_class(protocol)
 
         if io_config is not None:
             # TODO: look into support for other AzureConfig parameters
@@ -285,7 +288,7 @@ def _infer_filesystem(
             )
         else:
             fsspec_fs = fsspec_fs_cls()
-        resolved_filesystem, resolved_path = pafs_resolve_filesystem_and_path(path, fsspec_fs)
+        resolved_filesystem, resolved_path = pafs._resolve_filesystem_and_path(path, fsspec_fs)
         resolved_path = resolved_filesystem.normalize_path(_unwrap_protocol(resolved_path))
         return resolved_path, resolved_filesystem
 
@@ -294,9 +297,7 @@ def _infer_filesystem(
 
 
 def _unwrap_protocol(path):
-    """
-    Slice off any protocol prefixes on path.
-    """
+    """Slice off any protocol prefixes on path."""
     parsed = urllib.parse.urlparse(path, allow_fragments=False)  # support '#' in path
     query = "?" + parsed.query if parsed.query else ""  # support '?' in path
     return parsed.netloc + parsed.path + query
@@ -341,12 +342,33 @@ def glob_path_with_stats(
 ###
 
 
-def join_path(fs: FileSystem, base_path: str, *sub_paths: str) -> str:
-    """
-    Join a base path with sub-paths using the appropriate path separator
-    for the given filesystem.
-    """
-    if isinstance(fs, LocalFileSystem):
+def join_path(fs: pafs.FileSystem, base_path: str, *sub_paths: str) -> str:
+    """Join a base path with sub-paths using the appropriate path separator for the given filesystem."""
+    if isinstance(fs, pafs.LocalFileSystem):
         return os.path.join(base_path, *sub_paths)
     else:
         return f"{base_path.rstrip('/')}/{'/'.join(sub_paths)}"
+
+
+def overwrite_files(
+    manifest: DataFrame,
+    root_dir: str | pathlib.Path,
+    io_config: IOConfig | None,
+) -> None:
+    [resolved_path], fs = _resolve_paths_and_filesystem(root_dir, io_config=io_config)
+    file_selector = pafs.FileSelector(resolved_path, recursive=True)
+    try:
+        paths = [info.path for info in fs.get_file_info(file_selector) if info.type == pafs.FileType.File]
+    except FileNotFoundError:
+        # The root directory does not exist, so there are no files to delete.
+        return
+
+    all_file_paths_df = from_pydict({"path": paths})
+
+    assert manifest._result is not None
+    written_file_paths = manifest._result._get_merged_micropartition().get_column("path")
+    to_delete = all_file_paths_df.where(~(col("path").is_in(lit(written_file_paths))))
+
+    # TODO: Look into parallelizing this
+    for entry in to_delete:
+        fs.delete_file(entry["path"])
