@@ -1,7 +1,7 @@
 from __future__ import annotations
 
 import logging
-from collections.abc import Iterator
+from typing import TYPE_CHECKING
 
 import daft
 from daft.daft import (
@@ -16,13 +16,16 @@ from daft.hudi.pyhudi.table import HUDI_METAFIELD_PARTITION_PATH, HudiTable, Hud
 from daft.io.scan import PartitionField, ScanOperator
 from daft.logical.schema import Schema
 
+if TYPE_CHECKING:
+    from collections.abc import Iterator
+
 logger = logging.getLogger(__name__)
 
 
 class HudiScanOperator(ScanOperator):
     def __init__(self, table_uri: str, storage_config: StorageConfig) -> None:
         super().__init__()
-        resolved_path, self._resolved_fs = _resolve_paths_and_filesystem(table_uri, storage_config.config.io_config)
+        resolved_path, self._resolved_fs = _resolve_paths_and_filesystem(table_uri, storage_config.io_config)
         self._table = HudiTable(table_uri, self._resolved_fs, resolved_path[0])
         self._storage_config = storage_config
         self._schema = Schema.from_pyarrow_schema(self._table.schema)
@@ -34,6 +37,9 @@ class HudiScanOperator(ScanOperator):
     def schema(self) -> Schema:
         return self._schema
 
+    def name(self) -> str:
+        return "HudiScanOperator"
+
     def display_name(self) -> str:
         return f"HudiScanOperator({self._table.props.name})"
 
@@ -55,7 +61,7 @@ class HudiScanOperator(ScanOperator):
         files_metadata = hudi_table_metadata.files_metadata
 
         if len(self.partitioning_keys()) > 0 and pushdowns.partition_filters is None:
-            logging.warning(
+            logger.warning(
                 "%s has partitioning keys = %s, but no partition filter was specified. This will result in a full table scan.",
                 self.display_name(),
                 self.partitioning_keys(),
@@ -118,6 +124,7 @@ class HudiScanOperator(ScanOperator):
                 storage_config=self._storage_config,
                 num_rows=record_count,
                 size_bytes=size_bytes,
+                iceberg_delete_files=None,
                 pushdowns=pushdowns,
                 partition_values=partition_values,
                 stats=stats,
