@@ -4,7 +4,7 @@ from typing import Optional
 
 from daft import context
 from daft.api_annotations import PublicAPI
-from daft.daft import IOConfig, NativeStorageConfig, ScanOperatorHandle, StorageConfig
+from daft.daft import IOConfig, ScanOperatorHandle, StorageConfig
 from daft.dataframe import DataFrame
 from daft.logical.builder import LogicalPlanBuilder
 
@@ -32,8 +32,8 @@ def read_hudi(
 
     io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
 
-    multithreaded_io = not context.get_context().is_ray_runner
-    storage_config = StorageConfig.native(NativeStorageConfig(multithreaded_io, io_config))
+    multithreaded_io = context.get_context().get_or_create_runner().name != "ray"
+    storage_config = StorageConfig(multithreaded_io, io_config)
 
     hudi_operator = HudiScanOperator(table_uri, storage_config=storage_config)
 
