@@ -4,7 +4,7 @@ from typing import TYPE_CHECKING, Any, Dict, Optional
 
 from daft import context
 from daft.api_annotations import PublicAPI
-from daft.daft import IOConfig, NativeStorageConfig, ScanOperatorHandle, StorageConfig
+from daft.daft import IOConfig, ScanOperatorHandle, StorageConfig
 from daft.dataframe import DataFrame
 from daft.logical.builder import LogicalPlanBuilder
 
@@ -88,7 +88,7 @@ def read_iceberg(
     snapshot_id: Optional[int] = None,
     io_config: Optional["IOConfig"] = None,
 ) -> DataFrame:
-    """Create a DataFrame from an Iceberg table
+    """Create a DataFrame from an Iceberg table.
 
     Example:
         >>> import pyiceberg
@@ -122,8 +122,8 @@ def read_iceberg(
     )
     io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
 
-    multithreaded_io = not context.get_context().is_ray_runner
-    storage_config = StorageConfig.native(NativeStorageConfig(multithreaded_io, io_config))
+    multithreaded_io = context.get_context().get_or_create_runner().name != "ray"
+    storage_config = StorageConfig(multithreaded_io, io_config)
 
     iceberg_operator = IcebergScanOperator(pyiceberg_table, snapshot_id=snapshot_id, storage_config=storage_config)
 
