@@ -1,8 +1,9 @@
 from __future__ import annotations
 
-import pathlib
-from typing import TYPE_CHECKING
+import functools
+from typing import TYPE_CHECKING, Callable
 
+from daft.context import get_context
 from daft.daft import (
     CountMode,
     FileFormat,
@@ -10,37 +11,59 @@ from daft.daft import (
     JoinStrategy,
     JoinType,
     PyDaftExecutionConfig,
-    ResourceRequest,
     ScanOperatorHandle,
+    logical_plan_table_scan,
 )
 from daft.daft import LogicalPlanBuilder as _LogicalPlanBuilder
 from daft.expressions import Expression, col
 from daft.logical.schema import Schema
-from daft.runners.partitioning import PartitionCacheEntry
 
 if TYPE_CHECKING:
+    import pathlib
+
     from pyiceberg.table import Table as IcebergTable
 
     from daft.plan_scheduler.physical_plan_scheduler import (
         AdaptivePhysicalPlanScheduler,
         PhysicalPlanScheduler,
     )
+    from daft.runners.partitioning import PartitionCacheEntry
+
+
+def _apply_daft_planning_config_to_initializer(classmethod_func: Callable[..., LogicalPlanBuilder]):
+    """Decorator to be applied to any @classmethod instantiation method on LogicalPlanBuilder.
+
+    This decorator ensures that the current DaftPlanningConfig is applied to the instantiated LogicalPlanBuilder
+    """
+
+    @functools.wraps(classmethod_func)
+    def wrapper(cls: type[LogicalPlanBuilder], *args, **kwargs):
+        instantiated_logical_plan_builder = classmethod_func(cls, *args, **kwargs)
+
+        # Parametrize the builder with the current DaftPlanningConfig
+        inner = instantiated_logical_plan_builder._builder
+        inner = inner.with_planning_config(get_context().daft_planning_config)
+
+        return cls(inner)
+
+    return wrapper
 
 
 class LogicalPlanBuilder:
-    """
-    A logical plan builder for the Daft DataFrame.
-    """
+    """A logical plan builder for the Daft DataFrame."""
 
     def __init__(self, builder: _LogicalPlanBuilder) -> None:
         self._builder = builder
 
     def to_physical_plan_scheduler(self, daft_execution_config: PyDaftExecutionConfig) -> PhysicalPlanScheduler:
-        """
-        Convert the underlying logical plan to a physical plan scheduler, which is
-        used to generate executable tasks for the physical plan.
+        """Convert the underlying logical plan to a physical plan scheduler.
+
+        physical plan scheduler is used to generate executable tasks for the physical plan.
 
         This should be called after triggering optimization with self.optimize().
+
+        **Warning**: This function is not part of the stable API and may change
+        without notice. It is intended for internal or experimental use only.
         """
         from daft.plan_scheduler.physical_plan_scheduler import PhysicalPlanScheduler
 
@@ -62,32 +85,31 @@ class LogicalPlanBuilder:
         )
 
     def schema(self) -> Schema:
-        """
-        The schema of the current logical plan.
-        """
+        """The schema of the current logical plan."""
         pyschema = self._builder.schema()
         return Schema._from_pyschema(pyschema)
 
-    def pretty_print(self, simple: bool = False) -> str:
-        """
-        Pretty prints the current underlying logical plan.
-        """
-        if simple:
-            return self._builder.repr_ascii(simple=True)
+    def pretty_print(self, simple: bool = False, format: str = "ascii") -> str:
+        """Pretty prints the current underlying logical plan."""
+        from daft.dataframe.display import MermaidOptions
+
+        if format == "ascii":
+            return self._builder.repr_ascii(simple)
+        elif format == "mermaid":
+            return self._builder.repr_mermaid(MermaidOptions(simple))
         else:
-            return repr(self)
+            raise ValueError(f"Unknown format: {format}")
 
     def __repr__(self) -> str:
         return self._builder.repr_ascii(simple=False)
 
     def optimize(self) -> LogicalPlanBuilder:
-        """
-        Optimize the underlying logical plan.
-        """
+        """Optimize the underlying logical plan."""
         builder = self._builder.optimize()
         return LogicalPlanBuilder(builder)
 
     @classmethod
+    @_apply_daft_planning_config_to_initializer
     def from_in_memory_scan(
         cls,
         partition: PartitionCacheEntry,
@@ -107,12 +129,13 @@ class LogicalPlanBuilder:
         return cls(builder)
 
     @classmethod
+    @_apply_daft_planning_config_to_initializer
     def from_tabular_scan(
         cls,
         *,
         scan_operator: ScanOperatorHandle,
     ) -> LogicalPlanBuilder:
-        builder = _LogicalPlanBuilder.table_scan(scan_operator)
+        builder = logical_plan_table_scan(scan_operator)
         return cls(builder)
 
     def select(
@@ -123,9 +146,9 @@ class LogicalPlanBuilder:
         builder = self._builder.select(to_select_pyexprs)
         return LogicalPlanBuilder(builder)
 
-    def with_columns(self, columns: list[Expression], custom_resource_request: ResourceRequest) -> LogicalPlanBuilder:
+    def with_columns(self, columns: list[Expression]) -> LogicalPlanBuilder:
         column_pyexprs = [expr._expr for expr in columns]
-        builder = self._builder.with_columns(column_pyexprs, custom_resource_request)
+        builder = self._builder.with_columns(column_pyexprs)
         return LogicalPlanBuilder(builder)
 
     def exclude(self, to_exclude: list[str]) -> LogicalPlanBuilder:
@@ -172,11 +195,20 @@ class LogicalPlanBuilder:
         builder = self._builder.sample(fraction, with_replacement, seed)
         return LogicalPlanBuilder(builder)
 
-    def sort(self, sort_by: list[Expression], descending: list[bool] | bool = False) -> LogicalPlanBuilder:
+    def sort(
+        self,
+        sort_by: list[Expression],
+        descending: list[bool] | bool = False,
+        nulls_first: list[bool] | bool | None = None,
+    ) -> LogicalPlanBuilder:
         sort_by_pyexprs = [expr._expr for expr in sort_by]
         if not isinstance(descending, list):
             descending = [descending] * len(sort_by_pyexprs)
-        builder = self._builder.sort(sort_by_pyexprs, descending)
+        if nulls_first is None:
+            nulls_first = descending
+        elif isinstance(nulls_first, bool):
+            nulls_first = [nulls_first] * len(sort_by_pyexprs)
+        builder = self._builder.sort(sort_by_pyexprs, descending, nulls_first)
         return LogicalPlanBuilder(builder)
 
     def hash_repartition(self, num_partitions: int | None, partition_by: list[Expression]) -> LogicalPlanBuilder:
@@ -225,6 +257,8 @@ class LogicalPlanBuilder:
         right_on: list[Expression],
         how: JoinType = JoinType.Inner,
         strategy: JoinStrategy | None = None,
+        join_suffix: str | None = None,
+        join_prefix: str | None = None,
     ) -> LogicalPlanBuilder:
         builder = self._builder.join(
             right._builder,
@@ -232,6 +266,8 @@ class LogicalPlanBuilder:
             [expr._expr for expr in right_on],
             how,
             strategy,
+            join_suffix,
+            join_prefix,
         )
         return LogicalPlanBuilder(builder)
 
@@ -239,6 +275,10 @@ class LogicalPlanBuilder:
         builder = self._builder.concat(other._builder)
         return LogicalPlanBuilder(builder)
 
+    def intersect(self, other: LogicalPlanBuilder) -> LogicalPlanBuilder:
+        builder = self._builder.intersect(other._builder, False)
+        return LogicalPlanBuilder(builder)
+
     def add_monotonically_increasing_id(self, column_name: str | None) -> LogicalPlanBuilder:
         builder = self._builder.add_monotonically_increasing_id(column_name)
         return LogicalPlanBuilder(builder)
@@ -258,16 +298,26 @@ class LogicalPlanBuilder:
         return LogicalPlanBuilder(builder)
 
     def write_iceberg(self, table: IcebergTable) -> LogicalPlanBuilder:
+        from daft.iceberg.iceberg_write import get_missing_columns, partition_field_to_expr
         from daft.io._iceberg import _convert_iceberg_file_io_properties_to_io_config
 
         name = ".".join(table.name())
         location = f"{table.location()}/data"
-        spec_id = table.spec().spec_id
+        partition_spec = table.spec()
         schema = table.schema()
+        missing_columns = get_missing_columns(self.schema().to_pyarrow_schema(), schema)
+        builder = (
+            self._builder
+            if len(missing_columns) == 0
+            else self._builder.with_columns([c._expr for c in missing_columns])
+        )
+        partition_cols = [partition_field_to_expr(field, schema)._expr for field in partition_spec.fields]
         props = table.properties
         columns = [col.name for col in schema.columns]
         io_config = _convert_iceberg_file_io_properties_to_io_config(table.io.properties)
-        builder = self._builder.iceberg_write(name, location, spec_id, schema, props, columns, io_config)
+        builder = builder.iceberg_write(
+            name, location, partition_spec.spec_id, partition_cols, schema, props, columns, io_config
+        )
         return LogicalPlanBuilder(builder)
 
     def write_deltalake(
@@ -277,6 +327,7 @@ class LogicalPlanBuilder:
         version: int,
         large_dtypes: bool,
         io_config: IOConfig,
+        partition_cols: list[str] | None = None,
     ) -> LogicalPlanBuilder:
         columns_name = self.schema().column_names()
         builder = self._builder.delta_write(
@@ -285,6 +336,7 @@ class LogicalPlanBuilder:
             mode,
             version,
             large_dtypes,
+            partition_cols,
             io_config,
         )
         return LogicalPlanBuilder(builder)
