@@ -8,7 +8,7 @@ from daft.daft import PySchema as _PySchema
 from daft.daft import read_csv_schema as _read_csv_schema
 from daft.daft import read_json_schema as _read_json_schema
 from daft.daft import read_parquet_schema as _read_parquet_schema
-from daft.datatype import DataType, TimeUnit
+from daft.datatype import DataType, TimeUnit, _ensure_registered_super_ext_type
 
 if TYPE_CHECKING:
     import pyarrow as pa
@@ -64,7 +64,7 @@ class Schema:
 
     @classmethod
     def from_pyarrow_schema(cls, pa_schema: pa.Schema) -> Schema:
-        """Creates a Daft Schema from a PyArrow Schema
+        """Creates a Daft Schema from a PyArrow Schema.
 
         Args:
             pa_schema (pa.Schema): PyArrow schema to convert
@@ -77,11 +77,12 @@ class Schema:
         )
 
     def to_pyarrow_schema(self) -> pa.Schema:
-        """Converts a Daft Schema to a PyArrow Schema
+        """Converts a Daft Schema to a PyArrow Schema.
 
         Returns:
             pa.Schema: PyArrow schema that corresponds to the provided Daft schema
         """
+        _ensure_registered_super_ext_type()
         return self._schema.to_pyarrow_schema()
 
     @classmethod
@@ -142,14 +143,14 @@ class Schema:
     def _truncated_table_string(self) -> str:
         return self._schema._truncated_table_string()
 
+    def apply_hints(self, hints: Schema) -> Schema:
+        return Schema._from_pyschema(self._schema.apply_hints(hints._schema))
+
+    # Takes the unions between two schemas. Throws an error if the schemas contain overlapping keys.
     def union(self, other: Schema) -> Schema:
         if not isinstance(other, Schema):
             raise ValueError(f"Expected Schema, got other: {type(other)}")
 
-        intersecting_names = self.to_name_set().intersection(other.to_name_set())
-        if intersecting_names:
-            raise ValueError(f"Cannot union schemas with overlapping names: {intersecting_names}")
-
         return Schema._from_pyschema(self._schema.union(other._schema))
 
     def __reduce__(self) -> tuple:
