@@ -1,6 +1,6 @@
 from __future__ import annotations
 
-from typing import Iterator
+from typing import TYPE_CHECKING
 
 from daft.daft import (
     AdaptivePhysicalPlanScheduler as _AdaptivePhysicalPlanScheduler,
@@ -10,20 +10,17 @@ from daft.daft import (
     PyDaftExecutionConfig,
 )
 from daft.execution import physical_plan
-from daft.logical.builder import LogicalPlanBuilder
-from daft.runners.partitioning import (
-    MaterializedResult,
-    PartitionCacheEntry,
-    PartitionT,
-)
-from daft.runners.pyrunner import PyMaterializedResult
-from daft.table.micropartition import MicroPartition
+
+if TYPE_CHECKING:
+    from daft.logical.builder import LogicalPlanBuilder
+    from daft.runners.partitioning import (
+        PartitionCacheEntry,
+        PartitionT,
+    )
 
 
 class PhysicalPlanScheduler:
-    """
-    Generates executable tasks for an underlying physical plan.
-    """
+    """Generates executable tasks for an underlying physical plan."""
 
     def __init__(self, scheduler: _PhysicalPlanScheduler):
         self._scheduler = scheduler
@@ -38,25 +35,33 @@ class PhysicalPlanScheduler:
     def num_partitions(self) -> int:
         return self._scheduler.num_partitions()
 
-    def pretty_print(self, simple: bool = False) -> str:
-        """
-        Pretty prints the current underlying physical plan.
-        """
-        if simple:
-            return self._scheduler.repr_ascii(simple=True)
+    def pretty_print(self, simple: bool = False, format: str = "ascii") -> str:
+        """Pretty prints the current underlying physical plan."""
+        from daft.dataframe.display import MermaidOptions
+
+        if format == "ascii":
+            return self._scheduler.repr_ascii(simple)
+        elif format == "mermaid":
+            return self._scheduler.repr_mermaid(MermaidOptions(simple))
         else:
-            return repr(self)
+            raise ValueError(f"Unknown format: {format}")
 
     def __repr__(self) -> str:
         return self._scheduler.repr_ascii(simple=False)
 
-    def to_partition_tasks(self, psets: dict[str, list[PartitionT]]) -> physical_plan.MaterializedPhysicalPlan:
-        return physical_plan.materialize(self._scheduler.to_partition_tasks(psets))
+    def to_json_string(self) -> str:
+        return self._scheduler.to_json_string()
 
-    def run(self, psets: dict[str, list[MaterializedResult[PartitionT]]]) -> Iterator[PyMaterializedResult]:
-        psets_mp = {part_id: [part.vpartition()._micropartition for part in parts] for part_id, parts in psets.items()}
-        return (
-            PyMaterializedResult(MicroPartition._from_pymicropartition(part)) for part in self._scheduler.run(psets_mp)
+    def to_partition_tasks(
+        self,
+        psets: dict[str, list[PartitionT]],
+        actor_pool_manager: physical_plan.ActorPoolManager,
+        results_buffer_size: int | None,
+    ) -> physical_plan.MaterializedPhysicalPlan:
+        return iter(
+            physical_plan.Materialize(
+                self._scheduler.to_partition_tasks(psets, actor_pool_manager), results_buffer_size
+            )
         )
 
 
