@@ -1,43 +1,17 @@
 from __future__ import annotations
 
+import warnings
 from typing import Any, Literal, TypeVar
 
-import pyarrow as pa
-
 from daft.arrow_utils import ensure_array, ensure_chunked_array
-from daft.daft import CountMode, ImageFormat, PySeries
-from daft.datatype import DataType
+from daft.daft import CountMode, ImageFormat, ImageMode, PySeries, image
+from daft.datatype import DataType, _ensure_registered_super_ext_type
+from daft.dependencies import np, pa, pd
 from daft.utils import pyarrow_supports_fixed_shape_tensor
 
-_RAY_DATA_EXTENSIONS_AVAILABLE = True
-try:
-    from ray.data.extensions import (
-        ArrowTensorArray,
-        ArrowTensorType,
-        ArrowVariableShapedTensorType,
-    )
-except ImportError:
-    _RAY_DATA_EXTENSIONS_AVAILABLE = False
-
-_NUMPY_AVAILABLE = True
-try:
-    import numpy as np
-except ImportError:
-    _NUMPY_AVAILABLE = False
-
-_PANDAS_AVAILABLE = True
-try:
-    import pandas as pd
-except ImportError:
-    _PANDAS_AVAILABLE = False
-
-ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
-
 
 class Series:
-    """
-    A Daft Series is an array of data of a single type, and is usually a column in a DataFrame.
-    """
+    """A Daft Series is an array of data of a single type, and is usually a column in a DataFrame."""
 
     _series: PySeries
 
@@ -52,30 +26,19 @@ class Series:
 
     @staticmethod
     def from_arrow(array: pa.Array | pa.ChunkedArray, name: str = "arrow_series") -> Series:
-        """
-        Construct a Series from an pyarrow array or chunked array.
+        """Construct a Series from an pyarrow array or chunked array.
 
         Args:
             array: The pyarrow (chunked) array whose data we wish to put in the Series.
             name: The name associated with the Series; this is usually the column name.
         """
+        _ensure_registered_super_ext_type()
         if DataType.from_arrow_type(array.type) == DataType.python():
             # If the Arrow type is not natively supported, go through the Python list path.
             return Series.from_pylist(array.to_pylist(), name=name, pyobj="force")
         elif isinstance(array, pa.Array):
             array = ensure_array(array)
-            if _RAY_DATA_EXTENSIONS_AVAILABLE and isinstance(array.type, ArrowTensorType):
-                storage_series = Series.from_arrow(array.storage, name=name)
-                series = storage_series.cast(
-                    DataType.fixed_size_list(
-                        DataType.from_arrow_type(array.type.scalar_type),
-                        int(np.prod(array.type.shape)),
-                    )
-                )
-                return series.cast(DataType.from_arrow_type(array.type))
-            elif _RAY_DATA_EXTENSIONS_AVAILABLE and isinstance(array.type, ArrowVariableShapedTensorType):
-                return Series.from_numpy(array.to_numpy(zero_copy_only=False), name=name)
-            elif isinstance(array.type, getattr(pa, "FixedShapeTensorType", ())):
+            if isinstance(array.type, getattr(pa, "FixedShapeTensorType", ())):
                 series = Series.from_arrow(array.storage, name=name)
                 return series.cast(DataType.from_arrow_type(array.type))
             else:
@@ -109,7 +72,6 @@ class Series:
                 falling back to Python type representation, or ``"force"`` the data to only
                 have a Python type representation. Default is ``"allow"``.
         """
-
         if not isinstance(data, list):
             raise TypeError(f"expected a python list, got {type(data)}")
 
@@ -131,8 +93,7 @@ class Series:
 
     @classmethod
     def from_numpy(cls, data: np.ndarray, name: str = "numpy_series") -> Series:
-        """
-        Construct a Series from a NumPy ndarray.
+        """Construct a Series from a NumPy ndarray.
 
         If the provided NumPy ndarray is 1-dimensional, Daft will attempt to store the ndarray
         in a pyarrow Array. If the ndarray has more than 1 dimension OR storing the 1D array in Arrow failed,
@@ -158,8 +119,7 @@ class Series:
 
     @classmethod
     def from_pandas(cls, data: pd.Series, name: str = "pd_series") -> Series:
-        """
-        Construct a Series from a pandas Series.
+        """Construct a Series from a pandas Series.
 
         This will first try to convert the series into a pyarrow array, then will fall
         back to converting the series to a NumPy ndarray and going through that construction path,
@@ -242,45 +202,24 @@ class Series:
     def datatype(self) -> DataType:
         return DataType._from_pydatatype(self._series.data_type())
 
-    def to_arrow(self, cast_tensors_to_ray_tensor_dtype: bool = False) -> pa.Array:
-        """
-        Convert this Series to an pyarrow array.
-        """
+    def to_arrow(self) -> pa.Array:
+        """Convert this Series to an pyarrow array."""
+        _ensure_registered_super_ext_type()
+
         dtype = self.datatype()
-        if cast_tensors_to_ray_tensor_dtype and (dtype._is_tensor_type() or dtype._is_fixed_shape_tensor_type()):
-            if not _RAY_DATA_EXTENSIONS_AVAILABLE:
-                raise ValueError("Trying to convert tensors to Ray tensor dtypes, but Ray is not installed.")
-            pyarrow_dtype = dtype.to_arrow_dtype(cast_tensor_to_ray_type=True)
-            if isinstance(pyarrow_dtype, ArrowTensorType):
-                assert dtype._is_fixed_shape_tensor_type()
-                arrow_series = self._series.to_arrow()
-                storage = arrow_series.storage
-                list_size = storage.type.list_size
-                storage = pa.ListArray.from_arrays(
-                    pa.array(
-                        list(range(0, (len(arrow_series) + 1) * list_size, list_size)),
-                        pa.int32(),
-                    ),
-                    storage.values,
-                )
-                return pa.ExtensionArray.from_storage(pyarrow_dtype, storage)
-            else:
-                # Variable-shaped tensor columns can't be converted directly to Ray's variable-shaped tensor extension
-                # type since it expects all tensor elements to have the same number of dimensions, which Daft does not enforce.
-                # TODO(Clark): Convert directly to Ray's variable-shaped tensor extension type when all tensor
-                # elements have the same number of dimensions, without going through pylist roundtrip.
-                return ArrowTensorArray.from_numpy(self.to_pylist())
-        elif dtype._is_fixed_shape_tensor_type() and pyarrow_supports_fixed_shape_tensor():
-            pyarrow_dtype = dtype.to_arrow_dtype(cast_tensor_to_ray_type=False)
+        arrow_arr = self._series.to_arrow()
+
+        # Special-case for PyArrow FixedShapeTensor if it is supported by the version of PyArrow
+        # TODO: Push this down into self._series.to_arrow()?
+        if dtype._is_fixed_shape_tensor_type() and pyarrow_supports_fixed_shape_tensor():
+            pyarrow_dtype = dtype.to_arrow_dtype()
             arrow_series = self._series.to_arrow()
             return pa.ExtensionArray.from_storage(pyarrow_dtype, arrow_series.storage)
-        else:
-            return self._series.to_arrow()
+
+        return arrow_arr
 
     def to_pylist(self) -> list:
-        """
-        Convert this Series to a Python list.
-        """
+        """Convert this Series to a Python list."""
         if self.datatype()._is_python_type():
             return self._series.to_pylist()
         elif self.datatype()._should_cast_to_python():
@@ -306,17 +245,20 @@ class Series:
 
         return Series._from_pyseries(self._series.slice(start, end))
 
-    def argsort(self, descending: bool = False) -> Series:
+    def argsort(self, descending: bool = False, nulls_first: bool | None = None) -> Series:
         if not isinstance(descending, bool):
             raise TypeError(f"expected `descending` to be bool, got {type(descending)}")
+        if nulls_first is None:
+            nulls_first = descending
 
-        return Series._from_pyseries(self._series.argsort(descending))
+        return Series._from_pyseries(self._series.argsort(descending, nulls_first))
 
-    def sort(self, descending: bool = False) -> Series:
+    def sort(self, descending: bool = False, nulls_first: bool | None = None) -> Series:
         if not isinstance(descending, bool):
             raise TypeError(f"expected `descending` to be bool, got {type(descending)}")
-
-        return Series._from_pyseries(self._series.sort(descending))
+        if nulls_first is None:
+            nulls_first = descending
+        return Series._from_pyseries(self._series.sort(descending, nulls_first))
 
     def hash(self, seed: Series | None = None) -> Series:
         if not isinstance(seed, Series) and seed is not None:
@@ -365,9 +307,15 @@ class Series:
     def round(self, decimal: int) -> Series:
         return Series._from_pyseries(self._series.round(decimal))
 
+    def clip(self, min: Series, max: Series) -> Series:
+        return Series._from_pyseries(self._series.clip(min._series, max._series))
+
     def sqrt(self) -> Series:
         return Series._from_pyseries(self._series.sqrt())
 
+    def cbrt(self) -> Series:
+        return Series._from_pyseries(self._series.cbrt())
+
     def sin(self) -> Series:
         """The elementwise sine of a numeric series."""
         return Series._from_pyseries(self._series.sin())
@@ -381,68 +329,69 @@ class Series:
         return Series._from_pyseries(self._series.tan())
 
     def cot(self) -> Series:
-        """The elementwise cotangent of a numeric series"""
+        """The elementwise cotangent of a numeric series."""
         return Series._from_pyseries(self._series.cot())
 
     def arcsin(self) -> Series:
-        """The elementwise arc sine of a numeric series"""
+        """The elementwise arc sine of a numeric series."""
         return Series._from_pyseries(self._series.arcsin())
 
     def arccos(self) -> Series:
-        """The elementwise arc cosine of a numeric series"""
+        """The elementwise arc cosine of a numeric series."""
         return Series._from_pyseries(self._series.arccos())
 
     def arctan(self) -> Series:
-        """The elementwise arc tangent of a numeric series"""
+        """The elementwise arc tangent of a numeric series."""
         return Series._from_pyseries(self._series.arctan())
 
     def arctan2(self, other: Series) -> Series:
-        """Calculates the four quadrant arctangent of coordinates (y, x)"""
+        """Calculates the four quadrant arctangent of coordinates (y, x)."""
         if not isinstance(other, Series):
             raise TypeError(f"expected another Series but got {type(other)}")
         return Series._from_pyseries(self._series.arctan2(other._series))
 
     def arctanh(self) -> Series:
-        """The elementwise inverse hyperbolic tangent of a numeric series"""
+        """The elementwise inverse hyperbolic tangent of a numeric series."""
         return Series._from_pyseries(self._series.arctanh())
 
     def arccosh(self) -> Series:
-        """The elementwise inverse hyperbolic cosine of a numeric series"""
+        """The elementwise inverse hyperbolic cosine of a numeric series."""
         return Series._from_pyseries(self._series.arccosh())
 
     def arcsinh(self) -> Series:
-        """The elementwise inverse hyperbolic sine of a numeric series"""
+        """The elementwise inverse hyperbolic sine of a numeric series."""
         return Series._from_pyseries(self._series.arcsinh())
 
     def radians(self) -> Series:
-        """The elementwise radians of a numeric series"""
+        """The elementwise radians of a numeric series."""
         return Series._from_pyseries(self._series.radians())
 
     def degrees(self) -> Series:
-        """The elementwise degrees of a numeric series"""
+        """The elementwise degrees of a numeric series."""
         return Series._from_pyseries(self._series.degrees())
 
     def log2(self) -> Series:
-        """The elementwise log2 of a numeric series"""
+        """The elementwise log2 of a numeric series."""
         return Series._from_pyseries(self._series.log2())
 
     def log10(self) -> Series:
-        """The elementwise log10 of a numeric series"""
+        """The elementwise log10 of a numeric series."""
         return Series._from_pyseries(self._series.log10())
 
     def log(self, base: float) -> Series:
         """The elementwise log with given base, of a numeric series.
+
         Args:
             base: The base of the logarithm.
         """
         return Series._from_pyseries(self._series.log(base))
 
     def ln(self) -> Series:
-        """The elementwise ln of a numeric series"""
+        """The elementwise ln of a numeric series."""
         return Series._from_pyseries(self._series.ln())
 
     def exp(self) -> Series:
-        """The e^self of a numeric series"""
+        """The e^self of a numeric series."""
         return Series._from_pyseries(self._series.exp())
 
     def __add__(self, other: object) -> Series:
@@ -545,6 +494,12 @@ class Series:
         assert self._series is not None and other._series is not None
         return Series._from_pyseries(self._series ^ other._series)
 
+    def __floordiv__(self, other: object) -> Series:
+        if not isinstance(other, Series):
+            raise TypeError(f"expected another Series but got {type(other)}")
+        assert self._series is not None and other._series is not None
+        return Series._from_pyseries(self._series // other._series)
+
     def count(self, mode: CountMode = CountMode.Valid) -> Series:
         assert self._series is not None
         return Series._from_pyseries(self._series.count(mode))
@@ -561,6 +516,10 @@ class Series:
         assert self._series is not None
         return Series._from_pyseries(self._series.mean())
 
+    def stddev(self) -> Series:
+        assert self._series is not None
+        return Series._from_pyseries(self._series.stddev())
+
     def sum(self) -> Series:
         assert self._series is not None
         return Series._from_pyseries(self._series.sum())
@@ -605,9 +564,9 @@ class Series:
         num_hashes: int,
         ngram_size: int,
         seed: int = 1,
+        hash_function: Literal["murmurhash3", "xxhash", "sha1"] = "murmurhash3",
     ) -> Series:
-        """
-        Runs the MinHash algorithm on the series.
+        """Runs the MinHash algorithm on the series.
 
         For a string, calculates the minimum hash over all its ngrams,
         repeating with `num_hashes` permutations. Returns as a list of 32-bit unsigned integers.
@@ -619,6 +578,7 @@ class Series:
             num_hashes: The number of hash permutations to compute.
             ngram_size: The number of tokens in each shingle/ngram.
             seed (optional): Seed used for generating permutations and the initial string hashes. Defaults to 1.
+            hash_function (optional): Hash function to use for initial string hashing. One of "murmur3", "xxhash", or "sha1". Defaults to "murmur3".
         """
         if not isinstance(num_hashes, int):
             raise ValueError(f"expected an integer for num_hashes but got {type(num_hashes)}")
@@ -626,8 +586,15 @@ class Series:
             raise ValueError(f"expected an integer for ngram_size but got {type(ngram_size)}")
         if seed is not None and not isinstance(seed, int):
             raise ValueError(f"expected an integer or None for seed but got {type(seed)}")
+        if not isinstance(hash_function, str):
+            raise ValueError(f"expected str for hash_function but got {type(hash_function)}")
+        assert hash_function in [
+            "murmurhash3",
+            "xxhash",
+            "sha1",
+        ], f"hash_function must be one of 'murmurhash3', 'xxhash', 'sha1', got {hash_function}"
 
-        return Series._from_pyseries(self._series.minhash(num_hashes, ngram_size, seed))
+        return Series._from_pyseries(self._series.minhash(num_hashes, ngram_size, seed, hash_function))
 
     def _to_str_values(self) -> Series:
         return Series._from_pyseries(self._series.to_str_values())
@@ -677,13 +644,13 @@ class Series:
 def item_to_series(name: str, item: Any) -> Series:
     if isinstance(item, list):
         series = Series.from_pylist(item, name)
-    elif _NUMPY_AVAILABLE and isinstance(item, np.ndarray):
+    elif np.module_available() and isinstance(item, np.ndarray):
         series = Series.from_numpy(item, name)
     elif isinstance(item, Series):
         series = item
     elif isinstance(item, (pa.Array, pa.ChunkedArray)):
         series = Series.from_arrow(item, name)
-    elif _PANDAS_AVAILABLE and isinstance(item, pd.Series):
+    elif pd.module_available() and isinstance(item, pd.Series):
         series = Series.from_pandas(item, name)
     else:
         raise ValueError(f"Creating a Series from data of type {type(item)} not implemented")
@@ -784,6 +751,10 @@ class SeriesStringNamespace(SeriesNamespace):
         assert self._series is not None
         return Series._from_pyseries(self._series.utf8_length())
 
+    def length_bytes(self) -> Series:
+        assert self._series is not None
+        return Series._from_pyseries(self._series.utf8_length_bytes())
+
     def lower(self) -> Series:
         assert self._series is not None
         return Series._from_pyseries(self._series.utf8_lower())
@@ -888,10 +859,10 @@ class SeriesStringNamespace(SeriesNamespace):
     def normalize(
         self,
         *,
-        remove_punct: bool = True,
-        lowercase: bool = True,
-        nfd_unicode: bool = True,
-        white_space: bool = True,
+        remove_punct: bool = False,
+        lowercase: bool = False,
+        nfd_unicode: bool = False,
+        white_space: bool = False,
     ) -> Series:
         if not isinstance(remove_punct, bool):
             raise ValueError(f"expected bool for remove_punct but got {type(remove_punct)}")
@@ -904,6 +875,16 @@ class SeriesStringNamespace(SeriesNamespace):
         assert self._series is not None
         return Series._from_pyseries(self._series.utf8_normalize(remove_punct, lowercase, nfd_unicode, white_space))
 
+    def count_matches(self, patterns: Series, whole_words: bool = False, case_sensitive: bool = True) -> Series:
+        if not isinstance(patterns, Series):
+            raise ValueError(f"expected another Series but got {type(patterns)}")
+        if not isinstance(whole_words, bool):
+            raise ValueError(f"expected bool for whole_word but got {type(whole_words)}")
+        if not isinstance(case_sensitive, bool):
+            raise ValueError(f"expected bool for case_sensitive but got {type(case_sensitive)}")
+        assert self._series is not None and patterns._series is not None
+        return Series._from_pyseries(self._series.utf8_count_matches(patterns._series, whole_words, case_sensitive))
+
 
 class SeriesDateNamespace(SeriesNamespace):
     def date(self) -> Series:
@@ -963,11 +944,29 @@ class SeriesPartitioningNamespace(SeriesNamespace):
 
 class SeriesListNamespace(SeriesNamespace):
     def lengths(self) -> Series:
+        warnings.warn(
+            "This function will be deprecated from Daft version >= 0.3.5!  Instead, please use 'length'",
+            category=DeprecationWarning,
+        )
+
+        return Series._from_pyseries(self._series.list_count(CountMode.All))
+
+    def length(self) -> Series:
         return Series._from_pyseries(self._series.list_count(CountMode.All))
 
     def get(self, idx: Series, default: Series) -> Series:
         return Series._from_pyseries(self._series.list_get(idx._series, default._series))
 
+    def sort(self, desc: bool | Series = False, nulls_first: bool | Series | None = None) -> Series:
+        if isinstance(desc, bool):
+            desc = Series.from_pylist([desc], name="desc")
+        if nulls_first is None:
+            nulls_first = desc
+        elif isinstance(nulls_first, bool):
+            nulls_first = Series.from_pylist([nulls_first], name="nulls_first")
+
+        return Series._from_pyseries(self._series.list_sort(desc._series, nulls_first._series))
+
 
 class SeriesMapNamespace(SeriesNamespace):
     def get(self, key: Series) -> Series:
@@ -975,7 +974,11 @@ class SeriesMapNamespace(SeriesNamespace):
 
 
 class SeriesImageNamespace(SeriesNamespace):
-    def decode(self, on_error: Literal["raise"] | Literal["null"] = "raise") -> Series:
+    def decode(
+        self,
+        on_error: Literal["raise", "null"] = "raise",
+        mode: str | ImageMode | None = None,
+    ) -> Series:
         raise_on_error = False
         if on_error == "raise":
             raise_on_error = True
@@ -983,14 +986,19 @@ class SeriesImageNamespace(SeriesNamespace):
             raise_on_error = False
         else:
             raise NotImplementedError(f"Unimplemented on_error option: {on_error}.")
-        return Series._from_pyseries(self._series.image_decode(raise_error_on_failure=raise_on_error))
+        if mode is not None:
+            if isinstance(mode, str):
+                mode = ImageMode.from_mode_string(mode.upper())
+            if not isinstance(mode, ImageMode):
+                raise ValueError(f"mode must be a string or ImageMode variant, but got: {mode}")
+        return Series._from_pyseries(image.decode(self._series, raise_error_on_failure=raise_on_error, mode=mode))
 
     def encode(self, image_format: str | ImageFormat) -> Series:
         if isinstance(image_format, str):
             image_format = ImageFormat.from_format_string(image_format.upper())
         if not isinstance(image_format, ImageFormat):
             raise ValueError(f"image_format must be a string or ImageFormat variant, but got: {image_format}")
-        return Series._from_pyseries(self._series.image_encode(image_format))
+        return Series._from_pyseries(image.encode(self._series, image_format))
 
     def resize(self, w: int, h: int) -> Series:
         if not isinstance(w, int):
@@ -998,4 +1006,11 @@ class SeriesImageNamespace(SeriesNamespace):
         if not isinstance(h, int):
             raise TypeError(f"expected int for h but got {type(h)}")
 
-        return Series._from_pyseries(self._series.image_resize(w, h))
+        return Series._from_pyseries(image.resize(self._series, w, h))
+
+    def to_mode(self, mode: str | ImageMode) -> Series:
+        if isinstance(mode, str):
+            mode = ImageMode.from_mode_string(mode.upper())
+        if not isinstance(mode, ImageMode):
+            raise ValueError(f"mode must be a string or ImageMode variant, but got: {mode}")
+        return Series._from_pyseries(image.to_mode(self._series, mode))
