@@ -3,13 +3,12 @@ from __future__ import annotations
 import logging
 from typing import TYPE_CHECKING, Any
 
-import pyarrow as pa
-
 from daft.daft import (
     CsvConvertOptions,
     CsvParseOptions,
     CsvReadOptions,
     IOConfig,
+    JoinSide,
     JoinType,
     JsonConvertOptions,
     JsonParseOptions,
@@ -26,14 +25,7 @@ from daft.table.table import Table
 
 if TYPE_CHECKING:
     import pandas as pd
-
-
-_PANDAS_AVAILABLE = True
-try:
-    import pandas as pd
-except ImportError:
-    _PANDAS_AVAILABLE = False
-
+    import pyarrow as pa
 
 logger = logging.getLogger(__name__)
 
@@ -139,10 +131,8 @@ class MicroPartition:
     def to_table(self) -> Table:
         return Table._from_pytable(self._micropartition.to_table())
 
-    def to_arrow(self, cast_tensors_to_ray_tensor_dtype: bool = False, convert_large_arrays: bool = False) -> pa.Table:
-        return self.to_table().to_arrow(
-            cast_tensors_to_ray_tensor_dtype=cast_tensors_to_ray_tensor_dtype, convert_large_arrays=convert_large_arrays
-        )
+    def to_arrow(self) -> pa.Table:
+        return self.to_table().to_arrow()
 
     def to_pydict(self) -> dict[str, list]:
         return self.to_table().to_pydict()
@@ -153,12 +143,10 @@ class MicroPartition:
     def to_pandas(
         self,
         schema: Schema | None = None,
-        cast_tensors_to_ray_tensor_dtype: bool = False,
         coerce_temporal_nanoseconds: bool = False,
     ) -> pd.DataFrame:
         return self.to_table().to_pandas(
             schema=schema,
-            cast_tensors_to_ray_tensor_dtype=cast_tensors_to_ray_tensor_dtype,
             coerce_temporal_nanoseconds=coerce_temporal_nanoseconds,
         )
 
@@ -167,7 +155,7 @@ class MicroPartition:
     ###
 
     def cast_to_schema(self, schema: Schema) -> MicroPartition:
-        """Casts a MicroPartition into the provided schema"""
+        """Casts a MicroPartition into the provided schema."""
         return MicroPartition._from_pymicropartition(self._micropartition.cast_to_schema(schema._schema))
 
     def eval_expression_list(self, exprs: ExpressionsProjection) -> MicroPartition:
@@ -187,7 +175,12 @@ class MicroPartition:
         pyexprs = [e._expr for e in exprs]
         return MicroPartition._from_pymicropartition(self._micropartition.filter(pyexprs))
 
-    def sort(self, sort_keys: ExpressionsProjection, descending: bool | list[bool] | None = None) -> MicroPartition:
+    def sort(
+        self,
+        sort_keys: ExpressionsProjection,
+        descending: bool | list[bool] | None = None,
+        nulls_first: bool | list[bool] | None = None,
+    ) -> MicroPartition:
         assert all(isinstance(e, Expression) for e in sort_keys)
         pyexprs = [e._expr for e in sort_keys]
         if descending is None:
@@ -202,7 +195,21 @@ class MicroPartition:
                 )
         else:
             raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
-        return MicroPartition._from_pymicropartition(self._micropartition.sort(pyexprs, descending))
+        if nulls_first is None:
+            nulls_first = descending
+        elif isinstance(nulls_first, bool):
+            nulls_first = [nulls_first for _ in pyexprs]
+        elif isinstance(nulls_first, list):
+            if len(nulls_first) != len(sort_keys):
+                raise ValueError(
+                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
+                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
+                )
+            else:
+                nulls_first = [bool(x) for x in nulls_first]
+        else:
+            raise TypeError(f"Expected a bool, list[bool] or None for `nulls_first` but got {type(nulls_first)}")
+        return MicroPartition._from_pymicropartition(self._micropartition.sort(pyexprs, descending, nulls_first))
 
     def sample(
         self,
@@ -243,7 +250,7 @@ class MicroPartition:
         return MicroPartition._from_pymicropartition(self._micropartition.quantiles(num))
 
     def explode(self, columns: ExpressionsProjection) -> MicroPartition:
-        """NOTE: Expressions here must be Explode expressions (Expression._explode())"""
+        """NOTE: Expressions here must be Explode expressions."""
         to_explode_pyexprs = [e._expr for e in columns]
         return MicroPartition._from_pymicropartition(self._micropartition.explode(to_explode_pyexprs))
 
@@ -261,6 +268,7 @@ class MicroPartition:
         right: MicroPartition,
         left_on: ExpressionsProjection,
         right_on: ExpressionsProjection,
+        null_equals_nulls: list[bool] | None = None,
         how: JoinType = JoinType.Inner,
     ) -> MicroPartition:
         if len(left_on) != len(right_on):
@@ -275,7 +283,13 @@ class MicroPartition:
         right_exprs = [e._expr for e in right_on]
 
         return MicroPartition._from_pymicropartition(
-            self._micropartition.hash_join(right._micropartition, left_on=left_exprs, right_on=right_exprs, how=how)
+            self._micropartition.hash_join(
+                right._micropartition,
+                left_on=left_exprs,
+                right_on=right_exprs,
+                null_equals_nulls=null_equals_nulls,
+                how=how,
+            )
         )
 
     def sort_merge_join(
@@ -305,6 +319,15 @@ class MicroPartition:
             )
         )
 
+    def cross_join(
+        self,
+        right: MicroPartition,
+        outer_loop_side: JoinSide,
+    ) -> MicroPartition:
+        return MicroPartition._from_pymicropartition(
+            self._micropartition.cross_join(right._micropartition, outer_loop_side=outer_loop_side)
+        )
+
     def partition_by_hash(self, exprs: ExpressionsProjection, num_partitions: int) -> list[MicroPartition]:
         if not isinstance(num_partitions, int):
             raise TypeError(f"Expected a num_partitions to be int, got {type(num_partitions)}")
@@ -355,7 +378,12 @@ class MicroPartition:
     # Compute methods (MicroPartition -> Series)
     ###
 
-    def argsort(self, sort_keys: ExpressionsProjection, descending: bool | list[bool] | None = None) -> Series:
+    def argsort(
+        self,
+        sort_keys: ExpressionsProjection,
+        descending: bool | list[bool] | None = None,
+        nulls_first: bool | list[bool] | None = None,
+    ) -> Series:
         assert all(isinstance(e, Expression) for e in sort_keys)
         pyexprs = [e._expr for e in sort_keys]
         if descending is None:
@@ -370,7 +398,21 @@ class MicroPartition:
                 )
         else:
             raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
-        return Series._from_pyseries(self._micropartition.argsort(pyexprs, descending))
+        if nulls_first is None:
+            nulls_first = descending
+        elif isinstance(nulls_first, bool):
+            nulls_first = [nulls_first for _ in pyexprs]
+        elif isinstance(nulls_first, list):
+            if len(nulls_first) != len(sort_keys):
+                raise ValueError(
+                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
+                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
+                )
+            else:
+                nulls_first = [bool(x) for x in nulls_first]
+        else:
+            raise TypeError(f"Expected a bool, list[bool] or None for `nulls_first` but got {type(nulls_first)}")
+        return Series._from_pyseries(self._micropartition.argsort(pyexprs, descending, nulls_first))
 
     def __reduce__(self) -> tuple:
         names = self.column_names()
