@@ -1,9 +1,7 @@
 from __future__ import annotations
 
 import logging
-from typing import TYPE_CHECKING, Any
-
-import pyarrow as pa
+from typing import TYPE_CHECKING, Any, Literal
 
 from daft.arrow_utils import ensure_table
 from daft.daft import (
@@ -25,30 +23,14 @@ from daft.daft import read_parquet_into_pyarrow as _read_parquet_into_pyarrow
 from daft.daft import read_parquet_into_pyarrow_bulk as _read_parquet_into_pyarrow_bulk
 from daft.daft import read_parquet_statistics as _read_parquet_statistics
 from daft.datatype import DataType, TimeUnit
+from daft.dependencies import pa, pd
 from daft.expressions import Expression, ExpressionsProjection
 from daft.logical.schema import Schema
 from daft.series import Series, item_to_series
 
-_NUMPY_AVAILABLE = True
-try:
-    import numpy as np
-except ImportError:
-    _NUMPY_AVAILABLE = False
-
-_PANDAS_AVAILABLE = True
-try:
-    import pandas as pd
-except ImportError:
-    _PANDAS_AVAILABLE = False
-
 if TYPE_CHECKING:
-    import numpy as np
-    import pandas as pd
-    import pyarrow as pa
-
     from daft.io import IOConfig
 
-
 logger = logging.getLogger(__name__)
 
 
@@ -111,6 +93,8 @@ class Table:
             if field.dtype == DataType.python()
             or field.dtype._is_tensor_type()
             or field.dtype._is_fixed_shape_tensor_type()
+            or field.dtype._is_sparse_tensor_type()
+            or field.dtype._is_fixed_shape_sparse_tensor_type()
         ]
         if non_native_fields:
             # If there are any contained Arrow types that are not natively supported, go through Table.from_pydict()
@@ -131,7 +115,7 @@ class Table:
 
     @staticmethod
     def from_pandas(pd_df: pd.DataFrame) -> Table:
-        if not _PANDAS_AVAILABLE:
+        if not pd.module_available():
             raise ImportError("Unable to import Pandas - please ensure that it is installed.")
         assert isinstance(pd_df, pd.DataFrame)
         try:
@@ -173,39 +157,12 @@ class Table:
     ###
 
     def to_table(self) -> Table:
-        """For compatibility with MicroPartition"""
+        """For compatibility with MicroPartition."""
         return self
 
-    def to_arrow(self, cast_tensors_to_ray_tensor_dtype: bool = False, convert_large_arrays: bool = False) -> pa.Table:
-        python_fields = set()
-        tensor_fields = set()
-        for field in self.schema():
-            if field.dtype._is_python_type():
-                python_fields.add(field.name)
-            elif field.dtype._is_tensor_type() or field.dtype._is_fixed_shape_tensor_type():
-                tensor_fields.add(field.name)
-        if python_fields or tensor_fields:
-            table = {}
-            for colname in self.column_names():
-                column_series = self.get_column(colname)
-                if colname in python_fields:
-                    column = column_series.to_pylist()
-                else:
-                    column = column_series.to_arrow(cast_tensors_to_ray_tensor_dtype)
-                table[colname] = column
-
-            tab = pa.Table.from_pydict(table)
-        else:
-            tab = pa.Table.from_batches([self._table.to_arrow_record_batch()])
-
-        if not convert_large_arrays:
-            return tab
-
-        new_columns = []
-        for col in tab.columns:
-            new_columns.append(_trim_pyarrow_large_arrays(col))
-
-        return pa.Table.from_arrays(new_columns, names=tab.column_names)
+    def to_arrow(self) -> pa.Table:
+        tab = pa.Table.from_pydict({colname: self.get_column(colname).to_arrow() for colname in self.column_names()})
+        return tab
 
     def to_pydict(self) -> dict[str, list]:
         return {colname: self.get_column(colname).to_pylist() for colname in self.column_names()}
@@ -220,13 +177,13 @@ class Table:
     def to_pandas(
         self,
         schema: Schema | None = None,
-        cast_tensors_to_ray_tensor_dtype: bool = False,
         coerce_temporal_nanoseconds: bool = False,
     ) -> pd.DataFrame:
         from packaging.version import parse
 
-        if not _PANDAS_AVAILABLE:
+        if not pd.module_available():
             raise ImportError("Unable to import Pandas - please ensure that it is installed.")
+
         python_fields = set()
         tensor_fields = set()
         for field in self.schema():
@@ -234,16 +191,17 @@ class Table:
                 python_fields.add(field.name)
             elif field.dtype._is_tensor_type() or field.dtype._is_fixed_shape_tensor_type():
                 tensor_fields.add(field.name)
+
         if python_fields or tensor_fields:
-            # Use Python list representation for Python typed columns.
             table = {}
             for colname in self.column_names():
                 column_series = self.get_column(colname)
-                if colname in python_fields or (colname in tensor_fields and not cast_tensors_to_ray_tensor_dtype):
+                # Use Python list representation for Python typed columns or tensor columns (return as numpy)
+                if colname in python_fields or colname in tensor_fields:
                     column = column_series.to_pylist()
                 else:
                     # Arrow-native field, so provide column as Arrow array.
-                    column_arrow = column_series.to_arrow(cast_tensors_to_ray_tensor_dtype)
+                    column_arrow = column_series.to_arrow()
                     if parse(pa.__version__) < parse("13.0.0"):
                         column = column_arrow.to_pandas()
                     else:
@@ -252,7 +210,7 @@ class Table:
 
             return pd.DataFrame.from_dict(table)
         else:
-            arrow_table = self.to_arrow(cast_tensors_to_ray_tensor_dtype)
+            arrow_table = self.to_arrow()
             if parse(pa.__version__) < parse("13.0.0"):
                 return arrow_table.to_pandas()
             else:
@@ -263,7 +221,7 @@ class Table:
     ###
 
     def cast_to_schema(self, schema: Schema) -> Table:
-        """Casts a Table into the provided schema"""
+        """Casts a Table into the provided schema."""
         return Table._from_pytable(self._table.cast_to_schema(schema._schema))
 
     def eval_expression_list(self, exprs: ExpressionsProjection) -> Table:
@@ -283,7 +241,12 @@ class Table:
         pyexprs = [e._expr for e in exprs]
         return Table._from_pytable(self._table.filter(pyexprs))
 
-    def sort(self, sort_keys: ExpressionsProjection, descending: bool | list[bool] | None = None) -> Table:
+    def sort(
+        self,
+        sort_keys: ExpressionsProjection,
+        descending: bool | list[bool] | None = None,
+        nulls_first: bool | list[bool] | None = None,
+    ) -> Table:
         assert all(isinstance(e, Expression) for e in sort_keys)
         pyexprs = [e._expr for e in sort_keys]
         if descending is None:
@@ -298,7 +261,19 @@ class Table:
                 )
         else:
             raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
-        return Table._from_pytable(self._table.sort(pyexprs, descending))
+        if nulls_first is None:
+            nulls_first = descending
+        elif isinstance(nulls_first, bool):
+            nulls_first = [nulls_first for _ in pyexprs]
+        elif isinstance(nulls_first, list):
+            if len(nulls_first) != len(sort_keys):
+                raise ValueError(
+                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
+                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
+                )
+            else:
+                nulls_first = [bool(x) for x in nulls_first]
+        return Table._from_pytable(self._table.sort(pyexprs, descending, nulls_first))
 
     def sample(
         self,
@@ -331,7 +306,7 @@ class Table:
         return Table._from_pytable(self._table.quantiles(num))
 
     def explode(self, columns: ExpressionsProjection) -> Table:
-        """NOTE: Expressions here must be Explode expressions (Expression._explode())"""
+        """NOTE: Expressions here must be Explode expressions."""
         to_explode_pyexprs = [e._expr for e in columns]
         return Table._from_pytable(self._table.explode(to_explode_pyexprs))
 
@@ -420,7 +395,12 @@ class Table:
     # Compute methods (Table -> Series)
     ###
 
-    def argsort(self, sort_keys: ExpressionsProjection, descending: bool | list[bool] | None = None) -> Series:
+    def argsort(
+        self,
+        sort_keys: ExpressionsProjection,
+        descending: bool | list[bool] | None = None,
+        nulls_first: bool | list[bool] | None = None,
+    ) -> Series:
         assert all(isinstance(e, Expression) for e in sort_keys)
         pyexprs = [e._expr for e in sort_keys]
         if descending is None:
@@ -435,7 +415,21 @@ class Table:
                 )
         else:
             raise TypeError(f"Expected a bool, list[bool] or None for `descending` but got {type(descending)}")
-        return Series._from_pyseries(self._table.argsort(pyexprs, descending))
+        if nulls_first is None:
+            nulls_first = descending
+        elif isinstance(nulls_first, bool):
+            nulls_first = [nulls_first for _ in pyexprs]
+        elif isinstance(nulls_first, list):
+            if len(nulls_first) != len(sort_keys):
+                raise ValueError(
+                    f"Expected length of `nulls_first` to be the same length as `sort_keys` since a list was passed in,"
+                    f"got {len(nulls_first)} instead of {len(sort_keys)}"
+                )
+            else:
+                nulls_first = [bool(x) for x in nulls_first]
+        else:
+            raise TypeError(f"Expected a bool, list[bool] or None for `nulls_first` but got {type(nulls_first)}")
+        return Series._from_pyseries(self._table.argsort(pyexprs, descending, nulls_first))
 
     def __reduce__(self) -> tuple:
         names = self.column_names()
@@ -559,30 +553,6 @@ class Table:
         )
 
 
-def _trim_pyarrow_large_arrays(arr: pa.ChunkedArray) -> pa.ChunkedArray:
-    if pa.types.is_large_binary(arr.type) or pa.types.is_large_string(arr.type):
-        if pa.types.is_large_binary(arr.type):
-            target_type = pa.binary()
-        else:
-            target_type = pa.string()
-
-        all_chunks = []
-        for chunk in arr.chunks:
-            if len(chunk) == 0:
-                continue
-            offsets = np.frombuffer(chunk.buffers()[1], dtype=np.int64)
-            if offsets[-1] < 2**31:
-                all_chunks.append(chunk.cast(target_type))
-            else:
-                raise ValueError(
-                    f"Can not convert {arr.type} into {target_type} due to the offset array being too large: {offsets[-1]}. Maximum: {2**31}"
-                )
-
-        return pa.chunked_array(all_chunks, type=target_type)
-    else:
-        return arr
-
-
 def read_parquet_into_pyarrow(
     path: str,
     columns: list[str] | None = None,
@@ -592,6 +562,7 @@ def read_parquet_into_pyarrow(
     io_config: IOConfig | None = None,
     multithreaded_io: bool | None = None,
     coerce_int96_timestamp_unit: TimeUnit = TimeUnit.ns(),
+    string_encoding: Literal["utf-8", "raw"] = "utf-8",
     file_timeout_ms: int | None = 900_000,  # 15 minutes
 ) -> pa.Table:
     fields, metadata, columns, num_rows_read = _read_parquet_into_pyarrow(
@@ -603,6 +574,7 @@ def read_parquet_into_pyarrow(
         io_config=io_config,
         multithreaded_io=multithreaded_io,
         coerce_int96_timestamp_unit=coerce_int96_timestamp_unit._timeunit,
+        string_encoding=string_encoding,
         file_timeout_ms=file_timeout_ms,
     )
     schema = pa.schema(fields, metadata=metadata)
