@@ -5,9 +5,25 @@ import random
 import statistics
 from typing import Any, Callable
 
-import pyarrow as pa
+from daft.dependencies import pa
 
-ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
+
+def get_arrow_version():
+    return tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
+
+
+def in_notebook():
+    """Check if we are in a Jupyter notebook."""
+    try:
+        from IPython import get_ipython
+
+        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
+            return False
+    except ImportError:
+        return False
+    except AttributeError:
+        return False
+    return True
 
 
 def pydict_to_rows(pydict: dict[str, list]) -> list[frozenset[tuple[str, Any]]]:
@@ -118,4 +134,6 @@ def pyarrow_supports_fixed_shape_tensor() -> bool:
     """Whether pyarrow supports the fixed_shape_tensor canonical extension type."""
     from daft.context import get_context
 
-    return hasattr(pa, "fixed_shape_tensor") and (not get_context().is_ray_runner or ARROW_VERSION >= (13, 0, 0))
+    return hasattr(pa, "fixed_shape_tensor") and (
+        (get_context().get_or_create_runner().name != "ray") or get_arrow_version() >= (13, 0, 0)
+    )
