@@ -9,12 +9,16 @@ import importlib
 import inspect
 import os
 import subprocess
+import sys
 
 import sphinx_autosummary_accessors
 
 # Set environment variable to help code determine whether or not we are running a Sphinx doc build process
 os.environ["DAFT_SPHINX_BUILD"] = "1"
 
+# Help Sphinx find local custom extensions/directives that we build
+sys.path.insert(0, os.path.abspath("ext"))
+
 # -- Project information -----------------------------------------------------
 # https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
 project = "Daft"
@@ -45,10 +49,15 @@ extensions = [
     "myst_nb",
     "sphinx_copybutton",
     "sphinx_autosummary_accessors",
+    "sphinx_tabs.tabs",
+    # Local extensions
+    "sql_autosummary",
 ]
 
 templates_path = ["_templates", sphinx_autosummary_accessors.templates_path]
 
+# Removes module names that prefix our classes
+add_module_names = False
 
 # -- Options for Notebook rendering
 # https://myst-nb.readthedocs.io/en/latest/configuration.html?highlight=nb_execution_mode#execution
@@ -67,6 +76,10 @@ html_theme_options = {
     # This is how many levels are shown on the secondary sidebar
     "show_toc_level": 2,
 }
+# Set canonical link
+# Based on Google Search's recommendation, we use an absolute path here. For more details, see:
+# https://developers.google.com/search/docs/crawling-indexing/consolidate-duplicate-urls#use-rel=canonical-link-annotations
+html_baseurl = "https://www.getdaft.io/projects/docs/en/stable/"
 
 # -- Copy button configuration
 
@@ -83,9 +96,16 @@ redirects = {
     "learn/install": "../install.html",
     "learn/user_guides/dataframes": "intro-dataframes.html",
     "learn/user_guides/types_and_ops": "intro-dataframes.html",
-    "learn/user_guides/remote_cluster_execution": "scaling-up.html",
+    "learn/user_guides/remote_cluster_execution": "distributed-computing.html",
     "learn/quickstart": "learn/10-min.html",
     "learn/10-min": "../10-min.html",
+    "user_guide/basic_concepts/expressions": "user_guide/expressions",
+    "user_guide/basic_concepts/dataframe_introduction": "user_guide/basic_concepts",
+    "user_guide/basic_concepts/introduction": "user_guide/basic_concepts",
+    "user_guide/daft_in_depth/aggregations": "user_guide/aggregations",
+    "user_guide/daft_in_depth/dataframe-operations": "user_guide/dataframe-operations",
+    "user_guide/daft_in_depth/datatypes": "user_guide/datatypes",
+    "user_guide/daft_in_depth/udf": "user_guide/udf",
 }
 
 # Resolving code links to github
