@@ -1,21 +1,26 @@
-use std::iter::repeat;
-use std::sync::Arc;
+use std::{iter::repeat, sync::Arc};
 
-use crate::datatypes::{Field, Int64Array, Utf8Array};
+use arrow2::offset::OffsetsBuffer;
+use common_error::DaftResult;
+use indexmap::{
+    map::{raw_entry_v1::RawEntryMut, RawEntryApiV1},
+    IndexMap,
+};
+
+use super::as_arrow::AsArrow;
 use crate::{
     array::{
         growable::{make_growable, Growable},
-        FixedSizeListArray, ListArray,
+        ops::arrow2::comparison::build_is_equal,
+        FixedSizeListArray, ListArray, StructArray,
     },
-    datatypes::UInt64Array,
+    count_mode::CountMode,
+    datatypes::{BooleanArray, DataType, Field, Int64Array, UInt64Array, Utf8Array},
+    kernels::search_sorted::build_is_valid,
+    prelude::MapArray,
+    series::{IntoSeries, Series},
+    utils::identity_hash_set::IdentityBuildHasher,
 };
-use crate::{CountMode, DataType};
-
-use crate::series::{IntoSeries, Series};
-
-use common_error::DaftResult;
-
-use super::as_arrow::AsArrow;
 
 fn join_arrow_list_of_utf8s(
     list_element: Option<&dyn arrow2::array::Array>,
@@ -28,7 +33,7 @@ fn join_arrow_list_of_utf8s(
                 .downcast_ref::<arrow2::array::Utf8Array<i64>>()
                 .unwrap()
                 .iter()
-                .fold(String::from(""), |acc, str_item| {
+                .fold(String::new(), |acc, str_item| {
                     acc + str_item.unwrap_or("") + delimiter_str
                 })
             // Remove trailing `delimiter_str`
@@ -46,7 +51,7 @@ fn join_arrow_list_of_utf8s(
 // Given an i64 array that may have either 1 or `self.len()` elements, create an iterator with
 // `self.len()` elements. If there was originally 1 element, we repeat this element `self.len()`
 // times, otherwise we simply take the original array.
-fn create_iter<'a>(arr: &'a Int64Array, len: usize) -> Box<dyn Iterator<Item = i64> + '_> {
+fn create_iter<'a>(arr: &'a Int64Array, len: usize) -> Box<dyn Iterator<Item = i64> + 'a> {
     match arr.len() {
         1 => Box::new(repeat(arr.get(0).unwrap()).take(len)),
         arr_len => {
@@ -196,7 +201,189 @@ fn get_chunks_helper(
     }
 }
 
+fn list_sort_helper(
+    flat_child: &Series,
+    offsets: &OffsetsBuffer<i64>,
+    desc_iter: impl Iterator<Item = bool>,
+    nulls_first_iter: impl Iterator<Item = bool>,
+    validity: impl Iterator<Item = bool>,
+) -> DaftResult<Vec<Series>> {
+    desc_iter
+        .zip(nulls_first_iter)
+        .zip(validity)
+        .enumerate()
+        .map(|(i, ((desc, nulls_first), valid))| {
+            let start = *offsets.get(i).unwrap() as usize;
+            let end = *offsets.get(i + 1).unwrap() as usize;
+            if valid {
+                flat_child.slice(start, end)?.sort(desc, nulls_first)
+            } else {
+                Ok(Series::full_null(
+                    flat_child.name(),
+                    flat_child.data_type(),
+                    end - start,
+                ))
+            }
+        })
+        .collect()
+}
+
+fn list_sort_helper_fixed_size(
+    flat_child: &Series,
+    fixed_size: usize,
+    desc_iter: impl Iterator<Item = bool>,
+    nulls_first_iter: impl Iterator<Item = bool>,
+    validity: impl Iterator<Item = bool>,
+) -> DaftResult<Vec<Series>> {
+    desc_iter
+        .zip(nulls_first_iter)
+        .zip(validity)
+        .enumerate()
+        .map(|(i, ((desc, nulls_first), valid))| {
+            let start = i * fixed_size;
+            let end = (i + 1) * fixed_size;
+            if valid {
+                flat_child.slice(start, end)?.sort(desc, nulls_first)
+            } else {
+                Ok(Series::full_null(
+                    flat_child.name(),
+                    flat_child.data_type(),
+                    end - start,
+                ))
+            }
+        })
+        .collect()
+}
+
 impl ListArray {
+    pub fn value_counts(&self) -> DaftResult<MapArray> {
+        struct IndexRef {
+            index: usize,
+            hash: u64,
+        }
+
+        impl std::hash::Hash for IndexRef {
+            fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
+                self.hash.hash(state);
+            }
+        }
+
+        let original_name = self.name();
+
+        let hashes = self.flat_child.hash(None)?;
+
+        let flat_child = self.flat_child.to_arrow();
+        let flat_child = &*flat_child;
+
+        let is_equal = build_is_equal(
+            flat_child, flat_child,
+            false, // this value does not matter; invalid (= nulls) are never included
+            true,  // NaNs are equal so we do not get a bunch of {Nan: 1, Nan: 1, ...}
+        )?;
+
+        let is_valid = build_is_valid(flat_child);
+
+        let key_type = self.flat_child.data_type().clone();
+        let count_type = DataType::UInt64;
+
+        let mut include_mask = Vec::with_capacity(self.flat_child.len());
+        let mut count_array = Vec::new();
+
+        let mut offsets = Vec::with_capacity(self.len());
+
+        offsets.push(0_i64);
+
+        let mut map: IndexMap<IndexRef, u64, IdentityBuildHasher> = IndexMap::default();
+        for range in self.offsets().ranges() {
+            map.clear();
+
+            for index in range {
+                let index = index as usize;
+                if !is_valid(index) {
+                    include_mask.push(false);
+                    // skip nulls
+                    continue;
+                }
+
+                let hash = hashes.get(index).unwrap();
+
+                let entry = map
+                    .raw_entry_mut_v1()
+                    .from_hash(hash, |other| is_equal(other.index, index));
+
+                match entry {
+                    RawEntryMut::Occupied(mut entry) => {
+                        include_mask.push(false);
+                        *entry.get_mut() += 1;
+                    }
+                    RawEntryMut::Vacant(vacant) => {
+                        include_mask.push(true);
+                        vacant.insert(IndexRef { index, hash }, 1);
+                    }
+                }
+            }
+
+            // IndexMap maintains insertion order, so we iterate over its values
+            // in the same order that elements were added. This ensures that
+            // the count_array values correspond to the same order in which
+            // the include_mask was set earlier in the loop. Each 'true' in
+            // include_mask represents a unique key, and its corresponding
+            // count is now added to count_array in the same sequence.
+            for v in map.values() {
+                count_array.push(*v);
+            }
+
+            offsets.push(count_array.len() as i64);
+        }
+
+        let values = UInt64Array::from(("count", count_array)).into_series();
+        let include_mask = BooleanArray::from(("boolean", include_mask.as_slice()));
+
+        let keys = self.flat_child.filter(&include_mask)?;
+
+        let keys = Series::try_from_field_and_arrow_array(
+            Field::new("key", key_type.clone()),
+            keys.to_arrow(),
+        )?;
+
+        let values = Series::try_from_field_and_arrow_array(
+            Field::new("value", count_type.clone()),
+            values.to_arrow(),
+        )?;
+
+        let struct_type = DataType::Struct(vec![
+            Field::new("key", key_type.clone()),
+            Field::new("value", count_type.clone()),
+        ]);
+
+        let struct_array = StructArray::new(
+            Arc::new(Field::new("entries", struct_type.clone())),
+            vec![keys, values],
+            None,
+        );
+
+        let list_type = DataType::List(Box::new(struct_type));
+
+        let offsets = OffsetsBuffer::try_from(offsets)?;
+
+        let list_array = Self::new(
+            Arc::new(Field::new("entries", list_type)),
+            struct_array.into_series(),
+            offsets,
+            None,
+        );
+
+        let map_type = DataType::Map {
+            key: Box::new(key_type),
+            value: Box::new(count_type),
+        };
+
+        Ok(MapArray::new(
+            Field::new(original_name, map_type),
+            list_array,
+        ))
+    }
+
     pub fn count(&self, mode: CountMode) -> DaftResult<UInt64Array> {
         let counts = match (mode, self.flat_child.validity()) {
             (CountMode::All, _) | (CountMode::Valid, None) => {
@@ -339,9 +526,12 @@ impl ListArray {
         self.get_children_helper(idx_iter, default)
     }
 
-    pub fn get_slices(&self, start: &Int64Array, end: &Int64Array) -> DaftResult<Series> {
+    pub fn get_slices(&self, start: &Int64Array, end: Option<&Int64Array>) -> DaftResult<Series> {
         let start_iter = create_iter(start, self.len());
-        let end_iter = create_iter(end, self.len());
+        let end_iter = match end {
+            Some(end) => create_iter(end, self.len()),
+            None => Box::new(self.offsets().windows(2).map(|w| (w[1] - w[0]))),
+        };
         get_slices_helper(
             self.offsets().iter().copied(),
             self.field.clone(),
@@ -380,9 +570,69 @@ impl ListArray {
             new_offsets,
         )
     }
+
+    // Sorts the lists within a list column
+    pub fn list_sort(&self, desc: &BooleanArray, nulls_first: &BooleanArray) -> DaftResult<Self> {
+        let offsets = self.offsets();
+        let child_series = if desc.len() == 1 {
+            let desc_iter = repeat(desc.get(0).unwrap()).take(self.len());
+            let nulls_first_iter = repeat(nulls_first.get(0).unwrap()).take(self.len());
+            if let Some(validity) = self.validity() {
+                list_sort_helper(
+                    &self.flat_child,
+                    offsets,
+                    desc_iter,
+                    nulls_first_iter,
+                    validity.iter(),
+                )?
+            } else {
+                list_sort_helper(
+                    &self.flat_child,
+                    offsets,
+                    desc_iter,
+                    nulls_first_iter,
+                    repeat(true).take(self.len()),
+                )?
+            }
+        } else {
+            let desc_iter = desc.as_arrow().values_iter();
+            let nulls_first_iter = nulls_first.as_arrow().values_iter();
+            if let Some(validity) = self.validity() {
+                list_sort_helper(
+                    &self.flat_child,
+                    offsets,
+                    desc_iter,
+                    nulls_first_iter,
+                    validity.iter(),
+                )?
+            } else {
+                list_sort_helper(
+                    &self.flat_child,
+                    offsets,
+                    desc_iter,
+                    nulls_first_iter,
+                    repeat(true).take(self.len()),
+                )?
+            }
+        };
+
+        let child_refs: Vec<&Series> = child_series.iter().collect();
+        let child = Series::concat(&child_refs)?;
+        Ok(Self::new(
+            self.field.clone(),
+            child,
+            self.offsets().clone(),
+            self.validity().cloned(),
+        ))
+    }
 }
 
 impl FixedSizeListArray {
+    pub fn value_counts(&self) -> DaftResult<MapArray> {
+        let list = self.to_list();
+        list.value_counts()
+    }
+
     pub fn count(&self, mode: CountMode) -> DaftResult<UInt64Array> {
         let size = self.fixed_element_len();
         let counts = match (mode, self.flat_child.validity()) {
@@ -511,11 +761,14 @@ impl FixedSizeListArray {
         self.get_children_helper(idx_iter, default)
     }
 
-    pub fn get_slices(&self, start: &Int64Array, end: &Int64Array) -> DaftResult<Series> {
-        let start_iter = create_iter(start, self.len());
-        let end_iter = create_iter(end, self.len());
-        let new_field = Arc::new(self.field.to_exploded_field()?.to_list_field()?);
+    pub fn get_slices(&self, start: &Int64Array, end: Option<&Int64Array>) -> DaftResult<Series> {
         let list_size = self.fixed_element_len();
+        let start_iter = create_iter(start, self.len());
+        let end_iter = match end {
+            Some(end) => create_iter(end, self.len()),
+            None => Box::new(repeat(list_size as i64).take(self.len())),
+        };
+        let new_field = Arc::new(self.field.to_exploded_field()?.to_list_field()?);
         get_slices_helper(
             (0..=((self.len() * list_size) as i64)).step_by(list_size),
             new_field,
@@ -554,6 +807,61 @@ impl FixedSizeListArray {
             new_offsets,
         )
     }
+
+    // Sorts the lists within a list column
+    pub fn list_sort(&self, desc: &BooleanArray, nulls_first: &BooleanArray) -> DaftResult<Self> {
+        let fixed_size = self.fixed_element_len();
+
+        let child_series = if desc.len() == 1 {
+            let desc_iter = repeat(desc.get(0).unwrap()).take(self.len());
+            let nulls_first_iter = repeat(nulls_first.get(0).unwrap()).take(self.len());
+            if let Some(validity) = self.validity() {
+                list_sort_helper_fixed_size(
+                    &self.flat_child,
+                    fixed_size,
+                    desc_iter,
+                    nulls_first_iter,
+                    validity.iter(),
+                )?
+            } else {
+                list_sort_helper_fixed_size(
+                    &self.flat_child,
+                    fixed_size,
+                    desc_iter,
+                    nulls_first_iter,
+                    repeat(true).take(self.len()),
+                )?
+            }
+        } else {
+            let desc_iter = desc.as_arrow().values_iter();
+            let nulls_first_iter = nulls_first.as_arrow().values_iter();
+            if let Some(validity) = self.validity() {
+                list_sort_helper_fixed_size(
+                    &self.flat_child,
+                    fixed_size,
+                    desc_iter,
+                    nulls_first_iter,
+                    validity.iter(),
+                )?
+            } else {
+                list_sort_helper_fixed_size(
+                    &self.flat_child,
+                    fixed_size,
+                    desc_iter,
+                    nulls_first_iter,
+                    repeat(true).take(self.len()),
+                )?
+            }
+        };
+
+        let child_refs: Vec<&Series> = child_series.iter().collect();
+        let child = Series::concat(&child_refs)?;
+        Ok(Self::new(
+            self.field.clone(),
+            child,
+            self.validity().cloned(),
+        ))
+    }
 }
 
 macro_rules! impl_aggs_list_array {
