@@ -1,17 +1,24 @@
+use core::str;
 use std::{collections::HashMap, num::NonZeroUsize, sync::Arc};
 
 use arrow2::{
     datatypes::Field,
-    io::csv::read_async::{read_rows, AsyncReaderBuilder, ByteRecord},
+    io::csv::{
+        read_async,
+        read_async::{read_rows, AsyncReaderBuilder},
+    },
 };
 use async_compat::{Compat, CompatExt};
 use common_error::{DaftError, DaftResult};
+use common_runtime::get_io_runtime;
 use csv_async::AsyncReader;
-use daft_core::{schema::Schema, utils::arrow::cast_array_for_daft_if_needed, Series};
+use daft_compression::CompressionCodec;
+use daft_core::{prelude::*, utils::arrow::cast_array_for_daft_if_needed};
+use daft_decoding::deserialize::deserialize_column;
 use daft_dsl::optimization::get_required_columns;
-use daft_io::{get_runtime, GetResult, IOClient, IOStatsRef};
+use daft_io::{parse_url, GetResult, IOClient, IOStatsRef, SourceType};
 use daft_table::Table;
-use futures::{Stream, StreamExt, TryStreamExt};
+use futures::{stream::BoxStream, Stream, StreamExt, TryStreamExt};
 use rayon::{
     iter::{IndexedParallelIterator, IntoParallelIterator},
     prelude::{IntoParallelRefIterator, ParallelIterator},
@@ -27,13 +34,18 @@ use tokio::{
 };
 use tokio_util::io::StreamReader;
 
-use crate::ArrowSnafu;
-use crate::{metadata::read_csv_schema_single, CsvConvertOptions, CsvParseOptions, CsvReadOptions};
-use daft_compression::CompressionCodec;
-use daft_decoding::deserialize::deserialize_column;
+use crate::{
+    metadata::read_csv_schema_single, ArrowSnafu, CsvConvertOptions, CsvParseOptions,
+    CsvReadOptions,
+};
 
-trait ByteRecordChunkStream: Stream<Item = super::Result<Vec<ByteRecord>>> {}
-impl<S> ByteRecordChunkStream for S where S: Stream<Item = super::Result<Vec<ByteRecord>>> {}
+trait ByteRecordChunkStream: Stream<Item = super::Result<Vec<read_async::ByteRecord>>> {}
+impl<S> ByteRecordChunkStream for S where
+    S: Stream<Item = super::Result<Vec<read_async::ByteRecord>>>
+{
+}
+
+use crate::local::{read_csv_local, stream_csv_local};
 
 type TableChunkResult =
     super::Result<Context<JoinHandle<DaftResult<Table>>, super::JoinSnafu, super::Error>>;
@@ -51,9 +63,8 @@ pub fn read_csv(
     multithreaded_io: bool,
     max_chunks_in_flight: Option<usize>,
 ) -> DaftResult<Table> {
-    let runtime_handle = get_runtime(multithreaded_io)?;
-    let _rt_guard = runtime_handle.enter();
-    runtime_handle.block_on(async {
+    let runtime_handle = get_io_runtime(multithreaded_io);
+    runtime_handle.block_on_current_thread(async {
         read_csv_single_into_table(
             uri,
             convert_options,
@@ -79,13 +90,12 @@ pub fn read_csv_bulk(
     max_chunks_in_flight: Option<usize>,
     num_parallel_tasks: usize,
 ) -> DaftResult<Vec<Table>> {
-    let runtime_handle = get_runtime(multithreaded_io)?;
-    let _rt_guard = runtime_handle.enter();
-    let tables = runtime_handle.block_on(async move {
+    let runtime_handle = get_io_runtime(multithreaded_io);
+    let tables = runtime_handle.block_on_current_thread(async move {
         // Launch a read task per URI, throttling the number of concurrent file reads to num_parallel tasks.
         let task_stream = futures::stream::iter(uris.iter().map(|uri| {
             let (uri, convert_options, parse_options, read_options, io_client, io_stats) = (
-                uri.to_string(),
+                (*uri).to_string(),
                 convert_options.clone(),
                 parse_options.clone(),
                 read_options.clone(),
@@ -135,9 +145,47 @@ pub fn read_csv_bulk(
     tables.into_iter().collect::<DaftResult<Vec<_>>>()
 }
 
-// Parallel version of table concat
-// get rid of this once Table APIs are parallel
-fn tables_concat(mut tables: Vec<Table>) -> DaftResult<Table> {
+#[allow(clippy::too_many_arguments)]
+pub async fn stream_csv(
+    uri: String,
+    convert_options: Option<CsvConvertOptions>,
+    parse_options: Option<CsvParseOptions>,
+    read_options: Option<CsvReadOptions>,
+    io_client: Arc<IOClient>,
+    io_stats: Option<IOStatsRef>,
+    max_chunks_in_flight: Option<usize>,
+) -> DaftResult<BoxStream<'static, DaftResult<Table>>> {
+    let uri = uri.as_str();
+    let (source_type, _) = parse_url(uri)?;
+    let is_compressed = CompressionCodec::from_uri(uri).is_some();
+    if matches!(source_type, SourceType::File) && !is_compressed {
+        let stream = stream_csv_local(
+            uri,
+            convert_options,
+            parse_options.unwrap_or_default(),
+            read_options,
+            io_client,
+            io_stats,
+            max_chunks_in_flight,
+        )
+        .await?;
+        Ok(Box::pin(stream))
+    } else {
+        let stream = stream_csv_single(
+            uri,
+            convert_options,
+            parse_options,
+            read_options,
+            io_client,
+            io_stats,
+            max_chunks_in_flight,
+        )
+        .await?;
+        Ok(Box::pin(stream))
+    }
+}
+
+pub fn tables_concat(mut tables: Vec<Table>) -> DaftResult<Table> {
     if tables.is_empty() {
         return Err(DaftError::ValueError(
             "Need at least 1 Table to perform concat".to_string(),
@@ -171,10 +219,11 @@ fn tables_concat(mut tables: Vec<Table>) -> DaftResult<Table> {
     Table::new_with_size(
         first_table.schema.clone(),
         new_series,
-        tables.iter().map(|t| t.len()).sum(),
+        tables.iter().map(daft_table::Table::len).sum(),
     )
 }
 
+#[allow(clippy::too_many_arguments)]
 async fn read_csv_single_into_table(
     uri: &str,
     convert_options: Option<CsvConvertOptions>,
@@ -184,6 +233,21 @@ async fn read_csv_single_into_table(
     io_stats: Option<IOStatsRef>,
     max_chunks_in_flight: Option<usize>,
 ) -> DaftResult<Table> {
+    let (source_type, _) = parse_url(uri)?;
+    let is_compressed = CompressionCodec::from_uri(uri).is_some();
+    if matches!(source_type, SourceType::File) && !is_compressed {
+        return read_csv_local(
+            uri,
+            convert_options,
+            parse_options.unwrap_or_default(),
+            read_options,
+            io_client,
+            io_stats,
+            max_chunks_in_flight,
+        )
+        .await;
+    }
+
     let predicate = convert_options
         .as_ref()
         .and_then(|opts| opts.predicate.clone());
@@ -202,7 +266,7 @@ async fn read_csv_single_into_table(
                 let required_columns_for_predicate = get_required_columns(predicate);
                 for rc in required_columns_for_predicate {
                     if include_columns.iter().all(|c| c.as_str() != rc.as_str()) {
-                        include_columns.push(rc)
+                        include_columns.push(rc);
                     }
                 }
             }
@@ -301,6 +365,97 @@ async fn read_csv_single_into_table(
     }
 }
 
+pub async fn stream_csv_single(
+    uri: &str,
+    convert_options: Option<CsvConvertOptions>,
+    parse_options: Option<CsvParseOptions>,
+    read_options: Option<CsvReadOptions>,
+    io_client: Arc<IOClient>,
+    io_stats: Option<IOStatsRef>,
+    max_chunks_in_flight: Option<usize>,
+) -> DaftResult<impl Stream<Item = DaftResult<Table>> + Send> {
+    let predicate = convert_options
+        .as_ref()
+        .and_then(|opts| opts.predicate.clone());
+
+    let limit = convert_options.as_ref().and_then(|opts| opts.limit);
+
+    let include_columns = convert_options
+        .as_ref()
+        .and_then(|opts| opts.include_columns.clone());
+
+    let convert_options_with_predicate_columns = match (convert_options, &predicate) {
+        (None, _) => None,
+        (co, None) => co,
+        (Some(mut co), Some(predicate)) => {
+            if let Some(ref mut include_columns) = co.include_columns {
+                let required_columns_for_predicate = get_required_columns(predicate);
+                for rc in required_columns_for_predicate {
+                    if include_columns.iter().all(|c| c.as_str() != rc.as_str()) {
+                        include_columns.push(rc);
+                    }
+                }
+            }
+            // if we have a limit and a predicate, remove limit for stream
+            co.limit = None;
+            Some(co)
+        }
+    };
+
+    let (chunk_stream, _fields) = read_csv_single_into_stream(
+        uri,
+        convert_options_with_predicate_columns.unwrap_or_default(),
+        parse_options.unwrap_or_default(),
+        read_options,
+        io_client,
+        io_stats,
+    )
+    .await?;
+    // Default max chunks in flight is set to 2x the number of cores, which should ensure pipelining of reading chunks
+    // with the parsing of chunks on the rayon threadpool.
+    let max_chunks_in_flight = max_chunks_in_flight.unwrap_or_else(|| {
+        std::thread::available_parallelism()
+            .unwrap_or(NonZeroUsize::new(2).unwrap())
+            .checked_mul(2.try_into().unwrap())
+            .unwrap()
+            .into()
+    });
+    // Collect all chunks in chunk x column form.
+    let tables = chunk_stream
+        // Limit the number of chunks we have in flight at any given time.
+        .try_buffered(max_chunks_in_flight);
+
+    let filtered_tables = tables.map(move |table| {
+        let table = table?;
+        if let Some(predicate) = &predicate {
+            let filtered = table?.filter(&[predicate.clone()])?;
+            if let Some(include_columns) = &include_columns {
+                filtered.get_columns(include_columns.as_slice())
+            } else {
+                Ok(filtered)
+            }
+        } else {
+            table
+        }
+    });
+
+    let mut remaining_rows = limit.map(|limit| limit as i64);
+    let tables = filtered_tables.try_take_while(move |table| {
+        match remaining_rows {
+            // Limit has been met, early-terminate.
+            Some(rows_left) if rows_left <= 0 => futures::future::ready(Ok(false)),
+            // Limit has not yet been met, update remaining limit slack and continue.
+            Some(rows_left) => {
+                remaining_rows = Some(rows_left - table.len() as i64);
+                futures::future::ready(Ok(true))
+            }
+            // No limit, never early-terminate.
+            None => futures::future::ready(Ok(true)),
+        }
+    });
+    Ok(tables)
+}
+
 async fn read_csv_single_into_stream(
     uri: &str,
     convert_options: CsvConvertOptions,
@@ -309,10 +464,10 @@ async fn read_csv_single_into_stream(
     io_client: Arc<IOClient>,
     io_stats: Option<IOStatsRef>,
 ) -> DaftResult<(impl TableStream + Send, Vec<Field>)> {
-    let (mut schema, estimated_mean_row_size, estimated_std_row_size) = match convert_options.schema
-    {
-        Some(schema) => (schema.to_arrow()?, None, None),
-        None => {
+    let (mut schema, estimated_mean_row_size, estimated_std_row_size) =
+        if let Some(schema) = convert_options.schema {
+            (schema.to_arrow()?, None, None)
+        } else {
             let (schema, read_stats) = read_csv_schema_single(
                 uri,
                 parse_options.clone(),
@@ -327,8 +482,7 @@ async fn read_csv_single_into_stream(
                 Some(read_stats.mean_record_size_bytes),
                 Some(read_stats.stddev_record_size_bytes),
             )
-        }
-    };
+        };
     // Rename fields, if necessary.
     if let Some(column_names) = convert_options.column_names {
         schema = schema
@@ -442,7 +596,7 @@ where
                 estimated_rows_per_desired_chunk.max(8).min(num_rows - total_rows_read)
             };
             let mut chunk_buffer = vec![
-                ByteRecord::with_capacity(record_buffer_size, num_fields);
+                read_async::ByteRecord::with_capacity(record_buffer_size, num_fields);
                 chunk_size_rows
             ];
 
@@ -461,7 +615,7 @@ where
 
             chunk_buffer.truncate(rows_read);
             if rows_read > 0 {
-                yield chunk_buffer
+                yield chunk_buffer;
             }
         }
     }
@@ -479,7 +633,7 @@ fn parse_into_column_array_chunk_stream(
         .iter()
         .map(|i| fields.get(*i).unwrap().into())
         .collect::<Vec<daft_core::datatypes::Field>>();
-    let read_schema = Arc::new(daft_core::schema::Schema::new(fields_subset)?);
+    let read_schema = Arc::new(daft_core::prelude::Schema::new(fields_subset)?);
     let read_daft_fields = Arc::new(
         read_schema
             .fields
@@ -512,7 +666,7 @@ fn parse_into_column_array_chunk_stream(
                             )
                         })
                         .collect::<DaftResult<Vec<Series>>>()?;
-                    let num_rows = chunk.first().map(|s| s.len()).unwrap_or(0);
+                    let num_rows = chunk.first().map_or(0, daft_core::series::Series::len);
                     Ok(Table::new_unchecked(read_schema, chunk, num_rows))
                 })();
                 let _ = send.send(result);
@@ -523,7 +677,7 @@ fn parse_into_column_array_chunk_stream(
     }))
 }
 
-fn fields_to_projection_indices(
+pub fn fields_to_projection_indices(
     fields: &[arrow2::datatypes::Field],
     include_columns: &Option<Vec<String>>,
 ) -> Arc<Vec<usize>> {
@@ -549,25 +703,21 @@ fn fields_to_projection_indices(
 mod tests {
     use std::sync::Arc;
 
-    use common_error::{DaftError, DaftResult};
-
     use arrow2::io::csv::read::{
         deserialize_batch, deserialize_column, infer, infer_schema, read_rows, ByteRecord,
         ReaderBuilder,
     };
+    use common_error::{DaftError, DaftResult};
     use daft_core::{
-        datatypes::Field,
-        schema::Schema,
+        prelude::*,
         utils::arrow::{cast_array_for_daft_if_needed, cast_array_from_daft_if_needed},
-        DataType,
     };
     use daft_io::{IOClient, IOConfig};
     use daft_table::Table;
     use rstest::rstest;
 
-    use crate::{char_to_byte, CsvConvertOptions, CsvParseOptions, CsvReadOptions};
-
     use super::read_csv;
+    use crate::{char_to_byte, CsvConvertOptions, CsvParseOptions, CsvReadOptions};
 
     #[allow(clippy::too_many_arguments)]
     fn check_equal_local_arrow2(
@@ -656,7 +806,7 @@ mod tests {
         let file = format!(
             "{}/test/iris_tiny.csv{}",
             env!("CARGO_MANIFEST_DIR"),
-            compression.map_or("".to_string(), |ext| format!(".{}", ext))
+            compression.map_or(String::new(), |ext| format!(".{ext}"))
         );
 
         let mut io_config = IOConfig::default();
@@ -717,10 +867,9 @@ mod tests {
         ];
         let table = read_csv(
             file.as_ref(),
-            Some(
-                CsvConvertOptions::default()
-                    .with_column_names(Some(column_names.iter().map(|s| s.to_string()).collect())),
-            ),
+            Some(CsvConvertOptions::default().with_column_names(Some(
+                column_names.iter().map(|s| (*s).to_string()).collect(),
+            ))),
             Some(CsvParseOptions::default().with_has_header(false)),
             None,
             io_client,
@@ -1123,7 +1272,9 @@ mod tests {
             file.as_ref(),
             Some(
                 CsvConvertOptions::default()
-                    .with_column_names(Some(column_names.iter().map(|s| s.to_string()).collect()))
+                    .with_column_names(Some(
+                        column_names.iter().map(|s| (*s).to_string()).collect(),
+                    ))
                     .with_include_columns(Some(vec![
                         "petal.length".to_string(),
                         "petal.width".to_string(),
@@ -1749,7 +1900,7 @@ mod tests {
     ) -> DaftResult<()> {
         let file = format!(
             "s3://daft-public-data/test_fixtures/csv-dev/mvp.csv{}",
-            compression.map_or("".to_string(), |ext| format!(".{}", ext))
+            compression.map_or(String::new(), |ext| format!(".{ext}"))
         );
 
         let mut io_config = IOConfig::default();
@@ -1783,10 +1934,9 @@ mod tests {
         let column_names = ["a", "b"];
         let table = read_csv(
             file,
-            Some(
-                CsvConvertOptions::default()
-                    .with_column_names(Some(column_names.iter().map(|s| s.to_string()).collect())),
-            ),
+            Some(CsvConvertOptions::default().with_column_names(Some(
+                column_names.iter().map(|s| (*s).to_string()).collect(),
+            ))),
             Some(CsvParseOptions::default().with_has_header(false)),
             None,
             io_client,
@@ -1821,7 +1971,9 @@ mod tests {
             file,
             Some(
                 CsvConvertOptions::default()
-                    .with_column_names(Some(column_names.iter().map(|s| s.to_string()).collect()))
+                    .with_column_names(Some(
+                        column_names.iter().map(|s| (*s).to_string()).collect(),
+                    ))
                     .with_include_columns(Some(vec!["b".to_string()])),
             ),
             Some(CsvParseOptions::default().with_has_header(false)),
