@@ -1,29 +1,34 @@
-use std::io::{SeekFrom, Write};
-use std::ops::Range;
-use std::path::PathBuf;
+use std::{
+    io::{SeekFrom, Write},
+    ops::Range,
+    path::PathBuf,
+    sync::Arc,
+};
 
-use crate::object_io::{self, FileMetadata, LSResult};
-use crate::stats::IOStatsRef;
-
-use super::object_io::{GetResult, ObjectSource};
-use super::Result;
 use async_trait::async_trait;
 use bytes::Bytes;
 use common_error::DaftError;
-use futures::stream::BoxStream;
-use futures::StreamExt;
-use futures::TryStreamExt;
+use futures::{stream::BoxStream, StreamExt, TryStreamExt};
 use snafu::{ResultExt, Snafu};
-use std::sync::Arc;
 use tokio::io::{AsyncReadExt, AsyncSeekExt};
 
+use super::{
+    object_io::{GetResult, ObjectSource},
+    Result,
+};
+use crate::{
+    object_io::{self, FileMetadata, LSResult},
+    stats::IOStatsRef,
+    FileFormat,
+};
+
 /// NOTE: We hardcode this even for Windows
 ///
 /// For the most part, Windows machines work quite well with POSIX-style paths
 /// as long as there is no "mix" of "\" and "/".
 const PATH_SEGMENT_DELIMITER: &str = "/";
 
-pub(crate) struct LocalSource {}
+pub struct LocalSource {}
 
 #[derive(Debug, Snafu)]
 enum Error {
@@ -77,39 +82,42 @@ enum Error {
 
 impl From<Error> for super::Error {
     fn from(error: Error) -> Self {
-        use Error::*;
+        use Error::{
+            UnableToFetchDirectoryEntries, UnableToFetchFileMetadata, UnableToOpenFile,
+            UnableToOpenFileForWriting, UnableToReadBytes, UnableToWriteToFile,
+        };
         match error {
             UnableToOpenFile { path, source } | UnableToFetchDirectoryEntries { path, source } => {
-                use std::io::ErrorKind::*;
+                use std::io::ErrorKind::NotFound;
                 match source.kind() {
-                    NotFound => super::Error::NotFound {
+                    NotFound => Self::NotFound {
                         path,
                         source: source.into(),
                     },
-                    _ => super::Error::UnableToOpenFile {
+                    _ => Self::UnableToOpenFile {
                         path,
                         source: source.into(),
                     },
                 }
             }
             UnableToFetchFileMetadata { path, source } => {
-                use std::io::ErrorKind::*;
+                use std::io::ErrorKind::{IsADirectory, NotFound};
                 match source.kind() {
-                    NotFound | IsADirectory => super::Error::NotFound {
+                    NotFound | IsADirectory => Self::NotFound {
                         path,
                         source: source.into(),
                     },
-                    _ => super::Error::UnableToOpenFile {
+                    _ => Self::UnableToOpenFile {
                         path,
                         source: source.into(),
                     },
                 }
             }
-            UnableToReadBytes { path, source } => super::Error::UnableToReadBytes { path, source },
+            UnableToReadBytes { path, source } => Self::UnableToReadBytes { path, source },
             UnableToWriteToFile { path, source } | UnableToOpenFileForWriting { path, source } => {
-                super::Error::UnableToWriteToFile { path, source }
+                Self::UnableToWriteToFile { path, source }
             }
-            _ => super::Error::Generic {
+            _ => Self::Generic {
                 store: super::SourceType::File,
                 source: error.into(),
             },
@@ -119,7 +127,7 @@ impl From<Error> for super::Error {
 
 impl LocalSource {
     pub async fn get_client() -> super::Result<Arc<Self>> {
-        Ok(LocalSource {}.into())
+        Ok(Self {}.into())
     }
 }
 
@@ -196,6 +204,7 @@ impl ObjectSource for LocalSource {
         _page_size: Option<i32>,
         limit: Option<usize>,
         io_stats: Option<IOStatsRef>,
+        _file_format: Option<FileFormat>,
     ) -> super::Result<BoxStream<'static, super::Result<FileMetadata>>> {
         use crate::object_store_glob::glob;
 
@@ -271,7 +280,7 @@ impl ObjectSource for LocalSource {
         if meta.file_type().is_file() {
             // Provided uri points to a file, so only return that file.
             return Ok(futures::stream::iter([Ok(FileMetadata {
-                filepath: format!("{}{}", LOCAL_PROTOCOL, uri),
+                filepath: format!("{LOCAL_PROTOCOL}{uri}"),
                 size: Some(meta.len()),
                 filetype: object_io::FileType::File,
             })])
@@ -328,7 +337,7 @@ impl ObjectSource for LocalSource {
     }
 }
 
-pub(crate) async fn collect_file(local_file: LocalFile) -> Result<Bytes> {
+pub async fn collect_file(local_file: LocalFile) -> Result<Bytes> {
     let path = &local_file.path;
     let mut file = tokio::fs::File::open(path)
         .await
@@ -367,14 +376,13 @@ pub(crate) async fn collect_file(local_file: LocalFile) -> Result<Bytes> {
 }
 
 #[cfg(test)]
-
 mod tests {
-    use std::default;
-    use std::io::Write;
+    use std::{default, io::Write};
 
-    use crate::object_io::{FileMetadata, FileType, ObjectSource};
-    use crate::Result;
-    use crate::{HttpSource, LocalSource};
+    use crate::{
+        object_io::{FileMetadata, FileType, ObjectSource},
+        HttpSource, LocalSource, Result,
+    };
 
     async fn write_remote_parquet_to_local_file(
         f: &mut tempfile::NamedTempFile,
