@@ -1,11 +1,13 @@
 use std::{collections::HashMap, num::NonZeroUsize, sync::Arc};
 
 use common_error::{DaftError, DaftResult};
-use daft_core::{schema::Schema, utils::arrow::cast_array_for_daft_if_needed, Series};
+use common_runtime::get_io_runtime;
+use daft_compression::CompressionCodec;
+use daft_core::{prelude::*, utils::arrow::cast_array_for_daft_if_needed};
 use daft_dsl::optimization::get_required_columns;
-use daft_io::{get_runtime, parse_url, GetResult, IOClient, IOStatsRef, SourceType};
+use daft_io::{parse_url, GetResult, IOClient, IOStatsRef, SourceType};
 use daft_table::Table;
-use futures::{Stream, StreamExt, TryStreamExt};
+use futures::{stream::BoxStream, Stream, StreamExt, TryStreamExt};
 use rayon::prelude::{IntoParallelIterator, ParallelIterator};
 use snafu::{
     futures::{try_future::Context, TryFutureExt, TryStreamExt as _},
@@ -18,11 +20,10 @@ use tokio::{
 };
 use tokio_util::io::StreamReader;
 
-use crate::{decoding::deserialize_records, local::read_json_local, ArrowSnafu, ChunkSnafu};
 use crate::{
-    schema::read_json_schema_single, JsonConvertOptions, JsonParseOptions, JsonReadOptions,
+    decoding::deserialize_records, local::read_json_local, schema::read_json_schema_single,
+    ArrowSnafu, ChunkSnafu, JsonConvertOptions, JsonParseOptions, JsonReadOptions,
 };
-use daft_compression::CompressionCodec;
 
 type TableChunkResult =
     super::Result<Context<JoinHandle<DaftResult<Table>>, super::JoinSnafu, super::Error>>;
@@ -46,9 +47,8 @@ pub fn read_json(
     multithreaded_io: bool,
     max_chunks_in_flight: Option<usize>,
 ) -> DaftResult<Table> {
-    let runtime_handle = get_runtime(multithreaded_io)?;
-    let _rt_guard = runtime_handle.enter();
-    runtime_handle.block_on(async {
+    let runtime_handle = get_io_runtime(multithreaded_io);
+    runtime_handle.block_on_current_thread(async {
         read_json_single_into_table(
             uri,
             convert_options,
@@ -74,13 +74,12 @@ pub fn read_json_bulk(
     max_chunks_in_flight: Option<usize>,
     num_parallel_tasks: usize,
 ) -> DaftResult<Vec<Table>> {
-    let runtime_handle = get_runtime(multithreaded_io)?;
-    let _rt_guard = runtime_handle.enter();
-    let tables = runtime_handle.block_on(async move {
+    let runtime_handle = get_io_runtime(multithreaded_io);
+    let tables = runtime_handle.block_on_current_thread(async move {
         // Launch a read task per URI, throttling the number of concurrent file reads to num_parallel tasks.
         let task_stream = futures::stream::iter(uris.iter().map(|uri| {
             let (uri, convert_options, parse_options, read_options, io_client, io_stats) = (
-                uri.to_string(),
+                (*uri).to_string(),
                 convert_options.clone(),
                 parse_options.clone(),
                 read_options.clone(),
@@ -166,7 +165,7 @@ pub(crate) fn tables_concat(mut tables: Vec<Table>) -> DaftResult<Table> {
     Table::new_with_size(
         first_table.schema.clone(),
         new_series,
-        tables.iter().map(|t| t.len()).sum(),
+        tables.iter().map(daft_table::Table::len).sum(),
     )
 }
 
@@ -207,7 +206,7 @@ async fn read_json_single_into_table(
                 let required_columns_for_predicate = get_required_columns(predicate);
                 for rc in required_columns_for_predicate {
                     if include_columns.iter().all(|c| c.as_str() != rc.as_str()) {
-                        include_columns.push(rc)
+                        include_columns.push(rc);
                     }
                 }
             }
@@ -288,6 +287,97 @@ async fn read_json_single_into_table(
     }
 }
 
+pub async fn stream_json(
+    uri: String,
+    convert_options: Option<JsonConvertOptions>,
+    parse_options: Option<JsonParseOptions>,
+    read_options: Option<JsonReadOptions>,
+    io_client: Arc<IOClient>,
+    io_stats: Option<IOStatsRef>,
+    max_chunks_in_flight: Option<usize>,
+) -> DaftResult<BoxStream<'static, DaftResult<Table>>> {
+    let predicate = convert_options
+        .as_ref()
+        .and_then(|opts| opts.predicate.clone());
+
+    let limit = convert_options.as_ref().and_then(|opts| opts.limit);
+    let include_columns = convert_options
+        .as_ref()
+        .and_then(|opts| opts.include_columns.clone());
+
+    let convert_options_with_predicate_columns = match (convert_options, &predicate) {
+        (None, _) => None,
+        (co, None) => co,
+        (Some(mut co), Some(predicate)) => {
+            if let Some(ref mut include_columns) = co.include_columns {
+                let required_columns_for_predicate = get_required_columns(predicate);
+                for rc in required_columns_for_predicate {
+                    if include_columns.iter().all(|c| c.as_str() != rc.as_str()) {
+                        include_columns.push(rc);
+                    }
+                }
+            }
+            // if we have a limit and a predicate, remove limit for stream
+            co.limit = None;
+            Some(co)
+        }
+    };
+
+    let (chunk_stream, _fields) = read_json_single_into_stream(
+        &uri,
+        convert_options_with_predicate_columns.unwrap_or_default(),
+        parse_options.unwrap_or_default(),
+        read_options,
+        io_client,
+        io_stats,
+    )
+    .await?;
+
+    // Default max chunks in flight is set to 2x the number of cores, which should ensure pipelining of reading chunks
+    // with the parsing of chunks on the rayon threadpool.
+    let max_chunks_in_flight = max_chunks_in_flight.unwrap_or_else(|| {
+        std::thread::available_parallelism()
+            .unwrap_or(NonZeroUsize::new(2).unwrap())
+            .checked_mul(2.try_into().unwrap())
+            .unwrap()
+            .into()
+    });
+    // Collect all chunks in chunk x column form.
+    let tables = chunk_stream
+        // Limit the number of chunks we have in flight at any given time.
+        .try_buffered(max_chunks_in_flight);
+
+    let filtered_tables = tables.map(move |table| {
+        let table = table?;
+        if let Some(predicate) = &predicate {
+            let filtered = table?.filter(&[predicate.clone()])?;
+            if let Some(include_columns) = &include_columns {
+                filtered.get_columns(include_columns.as_slice())
+            } else {
+                Ok(filtered)
+            }
+        } else {
+            table
+        }
+    });
+
+    let mut remaining_rows = limit.map(|limit| limit as i64);
+    let tables = filtered_tables.try_take_while(move |table| {
+        match remaining_rows {
+            // Limit has been met, early-terminate.
+            Some(rows_left) if rows_left <= 0 => futures::future::ready(Ok(false)),
+            // Limit has not yet been met, update remaining limit slack and continue.
+            Some(rows_left) => {
+                remaining_rows = Some(rows_left - table.len() as i64);
+                futures::future::ready(Ok(true))
+            }
+            // No limit, never early-terminate.
+            None => futures::future::ready(Ok(true)),
+        }
+    });
+    Ok(Box::pin(tables))
+}
+
 async fn read_json_single_into_stream(
     uri: &str,
     convert_options: JsonConvertOptions,
@@ -414,7 +504,7 @@ fn parse_into_column_array_chunk_stream(
     schema: Arc<arrow2::datatypes::Schema>,
     schema_is_projection: bool,
 ) -> DaftResult<impl TableChunkStream + Send> {
-    let daft_schema = Arc::new(daft_core::schema::Schema::try_from(schema.as_ref())?);
+    let daft_schema = Arc::new(daft_core::prelude::Schema::try_from(schema.as_ref())?);
     let daft_fields = Arc::new(
         daft_schema
             .fields
@@ -474,25 +564,21 @@ mod tests {
     use std::{collections::HashSet, io::BufRead, sync::Arc};
 
     use common_error::DaftResult;
-
     use daft_core::{
-        datatypes::{Field, TimeUnit},
-        schema::Schema,
+        prelude::*,
         utils::arrow::{cast_array_for_daft_if_needed, cast_array_from_daft_if_needed},
-        DataType,
     };
     use daft_io::{IOClient, IOConfig};
     use daft_table::Table;
     use indexmap::IndexMap;
     use rstest::rstest;
 
+    use super::read_json;
     use crate::{
         decoding::deserialize_records,
         inference::{column_types_map_to_fields, infer_records_schema},
+        JsonConvertOptions, JsonReadOptions,
     };
-    use crate::{JsonConvertOptions, JsonReadOptions};
-
-    use super::read_json;
 
     fn check_equal_local_arrow2(
         path: &str,
@@ -510,7 +596,7 @@ mod tests {
         // Get consolidated schema from parsed JSON.
         let mut column_types: IndexMap<String, HashSet<arrow2::datatypes::DataType>> =
             IndexMap::new();
-        parsed.iter().for_each(|record| {
+        for record in &parsed {
             let schema = infer_records_schema(record).unwrap();
             for field in schema.fields {
                 match column_types.entry(field.name) {
@@ -524,7 +610,7 @@ mod tests {
                     }
                 }
             }
-        });
+        }
         let fields = column_types_map_to_fields(column_types);
         let schema: arrow2::datatypes::Schema = fields.into();
         // Apply projection to schema.
@@ -588,7 +674,7 @@ mod tests {
         let file = format!(
             "{}/test/iris_tiny.jsonl{}",
             env!("CARGO_MANIFEST_DIR"),
-            compression.map_or("".to_string(), |ext| format!(".{}", ext))
+            compression.map_or(String::new(), |ext| format!(".{}", ext))
         );
 
         let mut io_config = IOConfig::default();
@@ -1108,7 +1194,7 @@ mod tests {
     ) -> DaftResult<()> {
         let file = format!(
             "s3://daft-public-data/test_fixtures/json-dev/iris_tiny.jsonl{}",
-            compression.map_or("".to_string(), |ext| format!(".{}", ext))
+            compression.map_or(String::new(), |ext| format!(".{}", ext))
         );
 
         let mut io_config = IOConfig::default();
