@@ -1,36 +1,40 @@
 use std::{
+    cmp::{max, min},
     collections::{BTreeMap, HashSet},
     sync::Arc,
 };
 
-use arrow2::io::parquet::read::schema::infer_schema_with_options;
+use arrow2::io::parquet::read::{column_iter_to_arrays, schema::infer_schema_with_options};
 use common_error::DaftResult;
-use daft_core::{
-    datatypes::Field, schema::Schema, utils::arrow::cast_array_for_daft_if_needed, Series,
-};
+use common_runtime::get_compute_runtime;
+use daft_core::{prelude::*, utils::arrow::cast_array_for_daft_if_needed};
 use daft_dsl::ExprRef;
 use daft_io::{IOClient, IOStatsRef};
 use daft_stats::TruthValue;
 use daft_table::Table;
-use futures::{future::try_join_all, StreamExt};
+use futures::{future::try_join_all, stream::BoxStream, StreamExt};
 use parquet2::{
     page::{CompressedPage, Page},
     read::get_owned_page_stream_from_column_start,
     FallibleStreamingIterator,
 };
 use snafu::ResultExt;
+use tokio_stream::wrappers::ReceiverStream;
 
 use crate::{
+    determine_parquet_parallelism,
     metadata::read_parquet_metadata,
     read::ParquetSchemaInferenceOptions,
     read_planner::{CoalescePass, RangesContainer, ReadPlanner, SplitLargeRequestPass},
-    statistics, JoinSnafu, OneShotRecvSnafu, UnableToConvertRowGroupMetadataToStatsSnafu,
+    statistics,
+    stream_reader::spawn_column_iters_to_table_task,
+    utils::combine_stream,
+    JoinSnafu, OneShotRecvSnafu, UnableToConvertRowGroupMetadataToStatsSnafu,
     UnableToConvertSchemaToDaftSnafu, UnableToCreateParquetPageStreamSnafu,
-    UnableToParseSchemaFromMetadataSnafu, UnableToRunExpressionOnStatsSnafu,
+    UnableToParseSchemaFromMetadataSnafu, UnableToRunExpressionOnStatsSnafu, PARQUET_MORSEL_SIZE,
 };
-use arrow2::io::parquet::read::column_iter_to_arrays;
 
-pub(crate) struct ParquetReaderBuilder {
+pub struct ParquetReaderBuilder {
     pub uri: String,
     pub metadata: parquet2::metadata::FileMetaData,
     selected_columns: Option<HashSet<String>>,
@@ -39,6 +43,7 @@ pub(crate) struct ParquetReaderBuilder {
     row_groups: Option<Vec<i64>>,
     schema_inference_options: ParquetSchemaInferenceOptions,
     predicate: Option<ExprRef>,
+    chunk_size: Option<usize>,
 }
 use parquet2::read::decompress;
 
@@ -67,7 +72,7 @@ where
     S: futures::Stream<Item = parquet2::error::Result<Page>> + std::marker::Unpin,
 {
     pub fn new(src: S, handle: tokio::runtime::Handle) -> Self {
-        StreamIterator {
+        Self {
             curr: None,
             src: tokio::sync::Mutex::new(src),
             handle,
@@ -82,9 +87,11 @@ where
     type Error = parquet2::error::Error;
     type Item = Page;
     fn advance(&mut self) -> Result<(), Self::Error> {
-        let val = self.handle.block_on(async {
-            let mut s_guard = self.src.lock().await;
-            s_guard.next().await
+        let val = tokio::task::block_in_place(|| {
+            self.handle.block_on(async {
+                let mut s_guard = self.src.lock().await;
+                s_guard.next().await
+            })
         });
         if let Some(val) = val {
             let val = val?;
@@ -100,7 +107,7 @@ where
     }
 }
 
-pub(crate) fn build_row_ranges(
+pub fn build_row_ranges(
     limit: Option<usize>,
     row_start_offset: usize,
     row_groups: Option<&[i64]>,
@@ -117,7 +124,7 @@ pub(crate) fn build_row_ranges(
         let mut rows_to_add: i64 = limit.unwrap_or(i64::MAX);
         for i in row_groups {
             let i = *i as usize;
-            if !(0..metadata.row_groups.len()).contains(&i) {
+            if !metadata.row_groups.keys().any(|x| *x == i) {
                 return Err(super::Error::ParquetRowGroupOutOfIndex {
                     path: uri.to_string(),
                     row_group: i as i64,
@@ -127,7 +134,7 @@ pub(crate) fn build_row_ranges(
             if rows_to_add <= 0 {
                 break;
             }
-            let rg = metadata.row_groups.get(i).unwrap();
+            let rg = metadata.row_groups.get(&i).unwrap();
             if let Some(ref pred) = predicate {
                 let stats = statistics::row_group_metadata_to_table_stats(rg, schema)
                     .with_context(|_| UnableToConvertRowGroupMetadataToStatsSnafu {
@@ -155,7 +162,7 @@ pub(crate) fn build_row_ranges(
     } else {
         let mut rows_to_add = limit.unwrap_or(metadata.num_rows as i64);
 
-        for (i, rg) in metadata.row_groups.iter().enumerate() {
+        for (i, rg) in &metadata.row_groups {
             if (curr_row_index + rg.num_rows()) < row_start_offset {
                 curr_row_index += rg.num_rows();
                 continue;
@@ -176,7 +183,7 @@ pub(crate) fn build_row_ranges(
                     }
                 }
                 let range_to_add = RowGroupRange {
-                    row_group_index: i,
+                    row_group_index: *i,
                     start: row_start_offset.saturating_sub(curr_row_index),
                     num_rows: rg.num_rows().min(rows_to_add as usize),
                 };
@@ -204,7 +211,7 @@ impl ParquetReaderBuilder {
             .await?;
         let metadata =
             read_parquet_metadata(uri, size, io_client, io_stats, field_id_mapping).await?;
-        Ok(ParquetReaderBuilder {
+        Ok(Self {
             uri: uri.into(),
             metadata,
             selected_columns: None,
@@ -213,6 +220,7 @@ impl ParquetReaderBuilder {
             row_groups: None,
             schema_inference_options: Default::default(),
             predicate: None,
+            chunk_size: None,
         })
     }
 
@@ -220,8 +228,8 @@ impl ParquetReaderBuilder {
         &self.metadata
     }
 
-    pub fn prune_columns<S: ToString + AsRef<str>>(mut self, columns: &[S]) -> super::Result<Self> {
-        self.selected_columns = Some(HashSet::from_iter(columns.iter().map(|s| s.to_string())));
+    pub fn prune_columns(mut self, columns: &[String]) -> super::Result<Self> {
+        self.selected_columns = Some(HashSet::from_iter(columns.iter().cloned()));
         Ok(self)
     }
 
@@ -252,12 +260,18 @@ impl ParquetReaderBuilder {
         self
     }
 
+    pub fn set_chunk_size(mut self, chunk_size: Option<usize>) -> Self {
+        self.chunk_size = chunk_size;
+        self
+    }
+
     pub fn build(self) -> super::Result<ParquetFileReader> {
-        let mut arrow_schema =
-            infer_schema_with_options(&self.metadata, &Some(self.schema_inference_options.into()))
-                .context(UnableToParseSchemaFromMetadataSnafu::<String> {
-                    path: self.uri.clone(),
-                })?;
+        let options = self.schema_inference_options.into();
+        let mut arrow_schema = infer_schema_with_options(&self.metadata, Some(options)).context(
+            UnableToParseSchemaFromMetadataSnafu {
+                path: self.uri.clone(),
+            },
+        )?;
 
         if let Some(names_to_keep) = self.selected_columns {
             arrow_schema
@@ -279,36 +293,50 @@ impl ParquetReaderBuilder {
             &self.uri,
         )?;
 
-        ParquetFileReader::new(self.uri, self.metadata, arrow_schema, row_ranges)
+        ParquetFileReader::new(
+            self.uri,
+            self.metadata,
+            arrow_schema,
+            row_ranges,
+            self.chunk_size,
+        )
     }
 }
 
 #[derive(Copy, Clone)]
-pub(crate) struct RowGroupRange {
+pub struct RowGroupRange {
     pub row_group_index: usize,
     pub start: usize,
     pub num_rows: usize,
 }
 
-pub(crate) struct ParquetFileReader {
+pub struct ParquetFileReader {
     uri: String,
     metadata: Arc<parquet2::metadata::FileMetaData>,
     arrow_schema: arrow2::datatypes::SchemaRef,
     row_ranges: Arc<Vec<RowGroupRange>>,
+    chunk_size: Option<usize>,
 }
 
 impl ParquetFileReader {
+    const DEFAULT_CHUNK_SIZE: usize = 2048;
+    // Set to 2GB because that's the maximum size of strings allowable by Parquet (using i32 offsets).
+    // See issue: https://github.com/Eventual-Inc/Daft/issues/3007
+    const MAX_PAGE_SIZE: usize = 2 * 1024 * 1024 * 1024;
+
     fn new(
         uri: String,
         metadata: parquet2::metadata::FileMetaData,
         arrow_schema: arrow2::datatypes::Schema,
         row_ranges: Vec<RowGroupRange>,
+        chunk_size: Option<usize>,
     ) -> super::Result<Self> {
-        Ok(ParquetFileReader {
+        Ok(Self {
             uri,
             metadata: Arc::new(metadata),
             arrow_schema: arrow_schema.into(),
             row_ranges: Arc::new(row_ranges),
+            chunk_size,
         })
     }
 
@@ -325,11 +353,11 @@ impl ParquetFileReader {
             let rg = self
                 .metadata
                 .row_groups
-                .get(row_group_range.row_group_index)
+                .get(&row_group_range.row_group_index)
                 .unwrap();
 
             let columns = rg.columns();
-            for field in arrow_fields.iter() {
+            for field in arrow_fields {
                 let field_name = field.name.clone();
                 let filtered_cols = columns
                     .iter()
@@ -369,6 +397,162 @@ impl ParquetFileReader {
         read_planner.collect(io_client, io_stats)
     }
 
+    pub async fn read_from_ranges_into_table_stream(
+        self,
+        ranges: Arc<RangesContainer>,
+        maintain_order: bool,
+        predicate: Option<ExprRef>,
+        original_columns: Option<Vec<String>>,
+        original_num_rows: Option<usize>,
+        delete_rows: Option<Vec<i64>>,
+    ) -> DaftResult<BoxStream<'static, DaftResult<Table>>> {
+        let daft_schema = Arc::new(daft_core::prelude::Schema::try_from(
+            self.arrow_schema.as_ref(),
+        )?);
+
+        let num_parallel_tasks = determine_parquet_parallelism(&daft_schema);
+        let semaphore = Arc::new(tokio::sync::Semaphore::new(num_parallel_tasks));
+
+        let (senders, receivers): (Vec<_>, Vec<_>) = self
+            .row_ranges
+            .iter()
+            .map(|_| tokio::sync::mpsc::channel(1))
+            .unzip();
+
+        let uri = self.uri.clone();
+        let chunk_size = self.chunk_size.unwrap_or(Self::DEFAULT_CHUNK_SIZE);
+        let chunk_iter_handles = <Vec<RowGroupRange> as Clone>::clone(&self.row_ranges)
+            .into_iter()
+            .map(move |row_range| {
+                let metadata = self.metadata.clone();
+                let arrow_schema = self.arrow_schema.clone();
+                let ranges = ranges.clone();
+                let uri = uri.clone();
+
+                let chunk_iter_task = tokio::task::spawn(async move {
+                    let arr_iter_handles = arrow_schema.fields.iter().map(|field| {
+                        let rt_handle = tokio::runtime::Handle::current();
+                        let ranges = ranges.clone();
+                        let uri = uri.clone();
+                        let field = field.clone();
+                        let metadata = metadata.clone();
+
+                        tokio::task::spawn(async move {
+                            let rg = metadata
+                                .row_groups
+                                .get(&row_range.row_group_index)
+                                .expect("Row Group index should be in bounds");
+                            let num_rows = rg.num_rows().min(row_range.start + row_range.num_rows);
+                            let filtered_columns = rg
+                                .columns()
+                                .iter()
+                                .filter(|x| x.descriptor().path_in_schema[0] == field.name)
+                                .collect::<Vec<_>>();
+                            let mut decompressed_iters = Vec::with_capacity(filtered_columns.len());
+                            let mut ptypes = Vec::with_capacity(filtered_columns.len());
+                            let mut num_values = Vec::with_capacity(filtered_columns.len());
+                            for col in filtered_columns {
+                                num_values.push(col.metadata().num_values as usize);
+                                ptypes.push(col.descriptor().descriptor.primitive_type.clone());
+
+                                let byte_range = {
+                                    let (start, len) = col.byte_range();
+                                    let end: u64 = start + len;
+                                    start as usize..end as usize
+                                };
+                                let range_reader =
+                                    Box::pin(ranges.get_range_reader(byte_range).await?);
+                                let compressed_page_stream =
+                                    get_owned_page_stream_from_column_start(
+                                        col,
+                                        range_reader,
+                                        vec![],
+                                        Arc::new(|_, _| true),
+                                        Self::MAX_PAGE_SIZE,
+                                    )
+                                    .with_context(|_| {
+                                        UnableToCreateParquetPageStreamSnafu::<String> {
+                                            path: uri.clone(),
+                                        }
+                                    })?;
+                                let page_stream = streaming_decompression(compressed_page_stream);
+                                let pinned_stream = Box::pin(page_stream);
+                                decompressed_iters
+                                    .push(StreamIterator::new(pinned_stream, rt_handle.clone()));
+                            }
+                            let arr_iter = column_iter_to_arrays(
+                                decompressed_iters,
+                                ptypes.iter().collect(),
+                                field,
+                                Some(chunk_size),
+                                num_rows,
+                                num_values,
+                            )?;
+                            Ok(arr_iter)
+                        })
+                    });
+
+                    let arr_iters = try_join_all(arr_iter_handles)
+                        .await
+                        .context(JoinSnafu { path: uri.clone() })?
+                        .into_iter()
+                        .collect::<DaftResult<Vec<_>>>()?;
+
+                    DaftResult::Ok(arr_iters)
+                });
+                (row_range, chunk_iter_task)
+            });
+
+        let compute_runtime = get_compute_runtime();
+        let uri = self.uri.clone();
+        let parquet_task = compute_runtime.spawn(async move {
+            let mut table_tasks = Vec::with_capacity(chunk_iter_handles.len());
+            for ((row_range, chunk_iter_handle), output_sender) in chunk_iter_handles.zip(senders) {
+                // We want to ensure that the channel capacity can hold one morsel worth of data for better deserialization performance.
+                let channel_size = {
+                    let chunks_per_morsel = max(PARQUET_MORSEL_SIZE / chunk_size, 1);
+                    let max_num_chunks = max(row_range.num_rows / chunk_size, 1);
+                    min(chunks_per_morsel, max_num_chunks)
+                };
+
+                let chunk_iter = chunk_iter_handle
+                    .await
+                    .context(JoinSnafu { path: uri.clone() })??;
+
+                let permit = semaphore.clone().acquire_owned().await.unwrap();
+                let table_task = spawn_column_iters_to_table_task(
+                    chunk_iter,
+                    row_range,
+                    daft_schema.clone(),
+                    uri.clone(),
+                    predicate.clone(),
+                    original_columns.clone(),
+                    original_num_rows,
+                    delete_rows.clone(),
+                    output_sender,
+                    permit,
+                    channel_size,
+                );
+                table_tasks.push(table_task);
+            }
+
+            try_join_all(table_tasks)
+                .await?
+                .into_iter()
+                .collect::<DaftResult<()>>()?;
+            DaftResult::Ok(())
+        });
+
+        let stream_of_streams =
+            futures::stream::iter(receivers.into_iter().map(ReceiverStream::new));
+        match maintain_order {
+            true => Ok(combine_stream(stream_of_streams.flatten(), parquet_task).boxed()),
+            false => {
+                Ok(combine_stream(stream_of_streams.flatten_unordered(None), parquet_task).boxed())
+            }
+        }
+    }
+
     pub async fn read_from_ranges_into_table(
         self,
         ranges: Arc<RangesContainer>,
@@ -390,9 +574,10 @@ impl ParquetFileReader {
                         let owned_uri = self.uri.clone();
                         let rg = metadata
                             .row_groups
-                            .get(row_range.row_group_index)
+                            .get(&row_range.row_group_index)
                             .expect("Row Group index should be in bounds");
                         let num_rows = rg.num_rows().min(row_range.start + row_range.num_rows);
+                        let chunk_size = self.chunk_size.unwrap_or(Self::DEFAULT_CHUNK_SIZE);
                         let columns = rg.columns();
                         let field_name = &field.name;
                         let filtered_cols_idx = columns
@@ -419,26 +604,27 @@ impl ParquetFileReader {
                         let handle = tokio::task::spawn(async move {
                             let mut range_readers = Vec::with_capacity(filtered_cols_idx.len());
 
-                            for range in needed_byte_ranges.into_iter() {
+                            for range in needed_byte_ranges {
                                 let range_reader = ranges.get_range_reader(range).await?;
-                                range_readers.push(Box::pin(range_reader))
+                                range_readers.push(Box::pin(range_reader));
                             }
 
                             let mut decompressed_iters =
                                 Vec::with_capacity(filtered_cols_idx.len());
                             let mut ptypes = Vec::with_capacity(filtered_cols_idx.len());
-
+                            let mut num_values = Vec::with_capacity(filtered_cols_idx.len());
                             for (col_idx, range_reader) in
                                 filtered_cols_idx.into_iter().zip(range_readers)
                             {
                                 let col = metadata
                                     .row_groups
-                                    .get(row_range.row_group_index)
+                                    .get(&row_range.row_group_index)
                                     .expect("Row Group index should be in bounds")
                                     .columns()
                                     .get(col_idx)
                                     .expect("Column index should be in bounds");
                                 ptypes.push(col.descriptor().descriptor.primitive_type.clone());
+                                num_values.push(col.metadata().num_values as usize);
 
                                 let compressed_page_stream =
                                     get_owned_page_stream_from_column_start(
@@ -446,12 +632,8 @@ impl ParquetFileReader {
                                         range_reader,
                                         vec![],
                                         Arc::new(|_, _| true),
-                                        // Set to a very high number 256MB to guard against unbounded large
-                                        // downloads from remote storage, which likely indicates corrupted Parquet data
-                                        // See: https://github.com/Eventual-Inc/Daft/issues/1551
-                                        256 * 1024 * 1024,
+                                        Self::MAX_PAGE_SIZE,
                                     )
-                                    .await
                                     .with_context(|_| {
                                         UnableToCreateParquetPageStreamSnafu::<String> {
                                             path: owned_uri.clone(),
@@ -460,7 +642,7 @@ impl ParquetFileReader {
                                 let page_stream = streaming_decompression(compressed_page_stream);
                                 let pinned_stream = Box::pin(page_stream);
                                 decompressed_iters
-                                    .push(StreamIterator::new(pinned_stream, rt_handle.clone()))
+                                    .push(StreamIterator::new(pinned_stream, rt_handle.clone()));
                             }
 
                             let (send, recv) = tokio::sync::oneshot::channel();
@@ -469,8 +651,9 @@ impl ParquetFileReader {
                                     decompressed_iters,
                                     ptypes.iter().collect(),
                                     field.clone(),
-                                    Some(2048),
+                                    Some(chunk_size),
                                     num_rows,
+                                    num_values,
                                 );
 
                                 let series = (|| {
@@ -547,10 +730,9 @@ impl ParquetFileReader {
             })?
             .into_iter()
             .collect::<DaftResult<Vec<_>>>()?;
-        let daft_schema = daft_core::schema::Schema::try_from(self.arrow_schema.as_ref())?;
 
         Table::new_with_size(
-            daft_schema,
+            Schema::new(all_series.iter().map(|s| s.field().clone()).collect())?,
             all_series,
             self.row_ranges.as_ref().iter().map(|rr| rr.num_rows).sum(),
         )
@@ -577,9 +759,10 @@ impl ParquetFileReader {
                         let owned_uri = self.uri.clone();
                         let rg = metadata
                             .row_groups
-                            .get(row_range.row_group_index)
+                            .get(&row_range.row_group_index)
                             .expect("Row Group index should be in bounds");
                         let num_rows = rg.num_rows().min(row_range.start + row_range.num_rows);
+                        let chunk_size = self.chunk_size.unwrap_or(PARQUET_MORSEL_SIZE);
                         let columns = rg.columns();
                         let field_name = &field.name;
                         let filtered_cols_idx = columns
@@ -603,26 +786,28 @@ impl ParquetFileReader {
                         let handle = tokio::task::spawn(async move {
                             let mut range_readers = Vec::with_capacity(filtered_cols_idx.len());
 
-                            for range in needed_byte_ranges.into_iter() {
+                            for range in needed_byte_ranges {
                                 let range_reader = ranges.get_range_reader(range).await?;
-                                range_readers.push(Box::pin(range_reader))
+                                range_readers.push(Box::pin(range_reader));
                             }
 
                             let mut decompressed_iters =
                                 Vec::with_capacity(filtered_cols_idx.len());
                             let mut ptypes = Vec::with_capacity(filtered_cols_idx.len());
+                            let mut num_values = Vec::with_capacity(filtered_cols_idx.len());
 
                             for (col_idx, range_reader) in
                                 filtered_cols_idx.into_iter().zip(range_readers)
                             {
                                 let col = metadata
                                     .row_groups
-                                    .get(row_range.row_group_index)
+                                    .get(&row_range.row_group_index)
                                     .expect("Row Group index should be in bounds")
                                     .columns()
                                     .get(col_idx)
                                     .expect("Column index should be in bounds");
                                 ptypes.push(col.descriptor().descriptor.primitive_type.clone());
+                                num_values.push(col.metadata().num_values as usize);
 
                                 let compressed_page_stream =
                                     get_owned_page_stream_from_column_start(
@@ -630,12 +815,8 @@ impl ParquetFileReader {
                                         range_reader,
                                         vec![],
                                         Arc::new(|_, _| true),
-                                        // Set to a very high number 256MB to guard against unbounded large
-                                        // downloads from remote storage, which likely indicates corrupted Parquet data
-                                        // See: https://github.com/Eventual-Inc/Daft/issues/1551
-                                        256 * 1024 * 1024,
+                                        Self::MAX_PAGE_SIZE,
                                     )
-                                    .await
                                     .with_context(|_| {
                                         UnableToCreateParquetPageStreamSnafu::<String> {
                                             path: owned_uri.clone(),
@@ -644,7 +825,7 @@ impl ParquetFileReader {
                                 let page_stream = streaming_decompression(compressed_page_stream);
                                 let pinned_stream = Box::pin(page_stream);
                                 decompressed_iters
-                                    .push(StreamIterator::new(pinned_stream, rt_handle.clone()))
+                                    .push(StreamIterator::new(pinned_stream, rt_handle.clone()));
                             }
 
                             let (send, recv) = tokio::sync::oneshot::channel();
@@ -653,8 +834,9 @@ impl ParquetFileReader {
                                     decompressed_iters,
                                     ptypes.iter().collect(),
                                     field.clone(),
-                                    Some(128 * 1024),
+                                    Some(chunk_size),
                                     num_rows,
+                                    num_values,
                                 );
 
                                 let ser = (|| {
