@@ -1,7 +1,10 @@
-use pyo3::{prelude::*, types::PyTuple, AsPyPointer};
-use serde::{Deserialize, Serialize};
+use std::hash::{Hash, Hasher};
 
 use common_py_serde::{deserialize_py_object, serialize_py_object};
+use pyo3::{prelude::*, types::PyTuple};
+use serde::{Deserialize, Serialize};
+
+use crate::storage_config::StorageConfig;
 
 #[derive(Debug, Clone, Serialize, Deserialize)]
 struct PyObjectSerializableWrapper(
@@ -14,71 +17,79 @@ struct PyObjectSerializableWrapper(
 
 /// Python arguments to a Python function that produces Tables
 #[derive(Debug, Clone, Serialize, Deserialize)]
-pub struct PythonTablesFactoryArgs(Vec<PyObjectSerializableWrapper>);
+pub struct PythonTablesFactoryArgs {
+    args: Vec<PyObjectSerializableWrapper>,
+    hash: u64,
+}
+
+impl Hash for PythonTablesFactoryArgs {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        self.hash.hash(state);
+    }
+}
 
 impl PythonTablesFactoryArgs {
     pub fn new(args: Vec<PyObject>) -> Self {
-        Self(args.into_iter().map(PyObjectSerializableWrapper).collect())
+        let mut hasher = std::collections::hash_map::DefaultHasher::new();
+        Python::with_gil(|py| {
+            for obj in &args {
+                // Only hash hashable PyObjects.
+                if let Ok(hash) = obj.bind(py).hash() {
+                    hash.hash(&mut hasher);
+                }
+            }
+        });
+        Self {
+            args: args.into_iter().map(PyObjectSerializableWrapper).collect(),
+            hash: hasher.finish(),
+        }
     }
 
-    pub fn to_pytuple<'a>(&self, py: Python<'a>) -> &'a PyTuple {
-        pyo3::types::PyTuple::new(py, self.0.iter().map(|x| x.0.as_ref(py)))
+    #[must_use]
+    pub fn to_pytuple<'a>(&self, py: Python<'a>) -> Bound<'a, PyTuple> {
+        pyo3::types::PyTuple::new_bound(py, self.args.iter().map(|x| x.0.bind(py)))
     }
 }
 
 impl PartialEq for PythonTablesFactoryArgs {
     fn eq(&self, other: &Self) -> bool {
-        if self.0.len() != other.0.len() {
+        if self.args.len() != other.args.len() {
             return false;
         }
-        self.0
+        self.args
             .iter()
-            .zip(other.0.iter())
+            .zip(other.args.iter())
             .all(|(s, o)| (s.0.as_ptr() as isize) == (o.0.as_ptr() as isize))
     }
 }
 
 pub mod pylib {
+    use std::{default, sync::Arc};
+
+    use common_daft_config::PyDaftExecutionConfig;
     use common_error::DaftResult;
-    use daft_core::python::field::PyField;
-    use daft_core::schema::SchemaRef;
-    use daft_dsl::python::PyExpr;
-
-    use daft_core::impl_bincode_py_state_serialization;
-    use daft_stats::PartitionSpec;
-    use daft_stats::TableMetadata;
-    use daft_stats::TableStatistics;
-    use daft_table::python::PyTable;
-    use daft_table::Table;
-    use pyo3::prelude::*;
-    use pyo3::types::PyBytes;
-    use pyo3::types::PyIterator;
-    use pyo3::types::PyList;
-    use pyo3::PyTypeInfo;
-
-    use std::sync::Arc;
-
-    use daft_core::python::schema::PySchema;
-
-    use pyo3::pyclass;
+    use common_file_formats::{python::PyFileFormatConfig, FileFormatConfig};
+    use common_py_serde::impl_bincode_py_state_serialization;
+    use common_scan_info::{
+        python::pylib::{PyPartitionField, PyPushdowns},
+        PartitionField, Pushdowns, ScanOperator, ScanOperatorRef, ScanTaskLike, ScanTaskLikeRef,
+    };
+    use daft_logical_plan::{LogicalPlanBuilder, PyLogicalPlanBuilder};
+    use daft_schema::{python::schema::PySchema, schema::SchemaRef};
+    use daft_stats::{PartitionSpec, TableMetadata, TableStatistics};
+    use daft_table::{python::PyTable, Table};
+    use pyo3::{
+        prelude::*,
+        pyclass,
+        types::{PyIterator, PyList},
+    };
     use serde::{Deserialize, Serialize};
 
-    use crate::anonymous::AnonymousScanOperator;
-    use crate::file_format::FileFormatConfig;
-    use crate::storage_config::PythonStorageConfig;
-    use crate::DataFileSource;
-    use crate::PartitionField;
-    use crate::Pushdowns;
-    use crate::ScanOperator;
-    use crate::ScanOperatorRef;
-    use crate::ScanTask;
-
-    use crate::file_format::PyFileFormatConfig;
-    use crate::glob::GlobScanOperator;
-    use crate::storage_config::PyStorageConfig;
-    use common_daft_config::PyDaftExecutionConfig;
-
     use super::PythonTablesFactoryArgs;
+    use crate::{
+        anonymous::AnonymousScanOperator, glob::GlobScanOperator, storage_config::StorageConfig,
+        DataSource, ScanTask,
+    };
     #[pyclass(module = "daft.daft", frozen)]
     #[derive(Debug, Clone)]
     pub struct ScanOperatorHandle {
@@ -97,7 +108,7 @@ pub mod pylib {
             files: Vec<String>,
             schema: PySchema,
             file_format_config: PyFileFormatConfig,
-            storage_config: PyStorageConfig,
+            storage_config: StorageConfig,
         ) -> PyResult<Self> {
             py.allow_threads(|| {
                 let schema = schema.schema;
@@ -107,32 +118,41 @@ pub mod pylib {
                     file_format_config.into(),
                     storage_config.into(),
                 ));
-                Ok(ScanOperatorHandle {
+                Ok(Self {
                     scan_op: ScanOperatorRef(operator),
                 })
             })
         }
 
         #[staticmethod]
+        #[allow(clippy::too_many_arguments)]
         pub fn glob_scan(
             py: Python,
-            glob_path: Vec<&str>,
+            glob_path: Vec<String>,
             file_format_config: PyFileFormatConfig,
-            storage_config: PyStorageConfig,
+            storage_config: StorageConfig,
+            hive_partitioning: bool,
             infer_schema: bool,
             schema: Option<PySchema>,
-            is_ray_runner: Option<bool>,
+            file_path_column: Option<String>,
         ) -> PyResult<Self> {
             py.allow_threads(|| {
-                let operator = Arc::new(GlobScanOperator::try_new(
-                    glob_path.as_slice(),
+                let executor = common_runtime::get_io_runtime(true);
+
+                let task = GlobScanOperator::try_new(
+                    glob_path,
                     file_format_config.into(),
                     storage_config.into(),
                     infer_schema,
                     schema.map(|s| s.schema),
-                    is_ray_runner.unwrap_or(false),
-                )?);
-                Ok(ScanOperatorHandle {
+                    file_path_column,
+                    hive_partitioning,
+                );
+
+                let operator = executor.block_on(task)??;
+                let operator = Arc::new(operator);
+
+                Ok(Self {
                     scan_op: ScanOperatorRef(operator),
                 })
             })
@@ -143,12 +163,13 @@ pub mod pylib {
             let scan_op = ScanOperatorRef(Arc::new(PythonScanOperatorBridge::from_python_abc(
                 py_scan, py,
             )?));
-            Ok(ScanOperatorHandle { scan_op })
+            Ok(Self { scan_op })
         }
     }
     #[pyclass(module = "daft.daft")]
     #[derive(Debug)]
     struct PythonScanOperatorBridge {
+        name: String,
         operator: PyObject,
         schema: SchemaRef,
         partitioning_keys: Vec<PartitionField>,
@@ -159,6 +180,10 @@ pub mod pylib {
     }
 
     impl PythonScanOperatorBridge {
+        fn _name(abc: &PyObject, py: Python) -> PyResult<String> {
+            let result = abc.call_method0(py, pyo3::intern!(py, "name"))?;
+            result.extract::<String>(py)
+        }
         fn _partitioning_keys(abc: &PyObject, py: Python) -> PyResult<Vec<PartitionField>> {
             let result = abc.call_method0(py, pyo3::intern!(py, "partitioning_keys"))?;
             let result = result.extract::<&PyList>(py)?;
@@ -201,6 +226,7 @@ pub mod pylib {
     impl PythonScanOperatorBridge {
         #[staticmethod]
         pub fn from_python_abc(abc: PyObject, py: Python) -> PyResult<Self> {
+            let name = Self::_name(&abc, py)?;
             let partitioning_keys = Self::_partitioning_keys(&abc, py)?;
             let schema = Self::_schema(&abc, py)?;
             let can_absorb_filter = Self::_can_absorb_filter(&abc, py)?;
@@ -209,6 +235,7 @@ pub mod pylib {
             let display_name = Self::_display_name(&abc, py)?;
 
             Ok(Self {
+                name,
                 operator: abc,
                 schema,
                 partitioning_keys,
@@ -221,12 +248,21 @@ pub mod pylib {
     }
 
     impl ScanOperator for PythonScanOperatorBridge {
-        fn partitioning_keys(&self) -> &[crate::PartitionField] {
+        fn name(&self) -> &str {
+            &self.name
+        }
+        fn partitioning_keys(&self) -> &[PartitionField] {
             &self.partitioning_keys
         }
-        fn schema(&self) -> daft_core::schema::SchemaRef {
+        fn schema(&self) -> daft_schema::schema::SchemaRef {
             self.schema.clone()
         }
+        fn file_path_column(&self) -> Option<&str> {
+            None
+        }
+        fn generated_fields(&self) -> Option<SchemaRef> {
+            None
+        }
         fn can_absorb_filter(&self) -> bool {
             self.can_absorb_filter
         }
@@ -242,18 +278,13 @@ pub mod pylib {
             lines
         }
 
-        fn to_scan_tasks(
-            &self,
-            pushdowns: Pushdowns,
-        ) -> common_error::DaftResult<
-            Box<dyn Iterator<Item = common_error::DaftResult<crate::ScanTaskRef>>>,
-        > {
+        fn to_scan_tasks(&self, pushdowns: Pushdowns) -> DaftResult<Vec<ScanTaskLikeRef>> {
             let scan_tasks = Python::with_gil(|py| {
-                let pypd = PyPushdowns(pushdowns.into()).into_py(py);
+                let pypd = PyPushdowns(pushdowns.clone().into()).into_py(py);
                 let pyiter =
                     self.operator
                         .call_method1(py, pyo3::intern!(py, "to_scan_tasks"), (pypd,))?;
-                let pyiter = PyIterator::from_object(py, &pyiter)?;
+                let pyiter = PyIterator::from_bound_object(pyiter.bind(py))?;
                 DaftResult::Ok(
                     pyiter
                         .map(|v| {
@@ -263,7 +294,11 @@ pub mod pylib {
                         .collect::<Vec<_>>(),
                 )
             })?;
-            Ok(Box::new(scan_tasks.into_iter()))
+
+            scan_tasks
+                .into_iter()
+                .map(|st| st.map(|task| task as Arc<dyn ScanTaskLike>))
+                .collect()
         }
     }
 
@@ -289,10 +324,6 @@ pub mod pylib {
             Ok(self.0.num_rows().map(i64::try_from).transpose()?)
         }
 
-        pub fn size_bytes(&self) -> PyResult<Option<i64>> {
-            Ok(self.0.size_bytes().map(i64::try_from).transpose()?)
-        }
-
         pub fn estimate_in_memory_size_bytes(
             &self,
             cfg: PyDaftExecutionConfig,
@@ -313,9 +344,10 @@ pub mod pylib {
             file: String,
             file_format: PyFileFormatConfig,
             schema: PySchema,
-            storage_config: PyStorageConfig,
+            storage_config: StorageConfig,
             num_rows: Option<i64>,
             size_bytes: Option<u64>,
+            iceberg_delete_files: Option<Vec<String>>,
             pushdowns: Option<PyPushdowns>,
             partition_values: Option<PyTable>,
             stats: Option<PyTable>,
@@ -328,7 +360,7 @@ pub mod pylib {
                 let eval_pred = table.eval_expression_list(&[partition_filters.clone()])?;
                 assert_eq!(eval_pred.num_columns(), 1);
                 let series = eval_pred.get_column_by_index(0)?;
-                assert_eq!(series.data_type(), &daft_core::DataType::Boolean);
+                assert_eq!(series.data_type(), &daft_core::datatypes::DataType::Boolean);
                 let boolean = series.bool()?;
                 assert_eq!(boolean.len(), 1);
                 let value = boolean.get(0);
@@ -340,34 +372,23 @@ pub mod pylib {
             // TODO(Clark): Filter out scan tasks with pushed down filters + table stats?
 
             let pspec = PartitionSpec {
-                keys: partition_values
-                    .map(|p| p.table)
-                    .unwrap_or_else(|| Table::empty(None).unwrap()),
+                keys: partition_values.map_or_else(|| Table::empty(None).unwrap(), |p| p.table),
             };
             let statistics = stats
                 .map(|s| TableStatistics::from_stats_table(&s.table))
                 .transpose()?;
 
-            let table_metadata = num_rows.map(|n| TableMetadata { length: n as usize });
+            let metadata = num_rows.map(|n| TableMetadata { length: n as usize });
 
-            let data_source = match table_metadata {
-                None => DataFileSource::AnonymousDataFile {
-                    path: file,
-                    chunk_spec: None,
-                    size_bytes,
-                    metadata: None,
-                    partition_spec: Some(pspec),
-                    statistics,
-                    parquet_metadata: None,
-                },
-                Some(tm) => DataFileSource::CatalogDataFile {
-                    path: file,
-                    chunk_spec: None,
-                    size_bytes,
-                    metadata: tm,
-                    partition_spec: pspec,
-                    statistics,
-                },
+            let data_source = DataSource::File {
+                path: file,
+                chunk_spec: None,
+                size_bytes,
+                iceberg_delete_files,
+                metadata,
+                partition_spec: Some(pspec),
+                statistics,
+                parquet_metadata: None,
             };
 
             let scan_task = ScanTask::new(
@@ -376,8 +397,9 @@ pub mod pylib {
                 schema.schema,
                 storage_config.into(),
                 pushdowns.map(|p| p.0.as_ref().clone()).unwrap_or_default(),
+                None,
             );
-            Ok(Some(PyScanTask(scan_task.into())))
+            Ok(Some(Self(scan_task.into())))
         }
 
         #[allow(clippy::too_many_arguments)]
@@ -386,7 +408,7 @@ pub mod pylib {
             url: String,
             file_format: PyFileFormatConfig,
             schema: PySchema,
-            storage_config: PyStorageConfig,
+            storage_config: StorageConfig,
             num_rows: Option<i64>,
             size_bytes: Option<u64>,
             pushdowns: Option<PyPushdowns>,
@@ -395,12 +417,10 @@ pub mod pylib {
             let statistics = stats
                 .map(|s| TableStatistics::from_stats_table(&s.table))
                 .transpose()?;
-            let data_source = DataFileSource::DatabaseDataSource {
+            let data_source = DataSource::Database {
                 path: url,
-                chunk_spec: None,
                 size_bytes,
                 metadata: num_rows.map(|n| TableMetadata { length: n as usize }),
-                partition_spec: None,
                 statistics,
             };
 
@@ -410,8 +430,9 @@ pub mod pylib {
                 schema.schema,
                 storage_config.into(),
                 pushdowns.map(|p| p.0.as_ref().clone()).unwrap_or_default(),
+                None,
             );
-            Ok(PyScanTask(scan_task.into()))
+            Ok(Self(scan_task.into()))
         }
 
         #[allow(clippy::too_many_arguments)]
@@ -420,7 +441,7 @@ pub mod pylib {
             py: Python,
             module: String,
             func_name: String,
-            func_args: Vec<&PyAny>,
+            func_args: Vec<Bound<PyAny>>,
             schema: PySchema,
             num_rows: Option<i64>,
             size_bytes: Option<u64>,
@@ -430,7 +451,7 @@ pub mod pylib {
             let statistics = stats
                 .map(|s| TableStatistics::from_stats_table(&s.table))
                 .transpose()?;
-            let data_source = DataFileSource::PythonFactoryFunction {
+            let data_source = DataSource::PythonFactoryFunction {
                 module,
                 func_name,
                 func_args: PythonTablesFactoryArgs::new(
@@ -450,12 +471,11 @@ pub mod pylib {
                 schema.schema,
                 // HACK: StorageConfig isn't used when running the Python function but this is a non-optional arg for
                 // ScanTask creation, so we just put in a placeholder here
-                Arc::new(crate::storage_config::StorageConfig::Python(Arc::new(
-                    PythonStorageConfig { io_config: None },
-                ))),
+                Arc::new(Default::default()),
                 pushdowns.map(|p| p.0.as_ref().clone()).unwrap_or_default(),
+                None,
             );
-            Ok(PyScanTask(scan_task.into()))
+            Ok(Self(scan_task.into()))
         }
 
         pub fn __repr__(&self) -> PyResult<String> {
@@ -477,132 +497,94 @@ pub mod pylib {
 
     impl_bincode_py_state_serialization!(PyScanTask);
 
-    #[pyclass(module = "daft.daft", name = "PartitionField", frozen)]
-    #[derive(Debug, Clone, Serialize, Deserialize)]
-    pub struct PyPartitionField(Arc<PartitionField>);
-
-    #[pymethods]
-    impl PyPartitionField {
-        #[new]
-        fn new(
-            field: PyField,
-            source_field: Option<PyField>,
-            transform: Option<PyPartitionTransform>,
-        ) -> PyResult<Self> {
-            let p_field = PartitionField::new(
-                field.field,
-                source_field.map(|f| f.into()),
-                transform.map(|e| e.0),
-            )?;
-            Ok(PyPartitionField(Arc::new(p_field)))
-        }
-
-        pub fn __repr__(&self) -> PyResult<String> {
-            Ok(format!("{}", self.0))
-        }
-
-        #[getter]
-        pub fn field(&self) -> PyResult<PyField> {
-            Ok(self.0.field.clone().into())
-        }
+    #[pyfunction]
+    pub fn logical_plan_table_scan(
+        scan_operator: ScanOperatorHandle,
+    ) -> PyResult<PyLogicalPlanBuilder> {
+        Ok(LogicalPlanBuilder::table_scan(scan_operator.into(), None)?.into())
     }
 
-    #[pyclass(module = "daft.daft", name = "PartitionTransform", frozen)]
-    #[derive(Debug, Clone, Serialize, Deserialize)]
-    pub struct PyPartitionTransform(crate::PartitionTransform);
-
-    #[pymethods]
-    impl PyPartitionTransform {
-        #[staticmethod]
-        pub fn identity() -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::Identity))
-        }
-
-        #[staticmethod]
-        pub fn year() -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::Year))
-        }
-
-        #[staticmethod]
-        pub fn month() -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::Month))
-        }
-
-        #[staticmethod]
-        pub fn day() -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::Day))
-        }
-
-        #[staticmethod]
-        pub fn hour() -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::Hour))
-        }
-
-        #[staticmethod]
-        pub fn void() -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::Void))
-        }
-
-        #[staticmethod]
-        pub fn iceberg_bucket(n: u64) -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::IcebergBucket(n)))
-        }
-
-        #[staticmethod]
-        pub fn iceberg_truncate(n: u64) -> PyResult<Self> {
-            Ok(Self(crate::PartitionTransform::IcebergTruncate(n)))
-        }
-
-        pub fn __repr__(&self) -> PyResult<String> {
-            Ok(format!("{}", self.0))
-        }
-    }
-
-    #[pyclass(module = "daft.daft", name = "Pushdowns", frozen)]
-    #[derive(Debug, Clone, Serialize, Deserialize)]
-    pub struct PyPushdowns(Arc<Pushdowns>);
-    #[pymethods]
-    impl PyPushdowns {
-        pub fn __repr__(&self) -> PyResult<String> {
-            Ok(format!("{:#?}", self.0))
-        }
-        #[getter]
-        pub fn limit(&self) -> Option<usize> {
-            self.0.limit
-        }
-
-        #[getter]
-        pub fn filters(&self) -> Option<PyExpr> {
-            self.0.filters.as_ref().map(|e| PyExpr { expr: e.clone() })
-        }
-
-        #[getter]
-        pub fn partition_filters(&self) -> Option<PyExpr> {
-            self.0
-                .partition_filters
-                .as_ref()
-                .map(|e| PyExpr { expr: e.clone() })
-        }
-
-        #[getter]
-        pub fn columns(&self) -> Option<Vec<String>> {
-            self.0.columns.as_deref().cloned()
-        }
-
-        pub fn filter_required_column_names(&self) -> Option<Vec<String>> {
-            self.0
-                .filters
-                .as_ref()
-                .map(daft_dsl::optimization::get_required_columns)
-        }
+    /// Estimates the in-memory size in bytes for a Parquet file.
+    ///
+    /// This function calculates an approximate size that the Parquet file would occupy
+    /// when loaded into memory, considering only the specified columns if provided.
+    ///
+    /// Used for testing only.
+    ///
+    /// # Arguments
+    ///
+    /// * `uri` - A string slice that holds the URI of the Parquet file.
+    /// * `file_size` - the size of the file on disk
+    /// * `columns` - An optional vector of strings representing the column names to consider.
+    ///               If None, all columns in the file will be considered.
+    /// * `has_metadata` - whether or not metadata is pre-populated in the ScanTask. Defaults to false.
+    ///
+    /// # Returns
+    ///
+    /// Returns an `i64` representing the estimated size in bytes.
+    ///
+    #[pyfunction]
+    pub fn estimate_in_memory_size_bytes(
+        uri: &str,
+        file_size: u64,
+        columns: Option<Vec<String>>,
+        has_metadata: Option<bool>,
+    ) -> PyResult<usize> {
+        let io_runtime = common_runtime::get_io_runtime(true);
+        let (schema, metadata) =
+            io_runtime.block_on_current_thread(daft_parquet::read::read_parquet_schema(
+                uri,
+                default::Default::default(),
+                None,
+                default::Default::default(),
+                None,
+            ))?;
+        let data_source = DataSource::File {
+            path: uri.to_string(),
+            chunk_spec: None,
+            size_bytes: Some(file_size),
+            iceberg_delete_files: None,
+            metadata: if has_metadata.unwrap_or(false) {
+                Some(TableMetadata {
+                    length: metadata.num_rows,
+                })
+            } else {
+                None
+            },
+            partition_spec: None,
+            statistics: None,
+            parquet_metadata: None,
+        };
+        let st = ScanTask::new(
+            vec![data_source],
+            Arc::new(FileFormatConfig::Parquet(default::Default::default())),
+            Arc::new(schema),
+            Arc::new(Default::default()),
+            Pushdowns::new(None, None, columns.map(Arc::new), None),
+            None,
+        );
+        Ok(st.estimate_in_memory_size_bytes(None).unwrap())
     }
 }
 
-pub fn register_modules(_py: Python, parent: &PyModule) -> PyResult<()> {
+pub fn register_modules(parent: &Bound<PyModule>) -> PyResult<()> {
+    parent.add_class::<StorageConfig>()?;
+
     parent.add_class::<pylib::ScanOperatorHandle>()?;
     parent.add_class::<pylib::PyScanTask>()?;
-    parent.add_class::<pylib::PyPartitionField>()?;
-    parent.add_class::<pylib::PyPartitionTransform>()?;
-    parent.add_class::<pylib::PyPushdowns>()?;
+    parent.add_function(wrap_pyfunction_bound!(
+        pylib::logical_plan_table_scan,
+        parent
+    )?)?;
+
+    Ok(())
+}
+
+pub fn register_testing_modules(parent: &Bound<PyModule>) -> PyResult<()> {
+    parent.add_function(wrap_pyfunction_bound!(
+        pylib::estimate_in_memory_size_bytes,
+        parent
+    )?)?;
+
     Ok(())
 }
