@@ -9,7 +9,7 @@ mod sort;
 pub use column_chunk_metadata::ColumnChunkMetaData;
 pub use column_descriptor::{ColumnDescriptor, Descriptor};
 pub use column_order::ColumnOrder;
-pub use file_metadata::{FileMetaData, KeyValue};
+pub use file_metadata::{FileMetaData, KeyValue, RowGroupList};
 pub use row_metadata::RowGroupMetaData;
 pub use schema_descriptor::SchemaDescriptor;
 pub use sort::*;
