@@ -8,7 +8,7 @@ from tests.utils import sort_arrow_table
 
 
 @pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
-def test_distinct_with_nulls(make_df, repartition_nparts):
+def test_distinct_with_nulls(make_df, repartition_nparts, with_morsel_size):
     daft_df = make_df(
         {
             "id": [1, None, None, None],
@@ -28,7 +28,7 @@ def test_distinct_with_nulls(make_df, repartition_nparts):
 
 
 @pytest.mark.parametrize("repartition_nparts", [1, 2, 5])
-def test_distinct_with_all_nulls(make_df, repartition_nparts):
+def test_distinct_with_all_nulls(make_df, repartition_nparts, with_morsel_size):
     daft_df = make_df(
         {
             "id": [None, None, None, None],
@@ -48,7 +48,7 @@ def test_distinct_with_all_nulls(make_df, repartition_nparts):
 
 
 @pytest.mark.parametrize("repartition_nparts", [1, 2])
-def test_distinct_with_empty(make_df, repartition_nparts):
+def test_distinct_with_empty(make_df, repartition_nparts, with_morsel_size):
     daft_df = make_df(
         {
             "id": [1],
