@@ -3,20 +3,23 @@ from __future__ import annotations
 import pyarrow as pa
 import pytest
 
+import daft
+from daft import col
 from daft.datatype import DataType
 from daft.errors import ExpressionTypeError
+from tests.conftest import get_tests_daft_runner_name
 from tests.utils import sort_arrow_table
 
 
 def skip_invalid_join_strategies(join_strategy, join_type):
-    if (join_strategy == "sort_merge" or join_strategy == "sort_merge_aligned_boundaries") and join_type != "inner":
-        pytest.skip("Sort merge currently only supports inner joins")
-    elif join_strategy == "broadcast" and join_type == "outer":
-        pytest.skip("Broadcast join does not support outer joins")
-    elif join_strategy == "broadcast" and join_type == "anti":
-        pytest.skip("Broadcast join does not support anti joins")
-    elif join_strategy == "broadcast" and join_type == "semi":
-        pytest.skip("Broadcast join does not support semi joins")
+    if get_tests_daft_runner_name() == "native":
+        if join_strategy not in [None, "hash"]:
+            pytest.skip("Native executor fails for these tests")
+    else:
+        if (join_strategy == "sort_merge" or join_strategy == "sort_merge_aligned_boundaries") and join_type != "inner":
+            pytest.skip("Sort merge currently only supports inner joins")
+        elif join_strategy == "broadcast" and join_type == "outer":
+            pytest.skip("Broadcast join does not support outer joins")
 
 
 def test_invalid_join_strategies(make_df):
@@ -52,6 +55,17 @@ def test_columns_after_join(make_df):
     assert set(joined_df2.schema().column_names()) == set(["A", "B"])
 
 
+def test_rename_join_keys_in_dataframe(make_df):
+    df1 = make_df({"A": [1, 2], "B": [2, 2]})
+
+    df2 = make_df({"A": [1, 2]})
+    joined_df1 = df1.join(df2, left_on=["A", "B"], right_on=["A", "A"])
+    joined_df2 = df1.join(df2, left_on=["B", "A"], right_on=["A", "A"])
+
+    assert set(joined_df1.schema().column_names()) == set(["A", "B"])
+    assert set(joined_df2.schema().column_names()) == set(["A", "B"])
+
+
 @pytest.mark.parametrize("n_partitions", [1, 2, 4])
 @pytest.mark.parametrize(
     "join_strategy",
@@ -59,7 +73,7 @@ def test_columns_after_join(make_df):
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_joins(join_strategy, join_type, make_df, n_partitions: int):
+def test_joins(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df = make_df(
@@ -91,7 +105,7 @@ def test_joins(join_strategy, join_type, make_df, n_partitions: int):
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_multicol_joins(join_strategy, join_type, make_df, n_partitions: int):
+def test_multicol_joins(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df = make_df(
@@ -125,7 +139,7 @@ def test_multicol_joins(join_strategy, join_type, make_df, n_partitions: int):
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int):
+def test_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df = make_df(
@@ -138,7 +152,7 @@ def test_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int):
     )
 
     joined = df.join(df, on="A", strategy=join_strategy, how=join_type)
-    joined = joined.sort(["A", "B"])
+    joined = joined.sort(["A", "B", "right.B"])
     joined_data = joined.to_pydict()
 
     assert joined_data == {
@@ -155,7 +169,7 @@ def test_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int):
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_multicol_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int):
+def test_multicol_dupes_join_key(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df = make_df(
@@ -169,14 +183,14 @@ def test_multicol_dupes_join_key(join_strategy, join_type, make_df, n_partitions
     )
 
     joined = df.join(df, on=["A", "B"], strategy=join_strategy, how=join_type)
-    joined = joined.sort(["A", "B", "C"])
+    joined = joined.sort(["A", "B", "C", "right.C"])
     joined_data = joined.to_pydict()
 
     assert joined_data == {
         "A": [1, 1, 1, 1, 2, 2, 2, 2, 3, 3],
         "B": ["a"] * 4 + ["b"] * 4 + ["c", "d"],
         "C": [0, 0, 1, 1, 0, 0, 1, 1, 1, 0],
-        "right.C": [1, 0, 1, 0, 1, 0, 1, 0, 1, 0],
+        "right.C": [0, 1, 0, 1, 0, 1, 0, 1, 1, 0],
     }
 
 
@@ -187,7 +201,7 @@ def test_multicol_dupes_join_key(join_strategy, join_type, make_df, n_partitions
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_joins_all_same_key(join_strategy, join_type, make_df, n_partitions: int):
+def test_joins_all_same_key(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df = make_df(
@@ -200,7 +214,7 @@ def test_joins_all_same_key(join_strategy, join_type, make_df, n_partitions: int
     )
 
     joined = df.join(df, on="A", strategy=join_strategy, how=join_type)
-    joined = joined.sort(["A", "B"])
+    joined = joined.sort(["A", "B", "right.B"])
     joined_data = joined.to_pydict()
 
     assert joined_data == {
@@ -261,7 +275,9 @@ def test_joins_all_same_key(join_strategy, join_type, make_df, n_partitions: int
         ),
     ],
 )
-def test_joins_no_overlap_disjoint(join_strategy, join_type, flip, expected, make_df, n_partitions: int):
+def test_joins_no_overlap_disjoint(
+    join_strategy, join_type, flip, expected, make_df, n_partitions: int, with_morsel_size
+):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df1 = make_df(
@@ -342,7 +358,9 @@ def test_joins_no_overlap_disjoint(join_strategy, join_type, flip, expected, mak
         ),
     ],
 )
-def test_joins_no_overlap_interleaved(join_strategy, join_type, flip, expected, make_df, n_partitions: int):
+def test_joins_no_overlap_interleaved(
+    join_strategy, join_type, flip, expected, make_df, n_partitions: int, with_morsel_size
+):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     df1 = make_df(
@@ -381,7 +399,7 @@ def test_joins_no_overlap_interleaved(join_strategy, join_type, flip, expected,
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_limit_after_join(join_strategy, join_type, make_df, n_partitions: int):
+def test_limit_after_join(join_strategy, join_type, make_df, n_partitions: int, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     data = {
@@ -448,7 +466,7 @@ def test_limit_after_join(join_strategy, join_type, make_df, n_partitions: int):
         ),
     ],
 )
-def test_join_with_null(join_strategy, join_type, expected, make_df, repartition_nparts):
+def test_join_with_null(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df = make_df(
@@ -514,7 +532,7 @@ def test_join_with_null(join_strategy, join_type, expected, make_df, repartition
         ),
     ],
 )
-def test_join_with_null_multikey(join_strategy, join_type, expected, make_df, repartition_nparts):
+def test_join_with_null_multikey(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df = make_df(
@@ -597,7 +615,9 @@ def test_join_with_null_multikey(join_strategy, join_type, expected, make_df, re
         ),
     ],
 )
-def test_join_with_null_asymmetric_multikey(join_strategy, join_type, expected, make_df, repartition_nparts):
+def test_join_with_null_asymmetric_multikey(
+    join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size
+):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df = make_df(
@@ -672,7 +692,7 @@ def test_join_with_null_asymmetric_multikey(join_strategy, join_type, expected,
         ),
     ],
 )
-def test_join_all_null(join_strategy, join_type, expected, make_df, repartition_nparts):
+def test_join_all_null(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df = make_df(
@@ -706,7 +726,7 @@ def test_join_all_null(join_strategy, join_type, expected, make_df, repartition_
     indirect=True,
 )
 @pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
-def test_join_null_type_column(join_strategy, join_type, make_df):
+def test_join_null_type_column(join_strategy, join_type, make_df, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df = make_df(
@@ -751,7 +771,7 @@ def test_join_null_type_column(join_strategy, join_type, make_df):
         ),
     ],
 )
-def test_join_semi_anti(join_strategy, join_type, expected, make_df, repartition_nparts):
+def test_join_semi_anti(join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df1 = make_df(
@@ -804,7 +824,9 @@ def test_join_semi_anti(join_strategy, join_type, expected, make_df, repartition
         ),
     ],
 )
-def test_join_semi_anti_different_names(join_strategy, join_type, expected, make_df, repartition_nparts):
+def test_join_semi_anti_different_names(
+    join_strategy, join_type, expected, make_df, repartition_nparts, with_morsel_size
+):
     skip_invalid_join_strategies(join_strategy, join_type)
 
     daft_df1 = make_df(
@@ -836,3 +858,399 @@ def test_join_semi_anti_different_names(join_strategy, join_type, expected, make
     assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "id_left") == sort_arrow_table(
         pa.Table.from_pydict(expected), "id_left"
     )
+
+
+@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer"])
+def test_join_true_join_keys(join_type, make_df, with_morsel_size):
+    daft_df = make_df(
+        {
+            "id": [1, 2, 3],
+            "values": ["a", "b", "c"],
+        }
+    )
+    daft_df2 = make_df(
+        {
+            "id": [2.0, 2.5, 3.0, 4.0],
+            "values": ["a2", "b2", "c2", "d2"],
+        }
+    )
+
+    result = daft_df.join(daft_df2, left_on=["id", "values"], right_on=["id", col("values").str.left(1)], how=join_type)
+
+    assert result.schema().column_names() == ["id", "values", "right.values"]
+    assert result.schema()["id"].dtype == daft_df.schema()["id"].dtype
+    assert result.schema()["values"].dtype == daft_df.schema()["values"].dtype
+    assert result.schema()["right.values"].dtype == daft_df2.schema()["values"].dtype
+
+
+@pytest.mark.parametrize(
+    "join_strategy",
+    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
+    indirect=True,
+)
+@pytest.mark.parametrize(
+    "join_type,expected",
+    [
+        (
+            "inner",
+            {
+                "a": [2, 3],
+                "b": [2, 3],
+            },
+        ),
+        (
+            "left",
+            {
+                "a": [1, 2, 3],
+                "b": [None, 2, 3],
+            },
+        ),
+        (
+            "right",
+            {
+                "a": [2, 3, None],
+                "b": [2, 3, 4],
+            },
+        ),
+        (
+            "outer",
+            {
+                "a": [1, 2, 3, None],
+                "b": [None, 2, 3, 4],
+            },
+        ),
+        (
+            "semi",
+            {
+                "a": [2, 3],
+            },
+        ),
+        (
+            "anti",
+            {
+                "a": [1],
+            },
+        ),
+    ],
+)
+def test_join_with_alias_in_key(join_strategy, join_type, expected, make_df, with_morsel_size):
+    skip_invalid_join_strategies(join_strategy, join_type)
+
+    daft_df1 = make_df(
+        {
+            "a": [1, 2, 3],
+        }
+    )
+    daft_df2 = make_df(
+        {
+            "b": [2, 3, 4],
+        }
+    )
+
+    daft_df = daft_df1.join(daft_df2, left_on=col("a").alias("x"), right_on="b", how=join_type, strategy=join_strategy)
+
+    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "a") == sort_arrow_table(
+        pa.Table.from_pydict(expected), "a"
+    )
+
+
+@pytest.mark.parametrize(
+    "join_strategy",
+    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
+    indirect=True,
+)
+@pytest.mark.parametrize(
+    "join_type,expected",
+    [
+        (
+            "inner",
+            {
+                "a": [2, 3],
+                "right.a": [2, 3],
+            },
+        ),
+        (
+            "left",
+            {
+                "a": [1, 2, 3],
+                "right.a": [None, 2, 3],
+            },
+        ),
+        (
+            "right",
+            {
+                "a": [2, 3, None],
+                "right.a": [2, 3, 4],
+            },
+        ),
+        (
+            "outer",
+            {
+                "a": [1, 2, 3, None],
+                "right.a": [None, 2, 3, 4],
+            },
+        ),
+        (
+            "semi",
+            {
+                "a": [2, 3],
+            },
+        ),
+        (
+            "anti",
+            {
+                "a": [1],
+            },
+        ),
+    ],
+)
+def test_join_same_name_alias(join_strategy, join_type, expected, make_df, with_morsel_size):
+    skip_invalid_join_strategies(join_strategy, join_type)
+
+    daft_df1 = make_df(
+        {
+            "a": [1, 2, 3],
+        }
+    )
+    daft_df2 = make_df(
+        {
+            "a": [2, 3, 4],
+        }
+    )
+
+    daft_df = daft_df1.join(daft_df2, left_on="a", right_on=col("a").alias("b"), how=join_type, strategy=join_strategy)
+
+    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "a") == sort_arrow_table(
+        pa.Table.from_pydict(expected), "a"
+    )
+
+
+@pytest.mark.parametrize(
+    "join_strategy",
+    [None, "hash", "sort_merge", "sort_merge_aligned_boundaries", "broadcast"],
+    indirect=True,
+)
+@pytest.mark.parametrize(
+    "join_type,expected",
+    [
+        (
+            "inner",
+            {
+                "a": [0.2, 0.3],
+                "right.a": [20, 30],
+            },
+        ),
+        (
+            "left",
+            {
+                "a": [0.1, 0.2, 0.3],
+                "right.a": [None, 20, 30],
+            },
+        ),
+        (
+            "right",
+            {
+                "a": [0.2, 0.3, None],
+                "right.a": [20, 30, 40],
+            },
+        ),
+        (
+            "outer",
+            {
+                "a": [0.1, 0.2, 0.3, None],
+                "right.a": [None, 20, 30, 40],
+            },
+        ),
+        (
+            "semi",
+            {
+                "a": [0.2, 0.3],
+            },
+        ),
+        (
+            "anti",
+            {
+                "a": [0.1],
+            },
+        ),
+    ],
+)
+def test_join_same_name_alias_with_compute(join_strategy, join_type, expected, make_df, with_morsel_size):
+    skip_invalid_join_strategies(join_strategy, join_type)
+
+    daft_df1 = make_df(
+        {
+            "a": [0.1, 0.2, 0.3],
+        }
+    )
+    daft_df2 = make_df(
+        {
+            "a": [20, 30, 40],
+        }
+    )
+
+    daft_df = daft_df1.join(
+        daft_df2, left_on=col("a") * 10, right_on=(col("a") / 10).alias("b"), how=join_type, strategy=join_strategy
+    )
+
+    assert sort_arrow_table(pa.Table.from_pydict(daft_df.to_pydict()), "a") == sort_arrow_table(
+        pa.Table.from_pydict(expected), "a"
+    )
+
+
+@pytest.mark.parametrize(
+    "suffix,prefix,expected",
+    [
+        (None, None, "right.score"),
+        ("_right", None, "score_right"),
+        (None, "left_", "left_score"),
+        ("_right", "prefix.", "prefix.score_right"),
+    ],
+)
+def test_join_suffix_and_prefix(suffix, prefix, expected, make_df, with_morsel_size):
+    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
+    df2 = daft.from_pydict({"idx": [3], "score": [0.1]})
+    df3 = daft.from_pydict({"idx": [1], "score": [0.1]})
+
+    df = df1.join(df2, on="idx").join(df3, on="idx", suffix=suffix, prefix=prefix)
+    assert df.column_names == ["idx", "val", "score", expected]
+
+
+@pytest.mark.parametrize("left_partitions", [1, 2, 4])
+@pytest.mark.parametrize("right_partitions", [1, 2, 4])
+def test_cross_join(left_partitions, right_partitions, make_df, with_morsel_size):
+    df1 = make_df(
+        {
+            "A": [1, 3, 5],
+            "B": ["a", "b", "c"],
+        },
+        repartition=left_partitions,
+        repartition_columns=["A"],
+    )
+
+    df2 = make_df(
+        {
+            "C": [2, 4, 6, 8],
+            "D": ["d", "e", "f", "g"],
+        },
+        repartition=right_partitions,
+        repartition_columns=["C"],
+    )
+
+    df = df1.join(df2, how="cross")
+    df = df.sort(["A", "C"])
+
+    assert df.to_pydict() == {
+        "A": [1, 1, 1, 1, 3, 3, 3, 3, 5, 5, 5, 5],
+        "B": ["a", "a", "a", "a", "b", "b", "b", "b", "c", "c", "c", "c"],
+        "C": [2, 4, 6, 8, 2, 4, 6, 8, 2, 4, 6, 8],
+        "D": ["d", "e", "f", "g", "d", "e", "f", "g", "d", "e", "f", "g"],
+    }
+
+
+@pytest.mark.parametrize("join_type", ["inner", "left", "right", "outer", "anti", "semi", "cross"])
+@pytest.mark.parametrize("repartition_nparts", [1, 2, 4])
+@pytest.mark.parametrize(
+    "left,right,expected",
+    [
+        # Case 1: Left has data, right is empty
+        (
+            {"a": [1, 2, 3, 4], "b": ["a", "b", "c", "d"]},
+            {"c": [], "d": []},
+            {
+                "inner": {"a": [], "b": [], "c": [], "d": []},
+                "left": {
+                    "a": [1, 2, 3, 4],
+                    "b": ["a", "b", "c", "d"],
+                    "c": [None, None, None, None],
+                    "d": [None, None, None, None],
+                },
+                "right": {"a": [], "b": [], "c": [], "d": []},
+                "outer": {
+                    "a": [1, 2, 3, 4],
+                    "b": ["a", "b", "c", "d"],
+                    "c": [None, None, None, None],
+                    "d": [None, None, None, None],
+                },
+                "anti": {"a": [1, 2, 3, 4], "b": ["a", "b", "c", "d"]},
+                "semi": {"a": [], "b": []},
+                "cross": {"a": [], "b": [], "c": [], "d": []},
+            },
+        ),
+        # Case 2: Left is empty, right has data
+        (
+            {"a": [], "b": []},
+            {"c": [5, 6, 7], "d": ["e", "f", "g"]},
+            {
+                "inner": {"a": [], "b": [], "c": [], "d": []},
+                "left": {"a": [], "b": [], "c": [], "d": []},
+                "right": {"a": [None, None, None], "b": [None, None, None], "c": [5, 6, 7], "d": ["e", "f", "g"]},
+                "outer": {"a": [None, None, None], "b": [None, None, None], "c": [5, 6, 7], "d": ["e", "f", "g"]},
+                "anti": {"a": [], "b": []},
+                "semi": {"a": [], "b": []},
+                "cross": {"a": [], "b": [], "c": [], "d": []},
+            },
+        ),
+        # Case 3: Both empty
+        (
+            {"a": [], "b": []},
+            {"c": [], "d": []},
+            {
+                "inner": {"a": [], "b": [], "c": [], "d": []},
+                "left": {"a": [], "b": [], "c": [], "d": []},
+                "right": {"a": [], "b": [], "c": [], "d": []},
+                "outer": {"a": [], "b": [], "c": [], "d": []},
+                "anti": {"a": [], "b": []},
+                "semi": {"a": [], "b": []},
+                "cross": {"a": [], "b": [], "c": [], "d": []},
+            },
+        ),
+    ],
+)
+def test_join_empty(join_type, repartition_nparts, left, right, expected, make_df, with_morsel_size):
+    left = pa.Table.from_pydict(
+        left,
+        schema=pa.schema(
+            [
+                ("a", pa.int32()),
+                ("b", pa.string()),
+            ]
+        ),
+    )
+    left_df = make_df(
+        left,
+        repartition=repartition_nparts,
+        repartition_columns=["a"],
+    )
+
+    right = pa.Table.from_pydict(
+        right,
+        schema=pa.schema(
+            [
+                ("c", pa.int32()),
+                ("d", pa.string()),
+            ]
+        ),
+    )
+    right_df = make_df(
+        right,
+        repartition=repartition_nparts,
+        repartition_columns=["c"],
+    )
+
+    if join_type == "cross":
+        left_on = None
+        right_on = None
+    else:
+        left_on = ["a"]
+        right_on = ["c"]
+
+    result = left_df.join(right_df, left_on=left_on, right_on=right_on, how=join_type)
+    if join_type in ["inner", "left", "right", "outer", "cross"]:
+        result = result.sort(["a", "b", "c", "d"])
+    else:
+        result = result.sort(["a", "b"])
+
+    expected_result = expected[join_type]
+
+    assert result.to_pydict() == expected_result
