@@ -19,6 +19,8 @@ def get_filesystem_from_path(path: str, **kwargs) -> fsspec.AbstractFileSystem:
     return fs
 
 
+# TODO(desmond): Expected some of these non-working cases to work after recent fixes, but that
+#                doesn't seem to be the case.
 # Taken from our spreadsheet of files that Daft should be able to handle
 DAFT_CAN_READ_FILES = [
     (
@@ -215,16 +217,11 @@ def set_split_config(request):
     max_size = 0 if request.param[0] else 384 * 1024 * 1024
     min_size = 0 if request.param[1] else 96 * 1024 * 1024
 
-    old_execution_config = daft.context.get_context().daft_execution_config
-
-    try:
-        daft.set_execution_config(
-            scan_tasks_max_size_bytes=max_size,
-            scan_tasks_min_size_bytes=min_size,
-        )
+    with daft.execution_config_ctx(
+        scan_tasks_max_size_bytes=max_size,
+        scan_tasks_min_size_bytes=min_size,
+    ):
         yield
-    finally:
-        daft.set_execution_config(old_execution_config)
 
 
 def read_parquet_with_pyarrow(path) -> pa.Table:
@@ -411,7 +408,7 @@ def test_row_groups_selection_into_pyarrow_bulk(public_storage_io_config, multit
     "multithreaded_io",
     [False, True],
 )
-def test_connect_timeout(multithreaded_io):
+def test_connect_timeout_s3(multithreaded_io):
     url = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
     connect_timeout_config = daft.io.IOConfig(
         s3=daft.io.S3Config(
@@ -432,7 +429,7 @@ def test_connect_timeout(multithreaded_io):
     "multithreaded_io",
     [False, True],
 )
-def test_read_timeout(multithreaded_io):
+def test_read_timeout_s3(multithreaded_io):
     url = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
     read_timeout_config = daft.io.IOConfig(
         s3=daft.io.S3Config(
@@ -462,3 +459,43 @@ def test_read_file_level_timeout():
 
     with pytest.raises((ReadTimeoutError), match=f"Parquet reader timed out while trying to read: {url}"):
         daft.table.read_parquet_into_pyarrow(url, io_config=read_timeout_config, file_timeout_ms=2)
+
+
+@pytest.mark.integration()
+@pytest.mark.parametrize(
+    "multithreaded_io",
+    [False, True],
+)
+def test_connect_timeout_gcs(multithreaded_io):
+    url = "gs://daft-public-data-gs/mvp.parquet"
+    connect_timeout_config = daft.io.IOConfig(
+        gcs=daft.io.GCSConfig(
+            anonymous=True,
+            connect_timeout_ms=1,
+            retry_initial_backoff_ms=10,
+            num_tries=3,
+        )
+    )
+
+    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match=f"timed out when trying to connect to {url}"):
+        MicroPartition.read_parquet(url, io_config=connect_timeout_config, multithreaded_io=multithreaded_io).to_arrow()
+
+
+@pytest.mark.integration()
+@pytest.mark.parametrize(
+    "multithreaded_io",
+    [False, True],
+)
+def test_read_timeout_gcs(multithreaded_io):
+    url = "gs://daft-public-data-gs/mvp.parquet"
+    read_timeout_config = daft.io.IOConfig(
+        gcs=daft.io.GCSConfig(
+            anonymous=True,
+            read_timeout_ms=1,
+            retry_initial_backoff_ms=10,
+            num_tries=3,
+        )
+    )
+
+    with pytest.raises((ReadTimeoutError, ConnectTimeoutError), match=f"Read timed out when trying to read {url}"):
+        MicroPartition.read_parquet(url, io_config=read_timeout_config, multithreaded_io=multithreaded_io).to_arrow()
