@@ -1,8 +1,5 @@
 from __future__ import annotations
 
-import contextlib
-import sys
-
 import pyarrow as pa
 import pytest
 
@@ -11,23 +8,10 @@ from daft.io.object_store_options import io_config_to_storage_options
 from daft.logical.schema import Schema
 from tests.utils import assert_pyarrow_tables_equal
 
-
-@contextlib.contextmanager
-def split_small_pq_files():
-    old_config = daft.context.get_context().daft_execution_config
-    daft.set_execution_config(
-        # Splits any parquet files >100 bytes in size
-        scan_tasks_min_size_bytes=1,
-        scan_tasks_max_size_bytes=100,
-    )
-    yield
-    daft.set_execution_config(config=old_config)
-
-
-PYARROW_LE_8_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (8, 0, 0)
-PYTHON_LT_3_8 = sys.version_info[:2] < (3, 8)
+PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
 pytestmark = pytest.mark.skipif(
-    PYARROW_LE_8_0_0 or PYTHON_LT_3_8, reason="deltalake only supported if pyarrow >= 8.0.0 and python >= 3.8"
+    PYARROW_LOWER_BOUND_SKIP,
+    reason="deltalake not supported on older versions of pyarrow",
 )
 
 
@@ -65,7 +49,10 @@ def test_deltalake_read_row_group_splits(tmp_path, base_table):
     deltalake.write_deltalake(path, base_table, min_rows_per_group=1, max_rows_per_group=2)
 
     # Force file splitting
-    with split_small_pq_files():
+    with daft.execution_config_ctx(
+        scan_tasks_min_size_bytes=1,
+        scan_tasks_max_size_bytes=100,
+    ):
         df = daft.read_deltalake(str(path))
         df.collect()
         assert len(df) == 3, "Length of non-materialized data when read through deltalake should be correct"
@@ -79,7 +66,10 @@ def test_deltalake_read_row_group_splits_with_filter(tmp_path, base_table):
     deltalake.write_deltalake(path, base_table, min_rows_per_group=1, max_rows_per_group=2)
 
     # Force file splitting
-    with split_small_pq_files():
+    with daft.execution_config_ctx(
+        scan_tasks_min_size_bytes=1,
+        scan_tasks_max_size_bytes=100,
+    ):
         df = daft.read_deltalake(str(path))
         df = df.where(df["a"] > 1)
         df.collect()
@@ -94,8 +84,33 @@ def test_deltalake_read_row_group_splits_with_limit(tmp_path, base_table):
     deltalake.write_deltalake(path, base_table, min_rows_per_group=1, max_rows_per_group=2)
 
     # Force file splitting
-    with split_small_pq_files():
+    with daft.execution_config_ctx(
+        scan_tasks_min_size_bytes=1,
+        scan_tasks_max_size_bytes=100,
+    ):
         df = daft.read_deltalake(str(path))
         df = df.limit(2)
         df.collect()
         assert len(df) == 2, "Length of non-materialized data when read through deltalake should be correct"
+
+
+def test_deltalake_read_versioned(tmp_path, base_table):
+    deltalake = pytest.importorskip("deltalake")
+    path = tmp_path / "some_table"
+    deltalake.write_deltalake(path, base_table)
+
+    updated_columns = base_table.columns + [pa.array(["x", "y", "z"])]
+    updated_column_names = base_table.column_names + ["new_column"]
+    updated_table = pa.Table.from_arrays(updated_columns, names=updated_column_names)
+    deltalake.write_deltalake(path, updated_table, mode="overwrite", schema_mode="overwrite")
+
+    for version in [None, 1]:
+        df = daft.read_deltalake(str(path), version=version)
+        expected_schema = Schema.from_pyarrow_schema(deltalake.DeltaTable(path).schema().to_pyarrow())
+        assert df.schema() == expected_schema
+        assert_pyarrow_tables_equal(df.to_arrow(), updated_table)
+
+    df = daft.read_deltalake(str(path), version=0)
+    expected_schema = Schema.from_pyarrow_schema(deltalake.DeltaTable(path, version=0).schema().to_pyarrow())
+    assert df.schema() == expected_schema
+    assert_pyarrow_tables_equal(df.to_arrow(), base_table)
