@@ -11,7 +11,7 @@ import ray
 
 import daft
 from daft import DataType
-from daft.context import get_context
+from tests.conftest import get_tests_daft_runner_name
 
 RAY_VERSION = tuple(int(s) for s in ray.__version__.split(".")[0:3])
 
@@ -36,7 +36,7 @@ def _row_to_pydict(row: ray.data.row.TableRow | dict) -> dict:
     return row.as_pydict()
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_to_ray_dataset_all_arrow(n_partitions: int):
     df = daft.from_pydict(DATA).repartition(n_partitions)
@@ -60,7 +60,7 @@ def test_to_ray_dataset_all_arrow(n_partitions: int):
     )
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.skipif(
     RAY_VERSION >= (2, 5, 0), reason="Ray Datasets versions >= 2.5.0 no longer support Python objects as rows"
 )
@@ -87,7 +87,7 @@ def test_to_ray_dataset_with_py(n_partitions: int):
     )
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_to_ray_dataset_with_numpy(n_partitions: int):
     df = daft.from_pydict(DATA).repartition(n_partitions)
@@ -119,7 +119,7 @@ def test_to_ray_dataset_with_numpy(n_partitions: int):
     )
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.skipif(RAY_VERSION < (2, 2, 0), reason="Variable-shaped tensor columns not supported in Ray < 2.1.0")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_to_ray_dataset_with_numpy_variable_shaped(n_partitions: int):
@@ -151,7 +151,7 @@ def test_to_ray_dataset_with_numpy_variable_shaped(n_partitions: int):
     )
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_from_ray_dataset_all_arrow(n_partitions: int):
     def add_float(table: pa.Table) -> pa.Table:
@@ -182,7 +182,7 @@ def test_from_ray_dataset_all_arrow(n_partitions: int):
     assert out_table.equals(expected_table), (out_table, expected_table)
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_from_ray_dataset_simple(n_partitions: int):
     ds = ray.data.range(8, parallelism=n_partitions)
@@ -195,7 +195,7 @@ def test_from_ray_dataset_simple(n_partitions: int):
     assert sorted(out[key]) == list(range(8))
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_from_ray_dataset_tensor(n_partitions: int):
     ds = ray.data.range(8)
@@ -220,7 +220,7 @@ def test_from_ray_dataset_tensor(n_partitions: int):
     np.testing.assert_equal(out_sorted, expected)
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_from_ray_dataset_pandas(n_partitions: int):
     def add_float(df: pd.DataFrame) -> pd.DataFrame:
@@ -242,7 +242,7 @@ def test_from_ray_dataset_pandas(n_partitions: int):
     pd.testing.assert_frame_equal(df.to_pandas(), expected_df)
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_from_ray_dataset_preview(n_partitions: int):
     ds = ray.data.range(3, parallelism=n_partitions)
@@ -252,7 +252,7 @@ def test_from_ray_dataset_preview(n_partitions: int):
     assert len(df._preview.preview_partition) == 3
 
 
-@pytest.mark.skipif(get_context().runner_config.name != "ray", reason="Needs to run on Ray runner")
+@pytest.mark.skipif(get_tests_daft_runner_name() != "ray", reason="Needs to run on Ray runner")
 @pytest.mark.parametrize("n_partitions", [1, 2])
 def test_from_ray_dataset_data_longer_than_preview(n_partitions: int):
     ds = ray.data.range(10, parallelism=n_partitions)
