@@ -5,9 +5,9 @@ import pyarrow as pa
 import pytest
 
 from daft import Series
-from daft.context import get_context
 from daft.datatype import DataType
 from daft.utils import pyarrow_supports_fixed_shape_tensor
+from tests.conftest import get_tests_daft_runner_name
 
 ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
 
@@ -368,7 +368,7 @@ def test_series_if_else_struct(if_true, if_false, expected) -> None:
 
 
 @pytest.mark.skipif(
-    get_context().runner_config.name == "ray",
+    get_tests_daft_runner_name() == "ray",
     reason="pyarrow extension types aren't supported on Ray clusters.",
 )
 @pytest.mark.parametrize(
