@@ -9,19 +9,12 @@ import tempfile
 import pytest
 
 import daft
-from daft.daft import NativeStorageConfig, PythonStorageConfig, StorageConfig
+from daft.daft import CsvParseOptions
 from daft.datatype import DataType
 from daft.logical.schema import Schema
 from daft.runners.partitioning import TableParseCSVOptions, TableReadOptions
-from daft.series import ARROW_VERSION
-from daft.table import MicroPartition, schema_inference, table_io
-
-
-def storage_config_from_use_native_downloader(use_native_downloader: bool) -> StorageConfig:
-    if use_native_downloader:
-        return StorageConfig.native(NativeStorageConfig(True, None))
-    else:
-        return StorageConfig.python(PythonStorageConfig(None))
+from daft.table import MicroPartition, table_io
+from daft.utils import get_arrow_version
 
 
 def test_read_input(tmpdir):
@@ -39,9 +32,6 @@ def test_read_input(tmpdir):
     assert table_io.read_csv(tmpdir / "file.csv", schema=schema).to_pydict() == data
     assert table_io.read_csv(str(tmpdir / "file.csv"), schema=schema).to_pydict() == data
 
-    with open(tmpdir / "file.csv", "rb") as f:
-        assert table_io.read_csv(f, schema=schema).to_pydict() == data
-
 
 @contextlib.contextmanager
 def _csv_write_helper(header: list[str] | None, data: list[list[str | None]], **kwargs):
@@ -64,8 +54,7 @@ def _csv_write_helper(header: list[str] | None, data: list[list[str | None]], **
         ("True", DataType.bool()),
     ],
 )
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_infer_schema(data, expected_dtype, use_native_downloader):
+def test_csv_infer_schema(data, expected_dtype):
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -74,13 +63,11 @@ def test_csv_infer_schema(data, expected_dtype, use_native_downloader):
             ["3", None],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        schema = schema_inference.from_csv(f, storage_config=storage_config)
+        schema = Schema.from_csv(f)
         assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", expected_dtype)])
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_infer_schema_custom_delimiter(use_native_downloader):
+def test_csv_infer_schema_custom_delimiter():
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -90,15 +77,11 @@ def test_csv_infer_schema_custom_delimiter(use_native_downloader):
         ],
         delimiter="|",
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        schema = schema_inference.from_csv(
-            f, storage_config=storage_config, csv_options=TableParseCSVOptions(delimiter="|")
-        )
+        schema = Schema.from_csv(f, parse_options=CsvParseOptions(delimiter="|"))
         assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_infer_schema_no_header(use_native_downloader):
+def test_csv_infer_schema_no_header():
     with _csv_write_helper(
         header=None,
         data=[
@@ -107,15 +90,8 @@ def test_csv_infer_schema_no_header(use_native_downloader):
             ["3", None],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        schema = schema_inference.from_csv(
-            f, storage_config=storage_config, csv_options=TableParseCSVOptions(header_index=None)
-        )
-        fields = (
-            [("column_1", DataType.int64()), ("column_2", DataType.int64())]
-            if use_native_downloader
-            else [("f0", DataType.int64()), ("f1", DataType.int64())]
-        )
+        schema = Schema.from_csv(f, parse_options=CsvParseOptions(has_header=False))
+        fields = [("column_1", DataType.int64()), ("column_2", DataType.int64())]
         assert schema == Schema._from_field_name_and_types(fields)
 
 
@@ -129,8 +105,7 @@ def test_csv_infer_schema_no_header(use_native_downloader):
         ("True", daft.Series.from_pylist([True, True, None])),
     ],
 )
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data(data, expected_data_series, use_native_downloader):
+def test_csv_read_data(data, expected_data_series):
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -139,7 +114,6 @@ def test_csv_read_data(data, expected_data_series, use_native_downloader):
             ["3", None],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
         schema = Schema._from_field_name_and_types(
             [("id", DataType.int64()), ("data", expected_data_series.datatype())]
         )
@@ -149,12 +123,11 @@ def test_csv_read_data(data, expected_data_series, use_native_downloader):
                 "data": expected_data_series,
             }
         )
-        table = table_io.read_csv(f, schema, storage_config=storage_config)
+        table = table_io.read_csv(f, schema)
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_csv_limit_rows(use_native_downloader):
+def test_csv_read_data_csv_limit_rows():
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -163,8 +136,6 @@ def test_csv_read_data_csv_limit_rows(use_native_downloader):
             ["3", None],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
         expected = MicroPartition.from_pydict(
             {
@@ -175,14 +146,12 @@ def test_csv_read_data_csv_limit_rows(use_native_downloader):
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             read_options=TableReadOptions(num_rows=2),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_csv_select_columns(use_native_downloader):
+def test_csv_read_data_csv_select_columns():
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -191,8 +160,6 @@ def test_csv_read_data_csv_select_columns(use_native_downloader):
             ["3", None],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
         expected = MicroPartition.from_pydict(
             {
@@ -202,14 +169,12 @@ def test_csv_read_data_csv_select_columns(use_native_downloader):
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             read_options=TableReadOptions(column_names=["data"]),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_csv_custom_delimiter(use_native_downloader):
+def test_csv_read_data_csv_custom_delimiter():
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -219,8 +184,6 @@ def test_csv_read_data_csv_custom_delimiter(use_native_downloader):
         ],
         delimiter="|",
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
         expected = MicroPartition.from_pydict(
             {
@@ -231,14 +194,12 @@ def test_csv_read_data_csv_custom_delimiter(use_native_downloader):
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(delimiter="|"),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_csv_no_header(use_native_downloader):
+def test_csv_read_data_csv_no_header():
     with _csv_write_helper(
         header=None,
         data=[
@@ -247,8 +208,6 @@ def test_csv_read_data_csv_no_header(use_native_downloader):
             ["3", None],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
         expected = MicroPartition.from_pydict(
             {
@@ -259,15 +218,13 @@ def test_csv_read_data_csv_no_header(use_native_downloader):
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(header_index=None),
             read_options=TableReadOptions(column_names=["id", "data"]),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_csv_custom_quote(use_native_downloader):
+def test_csv_read_data_csv_custom_quote():
     with _csv_write_helper(
         header=["'id'", "'data'"],
         data=[
@@ -276,8 +233,6 @@ def test_csv_read_data_csv_custom_quote(use_native_downloader):
             ["3", "aa"],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.string())])
         expected = MicroPartition.from_pydict(
             {
@@ -288,16 +243,13 @@ def test_csv_read_data_csv_custom_quote(use_native_downloader):
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(quote="'"),
         )
 
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-# TODO this test still fails with use_native_downloader = True
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_custom_escape(use_native_downloader):
+def test_csv_read_data_custom_escape():
     with _csv_write_helper(
         header=["id", "data"],
         data=[
@@ -310,8 +262,6 @@ def test_csv_read_data_custom_escape(use_native_downloader):
         doublequote=False,
         quoting=csv.QUOTE_ALL,
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.string())])
         expected = MicroPartition.from_pydict(
             {
@@ -322,16 +272,13 @@ def test_csv_read_data_custom_escape(use_native_downloader):
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(escape_char="\\", double_quote=False),
         )
 
         assert table.to_arrow() == expected.to_arrow(), f"Received:\n{table}\n\nExpected:\n{expected}"
 
 
-# TODO Not testing use_native_downloader = False, as pyarrow does not support comments directly
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_csv_read_data_custom_comment(use_native_downloader):
+def test_csv_read_data_custom_comment():
     with tempfile.TemporaryDirectory() as directory_name:
         file = os.path.join(directory_name, "tempfile")
         with open(file, "w", newline="") as f:
@@ -341,8 +288,6 @@ def test_csv_read_data_custom_comment(use_native_downloader):
             f.write("# comment line\n")
             writer.writerow(["3", "aa"])
 
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
         schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.string())])
         expected = MicroPartition.from_pydict(
             {
@@ -351,11 +296,11 @@ def test_csv_read_data_custom_comment(use_native_downloader):
             }
         )
         # Skipping test for arrow < 7.0.0 as comments are not supported in pyarrow
-        if ARROW_VERSION >= (7, 0, 0):
+        arrow_version = get_arrow_version()
+        if arrow_version >= (7, 0, 0):
             table = table_io.read_csv(
                 file,
                 schema,
-                storage_config=storage_config,
                 csv_options=TableParseCSVOptions(comment="#"),
             )
             assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
@@ -369,8 +314,6 @@ def test_csv_read_data_variable_missing_columns():
             ["2", "2"],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(True)
-
         schema = Schema._from_field_name_and_types(
             [
                 ("id", DataType.int64()),
@@ -386,7 +329,6 @@ def test_csv_read_data_variable_missing_columns():
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(allow_variable_columns=True),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
@@ -400,8 +342,6 @@ def test_csv_read_data_variable_extra_columns():
             ["2", "2", "2"],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(True)
-
         schema = Schema._from_field_name_and_types(
             [
                 ("id", DataType.int64()),
@@ -417,7 +357,6 @@ def test_csv_read_data_variable_extra_columns():
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(allow_variable_columns=True),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
@@ -431,8 +370,6 @@ def test_csv_read_data_variable_columns_with_non_matching_types():
             ["2", "2"],
         ],
     ) as f:
-        storage_config = storage_config_from_use_native_downloader(True)
-
         schema = Schema._from_field_name_and_types(
             [
                 ("id", DataType.int64()),
@@ -448,7 +385,6 @@ def test_csv_read_data_variable_columns_with_non_matching_types():
         table = table_io.read_csv(
             f,
             schema,
-            storage_config=storage_config,
             csv_options=TableParseCSVOptions(allow_variable_columns=True),
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
