@@ -5,21 +5,21 @@ import datetime
 import os
 import pathlib
 import tempfile
+from pathlib import Path
 
 import pyarrow as pa
 import pyarrow.parquet as papq
 import pytest
 
 import daft
-from daft.daft import NativeStorageConfig, PythonStorageConfig, StorageConfig
 from daft.datatype import DataType, TimeUnit
+from daft.exceptions import DaftCoreException
 from daft.logical.schema import Schema
 from daft.runners.partitioning import TableParseParquetOptions, TableReadOptions
 from daft.table import (
     MicroPartition,
     read_parquet_into_pyarrow,
     read_parquet_into_pyarrow_bulk,
-    schema_inference,
     table_io,
 )
 
@@ -27,13 +27,6 @@ PYARROW_GE_11_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumer
 PYARROW_GE_13_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (13, 0, 0)
 
 
-def storage_config_from_use_native_downloader(use_native_downloader: bool) -> StorageConfig:
-    if use_native_downloader:
-        return StorageConfig.native(NativeStorageConfig(True, None))
-    else:
-        return StorageConfig.python(PythonStorageConfig(None))
-
-
 def test_read_input(tmpdir):
     tmpdir = pathlib.Path(tmpdir)
     data = pa.Table.from_pydict({"foo": [1, 2, 3]})
@@ -46,12 +39,9 @@ def test_read_input(tmpdir):
     assert table_io.read_parquet(tmpdir / "file.parquet", schema=schema).to_arrow() == data
     assert table_io.read_parquet(str(tmpdir / "file.parquet"), schema=schema).to_arrow() == data
 
-    with open(tmpdir / "file.parquet", "rb") as f:
-        assert table_io.read_parquet(f, schema=schema).to_arrow() == data
-
 
 @contextlib.contextmanager
-def _parquet_write_helper(data: pa.Table, row_group_size: int = None, papq_write_table_kwargs: dict = {}):
+def _parquet_write_helper(data: pa.Table, row_group_size: int | None = None, papq_write_table_kwargs: dict = {}):
     with tempfile.TemporaryDirectory() as directory_name:
         file = os.path.join(directory_name, "tempfile")
         papq.write_table(data, file, row_group_size=row_group_size, **papq_write_table_kwargs)
@@ -74,14 +64,7 @@ def _parquet_write_helper(data: pa.Table, row_group_size: int = None, papq_write
         ([1, None, 2], DataType.list(DataType.int64())),
     ],
 )
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_parquet_infer_schema(data, expected_dtype, use_native_downloader):
-    # HACK: Pyarrow 13 changed their schema parsing behavior so we receive DataType.list(..) instead of DataType.list(..)
-    # However, our native downloader still parses DataType.list(..) regardless of PyArrow version
-    if PYARROW_GE_13_0_0 and not use_native_downloader and expected_dtype == DataType.list(DataType.int64()):
-        expected_dtype = DataType.list(DataType.int64())
-    storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-
+def test_parquet_infer_schema(data, expected_dtype):
     with _parquet_write_helper(
         pa.Table.from_pydict(
             {
@@ -90,17 +73,15 @@ def test_parquet_infer_schema(data, expected_dtype, use_native_downloader):
             }
         )
     ) as f:
-        schema = schema_inference.from_parquet(f, storage_config=storage_config)
+        schema = Schema.from_parquet(f)
         assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", expected_dtype)])
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_parquet_read_empty(use_native_downloader):
+def test_parquet_read_empty():
     with _parquet_write_helper(pa.Table.from_pydict({"foo": pa.array([], type=pa.int64())})) as f:
         schema = Schema._from_field_name_and_types([("foo", DataType.int64())])
         expected = MicroPartition.from_pydict({"foo": pa.array([], type=pa.int64())})
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        table = table_io.read_parquet(f, schema, storage_config=storage_config)
+        table = table_io.read_parquet(f, schema)
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
@@ -121,8 +102,7 @@ def test_parquet_read_empty(use_native_downloader):
         ({"foo": 1}, daft.Series.from_pylist([{"foo": 1}, {"foo": 1}, None])),
     ],
 )
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_parquet_read_data(data, expected_data_series, use_native_downloader):
+def test_parquet_read_data(data, expected_data_series):
     with _parquet_write_helper(
         pa.Table.from_pydict(
             {
@@ -140,14 +120,12 @@ def test_parquet_read_data(data, expected_data_series, use_native_downloader):
                 "data": expected_data_series,
             }
         )
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        table = table_io.read_parquet(f, schema, storage_config=storage_config)
+        table = table_io.read_parquet(f, schema)
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
 @pytest.mark.parametrize("row_group_size", [None, 1, 3])
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_parquet_read_data_limit_rows(row_group_size, use_native_downloader):
+def test_parquet_read_data_limit_rows(row_group_size):
     with _parquet_write_helper(
         pa.Table.from_pydict(
             {
@@ -164,10 +142,7 @@ def test_parquet_read_data_limit_rows(row_group_size, use_native_downloader):
                 "data": [1, 2],
             }
         )
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        table = table_io.read_parquet(
-            f, schema, read_options=TableReadOptions(num_rows=2), storage_config=storage_config
-        )
+        table = table_io.read_parquet(f, schema, read_options=TableReadOptions(num_rows=2))
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
@@ -178,8 +153,7 @@ def test_parquet_read_data_multi_row_groups():
     assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
-def test_parquet_read_data_select_columns(use_native_downloader):
+def test_parquet_read_data_select_columns():
     with _parquet_write_helper(
         pa.Table.from_pydict(
             {
@@ -194,10 +168,7 @@ def test_parquet_read_data_select_columns(use_native_downloader):
                 "data": [1, 2, None],
             }
         )
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
-        table = table_io.read_parquet(
-            f, schema, read_options=TableReadOptions(column_names=["data"]), storage_config=storage_config
-        )
+        table = table_io.read_parquet(f, schema, read_options=TableReadOptions(column_names=["data"]))
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
@@ -206,9 +177,8 @@ def test_parquet_read_data_select_columns(use_native_downloader):
 ###
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
 @pytest.mark.parametrize("use_deprecated_int96_timestamps", [True, False])
-def test_parquet_read_int96_timestamps(use_deprecated_int96_timestamps, use_native_downloader):
+def test_parquet_read_int96_timestamps(use_deprecated_int96_timestamps):
     data = {
         "timestamp_ms": pa.array([1, 2, 3], pa.timestamp("ms")),
         "timestamp_us": pa.array([1, 2, 3], pa.timestamp("us")),
@@ -235,19 +205,16 @@ def test_parquet_read_int96_timestamps(use_deprecated_int96_timestamps, use_nati
     ) as f:
         schema = Schema._from_field_name_and_types(schema)
         expected = MicroPartition.from_pydict(data)
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
         table = table_io.read_parquet(
             f,
             schema,
             read_options=TableReadOptions(column_names=schema.column_names()),
-            storage_config=storage_config,
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
 
-@pytest.mark.parametrize("use_native_downloader", [True, False])
 @pytest.mark.parametrize("coerce_to", [TimeUnit.ms(), TimeUnit.us()])
-def test_parquet_read_int96_timestamps_overflow(coerce_to, use_native_downloader):
+def test_parquet_read_int96_timestamps_overflow(coerce_to):
     # NOTE: datetime.datetime(3000, 1, 1) and datetime.datetime(1000, 1, 1) cannot be represented by our timestamp64(nanosecond)
     # type. However they can be written to Parquet's INT96 type. Here we test that a round-trip is possible if provided with
     # the appropriate flags.
@@ -273,13 +240,11 @@ def test_parquet_read_int96_timestamps_overflow(coerce_to, use_native_downloader
     ) as f:
         schema = Schema._from_field_name_and_types(schema)
         expected = MicroPartition.from_pydict(data)
-        storage_config = storage_config_from_use_native_downloader(use_native_downloader)
         table = table_io.read_parquet(
             f,
             schema,
             read_options=TableReadOptions(column_names=schema.column_names()),
             parquet_options=TableParseParquetOptions(coerce_int96_timestamp_unit=coerce_to),
-            storage_config=storage_config,
         )
         assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"
 
@@ -397,3 +362,22 @@ def test_read_parquet_file_missing_column_partial_read_with_pyarrow_bulk(tmpdir)
     read_back = read_parquet_into_pyarrow_bulk([file_path.as_posix()], columns=["x", "MISSING"])
     assert len(read_back) == 1
     assert tab.drop("y") == read_back[0]  # only read "x"
+
+
+@pytest.mark.parametrize(
+    "parquet_path", [Path(__file__).parents[2] / "assets" / "parquet-data" / "invalid_utf8.parquet"]
+)
+def test_parquet_read_string_utf8_into_binary(parquet_path: Path):
+    import pyarrow as pa
+
+    assert parquet_path.exists()
+
+    with pytest.raises(DaftCoreException, match="invalid utf-8 sequence"):
+        read_parquet_into_pyarrow(path=parquet_path.as_posix())
+
+    read_back = read_parquet_into_pyarrow(path=parquet_path.as_posix(), string_encoding="raw")
+    schema = read_back.schema
+    assert len(schema) == 1
+    assert schema[0].name == "invalid_string"
+    assert schema[0].type == pa.binary()
+    assert read_back["invalid_string"][0].as_py() == b"\x80\x80\x80"
