@@ -9,11 +9,12 @@ import pyarrow as pa
 import pyarrow.compute as pac
 import pytest
 
+import daft
 from daft import DataType, TimeUnit
-from daft.context import get_context
 from daft.series import Series
 from daft.table import MicroPartition
 from daft.utils import pyarrow_supports_fixed_shape_tensor
+from tests.conftest import get_tests_daft_runner_name
 
 ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
 
@@ -179,7 +180,7 @@ if pyarrow_supports_fixed_shape_tensor():
 
 
 def _with_uuid_ext_type(uuid_ext_type) -> tuple[dict, dict]:
-    if get_context().runner_config.name == "ray":
+    if get_tests_daft_runner_name() == "ray":
         # pyarrow extension types aren't supported in Ray clusters yet.
         return ARROW_ROUNDTRIP_TYPES, ARROW_TYPE_ARRAYS
     arrow_roundtrip_types = ARROW_ROUNDTRIP_TYPES.copy()
@@ -318,7 +319,7 @@ def test_from_pydict_arrow_struct_array() -> None:
 
 
 @pytest.mark.skipif(
-    get_context().runner_config.name == "ray",
+    get_tests_daft_runner_name() == "ray",
     reason="pyarrow extension types aren't supported on Ray clusters.",
 )
 def test_from_pydict_arrow_extension_array(uuid_ext_type) -> None:
@@ -527,7 +528,7 @@ def test_from_arrow_map_array() -> None:
 
 
 @pytest.mark.skipif(
-    get_context().runner_config.name == "ray",
+    get_tests_daft_runner_name() == "ray",
     reason="pyarrow extension types aren't supported on Ray clusters.",
 )
 def test_from_arrow_extension_array(uuid_ext_type) -> None:
@@ -649,3 +650,23 @@ def test_nested_struct_dates(levels: int) -> None:
 
     assert back_again.to_arrow().type == expected_arrow_type
     assert back_again.to_pylist() == data
+
+
+def test_from_arrow_iterable() -> None:
+    class CustomIterable:
+        def __iter__(self):
+            yield pa.table({"text": ["foo1", "bar2"]})
+            yield pa.table({"text": ["foo2", "bar2"]})
+            yield pa.table({"text": ["foo3", "bar3"]})
+
+    my_iter = CustomIterable()
+
+    table = daft.from_arrow(my_iter)
+    tbl = table.to_pydict()
+    assert tbl == {"text": ["foo1", "bar2", "foo2", "bar2", "foo3", "bar3"]}
+
+
+def test_to_arrow_iterator() -> None:
+    df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
+    it = df.to_arrow_iter()
+    assert isinstance(next(it), pa.RecordBatch)
